/*
 * map_cmdu.c: MultiAP cmdu handling
 *
 * Author: Vivek Dutta <vivek.dutta@iopsys.eu>
 * Author: Yashvardhan <y.yashvardhan@iopsys.eu>
 *
 * Copyright (C) 2020, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "map2.h"
#include "maputils.h"
#include "map_common.h"

// Change this number if number of tlvs in a cmdu
// exceeds
#define MAX_TLV_PER_CMDU  (15)

enum {
	TLIST_EMPTY,
	TLIST_ZEROORMORE_TLVS,
	TLIST_ZEROORONE_TLV,
	TLIST_ONE_TLV,
	TLIST_ONEORMORE_TLVS,
	TLIST_INVALID
};

struct tlv_list {
	uint8_t tlv_type;
	uint8_t count;
};

struct cmdu_tlv_association {
	uint16_t cmdu_type;
	bool relay;
	uint8_t transmission_type;
	struct tlv_list profile1[MAX_TLV_PER_CMDU];
	struct tlv_list profile2[MAX_TLV_PER_CMDU];
};

// Static/Internal data-structures and functions
static const struct cmdu_tlv_association CMDU_VALIDATOR_TEMPLATE[] = {
	{
		.cmdu_type = CMDU_TYPE_TOPOLOGY_DISCOVERY,
		.relay = false,
		.transmission_type = TRANSMISSION_NEIGHBOR_MULTICAST,
		.profile1 = {
			[0] = {TLV_TYPE_AL_MAC_ADDRESS_TYPE, TLIST_ONE_TLV},
			[1] = {TLV_TYPE_MAC_ADDRESS_TYPE, TLIST_ONE_TLV},
			[2 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {TLV_TYPE_AL_MAC_ADDRESS_TYPE, TLIST_ONE_TLV},
			[1] = {TLV_TYPE_MAC_ADDRESS_TYPE, TLIST_ONE_TLV},
			[2 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_TYPE_TOPOLOGY_NOTIFICATION,
		.relay = true,
		.transmission_type = TRANSMISSION_RELIABLE_MULTICAST,
		.profile1 = {
			[0] = {TLV_TYPE_AL_MAC_ADDRESS_TYPE, TLIST_ONE_TLV},
			[1] = {MAP_TLV_CLIENT_ASSOCIATION_EVENT,
				TLIST_ZEROORONE_TLV},
			[2 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {TLV_TYPE_AL_MAC_ADDRESS_TYPE, TLIST_ONE_TLV},
			[1] = {MAP_TLV_CLIENT_ASSOCIATION_EVENT,
				TLIST_ZEROORONE_TLV},
			[2 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_TYPE_TOPOLOGY_QUERY,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_MULTIAP_PROFILE, TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_TYPE_TOPOLOGY_RESPONSE,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {TLV_TYPE_DEVICE_INFORMATION_TYPE, TLIST_ONE_TLV},
			[1] = {TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES,
				TLIST_ZEROORMORE_TLVS},
			[2] = {TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST,
				TLIST_ZEROORMORE_TLVS},
			[3] = {TLV_TYPE_NEIGHBOR_DEVICE_LIST,
				TLIST_ZEROORMORE_TLVS},
			[4] = {MAP_TLV_SUPPORTED_SERVICE, TLIST_ZEROORONE_TLV},
			[5] = {MAP_TLV_AP_OPERATIONAL_BSS, TLIST_ONE_TLV},
			[6] = {MAP_TLV_ASSOCIATED_CLIENTS,
				TLIST_ZEROORONE_TLV},
			[7 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {TLV_TYPE_DEVICE_INFORMATION_TYPE,
				TLIST_ONE_TLV},
			[1] = {TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES,
				TLIST_ZEROORMORE_TLVS},
			[2] = {TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST,
				TLIST_ZEROORMORE_TLVS},
			[3] = {TLV_TYPE_NEIGHBOR_DEVICE_LIST,
				TLIST_ZEROORMORE_TLVS},
			[4] = {MAP_TLV_SUPPORTED_SERVICE, TLIST_ZEROORONE_TLV},
			[5] = {MAP_TLV_AP_OPERATIONAL_BSS, TLIST_ONE_TLV},
			[6] = {MAP_TLV_ASSOCIATED_CLIENTS,
				TLIST_ZEROORONE_TLV},
			[7] = {MAP_TLV_MULTIAP_PROFILE, TLIST_ONE_TLV},
			[8 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_TYPE_VENDOR_SPECIFIC,
		.relay = false,
		.transmission_type = TRANSMISSION_NEIGHBOR_MULTICAST,
		.profile1 = {
			[0] = {TLV_TYPE_VENDOR_SPECIFIC, TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {TLV_TYPE_VENDOR_SPECIFIC, TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH,
		.relay = true,
		.transmission_type = TRANSMISSION_RELAYED_MULTICAST,
		.profile1 = {
			[0] = {TLV_TYPE_AL_MAC_ADDRESS_TYPE, TLIST_ONE_TLV},
			[1] = {TLV_TYPE_SEARCHED_ROLE, TLIST_ONE_TLV},
			[2] = {TLV_TYPE_AUTOCONFIG_FREQ_BAND, TLIST_ONE_TLV},
			[3] = {MAP_TLV_SUPPORTED_SERVICE, TLIST_ZEROORONE_TLV},
			[4] = {MAP_TLV_SEARCHED_SERVICE, TLIST_ZEROORONE_TLV},
			[5 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {TLV_TYPE_AL_MAC_ADDRESS_TYPE, TLIST_ONE_TLV},
			[1] = {TLV_TYPE_SEARCHED_ROLE, TLIST_ONE_TLV},
			[2] = {TLV_TYPE_AUTOCONFIG_FREQ_BAND, TLIST_ONE_TLV},
			[3] = {MAP_TLV_SUPPORTED_SERVICE, TLIST_ZEROORONE_TLV},
			[4] = {MAP_TLV_SEARCHED_SERVICE, TLIST_ZEROORONE_TLV},
			[5] = {MAP_TLV_MULTIAP_PROFILE, TLIST_ONE_TLV},
			[6 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {TLV_TYPE_SUPPORTED_ROLE, TLIST_ONE_TLV},
			[1] = {TLV_TYPE_SUPPORTED_FREQ_BAND, TLIST_ONE_TLV},
			[2] = {MAP_TLV_SUPPORTED_SERVICE, TLIST_ZEROORONE_TLV},
			[3 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {TLV_TYPE_SUPPORTED_ROLE, TLIST_ONE_TLV},
			[1] = {TLV_TYPE_SUPPORTED_FREQ_BAND, TLIST_ONE_TLV},
			[2] = {MAP_TLV_SUPPORTED_SERVICE, TLIST_ZEROORONE_TLV},
			[3] = {MAP_TLV_MULTIAP_PROFILE, TLIST_ONE_TLV},
			[4 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_TYPE_AP_AUTOCONFIGURATION_WSC,
		/* manual verification required for this cmdu, as the
		 * validation differs for controller and agent.
		 */
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_TYPE_AP_AUTOCONFIGURATION_RENEW,
		.relay = true,
		.transmission_type = TRANSMISSION_RELAYED_MULTICAST,
		.profile1 = {
			[0] = {TLV_TYPE_AL_MAC_ADDRESS_TYPE, TLIST_ONE_TLV},
			[1] = {TLV_TYPE_SUPPORTED_ROLE, TLIST_ONE_TLV},
			[2] = {TLV_TYPE_SUPPORTED_FREQ_BAND, TLIST_ONE_TLV},
			[3 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {TLV_TYPE_AL_MAC_ADDRESS_TYPE, TLIST_ONE_TLV},
			[1] = {TLV_TYPE_SUPPORTED_ROLE, TLIST_ONE_TLV},
			[2] = {TLV_TYPE_SUPPORTED_FREQ_BAND, TLIST_ONE_TLV},
			[3 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	// MAP profile-1 CMDUs
	{
		.cmdu_type = CMDU_1905_ACK,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_ERROR_CODE, TLIST_ZEROORMORE_TLVS},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_ERROR_CODE, TLIST_ZEROORMORE_TLVS},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_AP_CAPABILITY_QUERY,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_AP_CAPABILITY_REPORT,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_AP_CAPABILITY, TLIST_ONE_TLV},
			[1] = {MAP_TLV_AP_RADIO_BASIC_CAPABILITIES,
				TLIST_ONEORMORE_TLVS},
			[2] = {MAP_TLV_AP_HT_CAPABILITIES,
				TLIST_ZEROORMORE_TLVS},
			[3] = {MAP_TLV_AP_VHT_CAPABILITIES,
				TLIST_ZEROORMORE_TLVS},
			[4] = {MAP_TLV_AP_HE_CAPABILITIES,
				TLIST_ZEROORMORE_TLVS},
			[5 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_AP_CAPABILITY, TLIST_ONE_TLV},
			[1] = {MAP_TLV_AP_RADIO_BASIC_CAPABILITIES,
				TLIST_ONEORMORE_TLVS},
			[2] = {MAP_TLV_AP_HT_CAPABILITIES,
				TLIST_ZEROORMORE_TLVS},
			[3] = {MAP_TLV_AP_VHT_CAPABILITIES,
				TLIST_ZEROORMORE_TLVS},
			[4] = {MAP_TLV_AP_HE_CAPABILITIES,
				TLIST_ZEROORMORE_TLVS},
			[5] = {MAP_TLV_CHANNEL_SCAN_CAPABILITY,
				TLIST_ONE_TLV},
			[6] = {MAP_TLV_CAC_CAPABILITY,
				TLIST_ONE_TLV},
			[7] = {MAP_TLV_PROFILE2_AP_CAPABILITY,
				TLIST_ONE_TLV},
			[8] = {MAP_TLV_METRIC_COLLECTION_INTERVAL,
				TLIST_ONE_TLV},
			[9 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_POLICY_CONFIG_REQ,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_STEERING_POLICY,
				TLIST_ZEROORONE_TLV},
			[1] = {MAP_TLV_METRIC_REPORTING_POLICY,
				TLIST_ZEROORONE_TLV},
			[2 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_STEERING_POLICY, TLIST_ZEROORONE_TLV},
			[1] = {MAP_TLV_METRIC_REPORTING_POLICY,
				TLIST_ZEROORONE_TLV},
			[2] = {MAP_TLV_DEFAULT_8021Q_SETTINGS,
				TLIST_ZEROORONE_TLV},
			[3] = {MAP_TLV_TRAFFIC_SEPARATION_POLICY,
				TLIST_ZEROORONE_TLV},
			[4] = {MAP_TLV_CHANNEL_SCAN_REPORTING_POLICY,
				TLIST_ZEROORONE_TLV},
			[5] = {MAP_TLV_UNSUCCESS_ASSOCIATION_POLICY,
				TLIST_ZEROORONE_TLV},
			[6] = {MAP_TLV_BACKHAUL_BSS_CONFIG,
				TLIST_ZEROORMORE_TLVS},
			[7 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_CHANNEL_PREFERENCE_QUERY,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_CHANNEL_PREFERENCE_REPORT,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_CHANNEL_PREFERENCE,
				TLIST_ZEROORMORE_TLVS},
			[1] = {MAP_TLV_RADIO_OPERATION_RESTRICTION,
				TLIST_ZEROORMORE_TLVS},
			[2 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_CHANNEL_PREFERENCE,
				TLIST_ZEROORMORE_TLVS},
			[1] = {MAP_TLV_RADIO_OPERATION_RESTRICTION,
				TLIST_ZEROORMORE_TLVS},
			[2] = {MAP_TLV_CAC_COMPLETION_REPORT,
				TLIST_ZEROORONE_TLV},
			[3] = {MAP_TLV_CAC_STATUS_REPORT,
				TLIST_ONE_TLV},
			[4 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_CHANNEL_SELECTION_REQ,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_CHANNEL_PREFERENCE,
				TLIST_ZEROORMORE_TLVS},
			[1] = {MAP_TLV_TRANSMIT_POWER_LIMIT,
				TLIST_ZEROORMORE_TLVS},
			[2 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_CHANNEL_PREFERENCE,
				TLIST_ZEROORMORE_TLVS},
			[1] = {MAP_TLV_TRANSMIT_POWER_LIMIT,
				TLIST_ZEROORMORE_TLVS},
			[2 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_CHANNEL_SELECTION_RESPONSE,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_CHANNEL_SELECTION_RESPONSE,
				TLIST_ONEORMORE_TLVS},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_CHANNEL_SELECTION_RESPONSE,
				TLIST_ONEORMORE_TLVS},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_OPERATING_CHANNEL_REPORT,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_OPERATING_CHANNEL_REPORT,
				TLIST_ONEORMORE_TLVS},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_OPERATING_CHANNEL_REPORT,
				TLIST_ONEORMORE_TLVS},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_CLIENT_CAPABILITY_QUERY,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_CLIENT_INFO, TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_CLIENT_INFO, TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_CLIENT_CAPABILITY_REPORT,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_CLIENT_INFO, TLIST_ONE_TLV},
			[1] = {MAP_TLV_CLIENT_CAPABILITY_REPORT,
				TLIST_ONE_TLV},
			[2] = {MAP_TLV_ERROR_CODE, TLIST_ZEROORONE_TLV},
			[3 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_CLIENT_INFO, TLIST_ONE_TLV},
			[1] = {MAP_TLV_CLIENT_CAPABILITY_REPORT,
				TLIST_ONE_TLV},
			[2] = {MAP_TLV_ERROR_CODE, TLIST_ZEROORONE_TLV},
			[3 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_AP_METRICS_QUERY,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_AP_METRIC_QUERY, TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_AP_METRIC_QUERY, TLIST_ONE_TLV},
			[1] = {MAP_TLV_AP_RADIO_IDENTIFIER,
				TLIST_ZEROORMORE_TLVS},
			[2 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_AP_METRICS_RESPONSE,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_AP_METRICS, TLIST_ONEORMORE_TLVS},
			[1] = {MAP_TLV_ASSOCIATED_STA_TRAFFIC_STATS,
				TLIST_ZEROORMORE_TLVS},
			[2] = {MAP_TLV_ASSOCIATED_STA_LINK_METRICS,
				TLIST_ZEROORMORE_TLVS},
			[3 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_AP_METRICS, TLIST_ONEORMORE_TLVS},
			[1] = {MAP_TLV_ASSOCIATED_STA_TRAFFIC_STATS,
				TLIST_ZEROORMORE_TLVS},
			[2] = {MAP_TLV_ASSOCIATED_STA_LINK_METRICS,
				TLIST_ZEROORMORE_TLVS},
			[3] = {MAP_TLV_AP_EXTENDED_METRICS,
				TLIST_ONEORMORE_TLVS},
			[4] = {MAP_TLV_RADIO_METRICS,
				TLIST_ZEROORMORE_TLVS},
			[5] = {MAP_TLV_ASSOCIATED_STA_EXT_LINK_METRICS,
				TLIST_ZEROORMORE_TLVS},
			[6 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_ASSOC_STA_LINK_METRICS_QUERY,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_STA_MAC_ADDRESS, TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_STA_MAC_ADDRESS, TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_ASSOC_STA_LINK_METRICS_RESPONSE,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_ASSOCIATED_STA_LINK_METRICS,
				TLIST_ONEORMORE_TLVS},
			[1] = {MAP_TLV_ERROR_CODE, TLIST_ZEROORONE_TLV},
			[2 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_ASSOCIATED_STA_LINK_METRICS,
				TLIST_ONEORMORE_TLVS},
			[1] = {MAP_TLV_ERROR_CODE, TLIST_ZEROORONE_TLV},
			[2] = {MAP_TLV_ASSOCIATED_STA_EXT_LINK_METRICS,
				TLIST_ONEORMORE_TLVS},
			[3 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_UNASSOC_STA_LINK_METRIC_QUERY,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_UNASSOCIATED_STA_LINK_METRICS_QUERY,
				TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_UNASSOCIATED_STA_LINK_METRICS_QUERY,
				TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_UNASSOC_STA_LINK_METRIC_RESPONSE,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_UNASSOCIATED_STA_LINK_METRICS_RESPONSE,
				TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_UNASSOCIATED_STA_LINK_METRICS_RESPONSE,
				TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_BEACON_METRICS_QUERY,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_BEACON_METRICS_QUERY, TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_BEACON_METRICS_QUERY, TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_BEACON_METRICS_RESPONSE,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_BEACON_METRICS_RESPONSE, TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_BEACON_METRICS_RESPONSE, TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_COMBINED_INFRA_METRICS,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_AP_METRICS, TLIST_ONEORMORE_TLVS},
			[1] = {TLV_TYPE_TRANSMITTER_LINK_METRIC,
				TLIST_ZEROORMORE_TLVS},
			[2] = {TLV_TYPE_RECEIVER_LINK_METRIC,
				TLIST_ZEROORMORE_TLVS},
			[3 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_AP_METRICS, TLIST_ONEORMORE_TLVS},
			[1] = {TLV_TYPE_TRANSMITTER_LINK_METRIC,
				TLIST_ZEROORMORE_TLVS},
			[2] = {TLV_TYPE_RECEIVER_LINK_METRIC,
				TLIST_ZEROORMORE_TLVS},
			[3 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_CLIENT_STEERING_REQUEST,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_STEERING_REQUEST, TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_STEERING_REQUEST, TLIST_ZEROORONE_TLV},
			[1] = {MAP_TLV_PROFILE2_STEERING_REQ,
				TLIST_ZEROORONE_TLV},
			[2 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_CLIENT_STEERING_BTM_REPORT,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_STEERING_BTM_REPORT, TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_STEERING_BTM_REPORT, TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_CLIENT_ASSOC_CONTROL_REQUEST,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_CLIENT_ASSOCIATION_CONTROL_REQUEST,
				TLIST_ONEORMORE_TLVS},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_CLIENT_ASSOCIATION_CONTROL_REQUEST,
				TLIST_ONEORMORE_TLVS},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_STEERING_COMPLETED,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_HIGHER_LAYER_DATA,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_HIGHER_LAYER_DATA, TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_HIGHER_LAYER_DATA, TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_BACKHAUL_STEER_REQUEST,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_BACKHAUL_STEERING_REQUEST,
				TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_BACKHAUL_STEERING_REQUEST,
				TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_BACKHAUL_STEER_RESPONSE,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0] = {MAP_TLV_BACKHAUL_STEERING_RESPONSE,
				TLIST_ONE_TLV},
			[1] = {MAP_TLV_ERROR_CODE, TLIST_ZEROORONE_TLV},
			[2 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_BACKHAUL_STEERING_RESPONSE,
				TLIST_ONE_TLV},
			[1] = {MAP_TLV_ERROR_CODE, TLIST_ZEROORONE_TLV},
			[2 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	// MAP-2 specific CMDUs
	{
		.cmdu_type = CMDU_CHANNEL_SCAN_REQUEST,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_CHANNEL_SCAN_REQ, TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_CHANNEL_SCAN_REPORT,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_TIMESTAMP, TLIST_ONE_TLV},
			[1] = {MAP_TLV_CHANNEL_SCAN_RES,
				TLIST_ONEORMORE_TLVS},
			[2 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_CAC_REQUEST,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_CAC_REQ, TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_CAC_TERMINATION,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_CAC_TERMINATION, TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_CLIENT_DISASSOCIATION_STATS,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_STA_MAC_ADDRESS, TLIST_ONE_TLV},
			[1] = {MAP_TLV_REASON_CODE, TLIST_ONE_TLV},
			[2] = {MAP_TLV_ASSOCIATED_STA_TRAFFIC_STATS,
				TLIST_ONE_TLV},
			[3 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_ERROR_RESPONSE,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_PROFILE2_ERR_CODE,
				TLIST_ONEORMORE_TLVS},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_ASSOCIATION_STATUS_NOTIFICATION,
		.relay = true,
		.transmission_type = TRANSMISSION_RELIABLE_MULTICAST,
		.profile1 = {
			[0 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_ASSOCIATION_STATUS_NOTIF,
				TLIST_ONE_TLV},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_TUNNELED,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_SOURCE_INFO, TLIST_ONE_TLV},
			[1] = {MAP_TLV_TUNNELED_MSG_TYPE, TLIST_ONE_TLV},
			[2] = {MAP_TLV_TUNNELED, TLIST_ONEORMORE_TLVS},
			[3 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_BACKHAUL_STA_CAPABILITY_QUERY,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_BACKHAUL_STA_CAPABILITY_REPORT,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_BACKHAUL_STA_RADIO_CAPABILITY,
				TLIST_ZEROORMORE_TLVS},
			[1 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
	{
		.cmdu_type = CMDU_FAILED_CONNECTION,
		.relay = false,
		.transmission_type = TRANSMISSION_UNICAST,
		.profile1 = {
			[0 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		},
		.profile2 = {
			[0] = {MAP_TLV_STA_MAC_ADDRESS, TLIST_ONE_TLV},
			[1] = {MAP_TLV_STATUS_CODE, TLIST_ONE_TLV},
			[2] = {MAP_TLV_REASON_CODE, TLIST_ZEROORONE_TLV},
			[3 ... MAX_TLV_PER_CMDU-1] = {0xFF, TLIST_EMPTY},
		}
	},
};

static const char *get_cmdu_transmission_string(uint8_t transmission_type)
{
	switch (transmission_type) {
	case TRANSMISSION_UNICAST:
		return "UNICAST";
	case TRANSMISSION_RELIABLE_MULTICAST:
		return "RELIABLE_MULTICAST";
	case TRANSMISSION_RELAYED_MULTICAST:
		return "RELAYED_MULTICAST";
	case TRANSMISSION_NEIGHBOR_MULTICAST:
		return "NEIGHBOR_MULTICAST";
	}
	return "Unknown";
}

static const char *get_tlv_count_string(uint8_t count)
{
	switch (count) {
	case TLIST_EMPTY:
		return "EMPTY";
	case TLIST_ZEROORMORE_TLVS:
		return "ZeroOrMore";
	case TLIST_ZEROORONE_TLV:
		return "ZeroOrOne";
	case TLIST_ONE_TLV:
		return "OnlyOne";
	case TLIST_ONEORMORE_TLVS:
		return "OneOrMore";
	}
	return "Invalid";
}

// Get the index of cmdu_type from db
// Input:
//       cmdu_type: two byte unsigned integer value as per specs
// Output:
//       return: -1, if cmdu_type is not available in db
//               <uint8_t>, index of cmdu_type in db
static int8_t get_map_cmdu_index(uint16_t cmdu_type)
{
	uint8_t num_cmdu = NUM_ELEMS(CMDU_VALIDATOR_TEMPLATE);

	for (uint8_t i = 0; i < num_cmdu; ++i) {
		if (cmdu_type == CMDU_VALIDATOR_TEMPLATE[i].cmdu_type)
			return i;
	}
	return -1;
}

static uint8_t get_tlv_tally_profile1(uint8_t cmdu_index, uint8_t tlv_type)
{
	uint8_t i;

	for (i = 0; i < MAX_TLV_PER_CMDU; ++i) {
		if (CMDU_VALIDATOR_TEMPLATE[cmdu_index].profile1[i].tlv_type ==
		    INVALID_TLV)
			break;

		if (tlv_type == CMDU_VALIDATOR_TEMPLATE[
		    cmdu_index].profile1[i].tlv_type)
			return (CMDU_VALIDATOR_TEMPLATE[
				cmdu_index].profile1[i].count);
	}

	return (TLIST_INVALID);
}

static uint8_t get_tlv_tally_profile2(uint8_t cmdu_index, uint8_t tlv_type)
{
	uint8_t i;

	for (i = 0; i < MAX_TLV_PER_CMDU; ++i) {
		if (CMDU_VALIDATOR_TEMPLATE[cmdu_index].profile2[i].tlv_type ==
		    INVALID_TLV)
			break;

		if (tlv_type == CMDU_VALIDATOR_TEMPLATE[
		    cmdu_index].profile2[i].tlv_type)
			return (CMDU_VALIDATOR_TEMPLATE[
				cmdu_index].profile2[i].count);
	}

	return TLIST_INVALID;
}

// Public functions
void map_dump_cmdu_validator(void)
{
	uint8_t i, j;
	uint8_t num_cmdu;
	const char *temp, *temp2;

	num_cmdu = NUM_ELEMS(CMDU_VALIDATOR_TEMPLATE);

	PRINT_DBG("########## Dump MAP cmdu validator ###########");
	for (i = 0; i < num_cmdu; ++i) {
		temp = map_stringify_cmdu_type(
					CMDU_VALIDATOR_TEMPLATE[i].cmdu_type);
		PRINT_DBG("\n## Type:(%s)", temp);

		PRINT_DBG("\tRelay:|%d|", CMDU_VALIDATOR_TEMPLATE[i].relay);

		temp = get_cmdu_transmission_string(
				CMDU_VALIDATOR_TEMPLATE[i].transmission_type);
		PRINT_DBG("\tTransmission:(%s)", temp);

		PRINT_DBG("\t# Profile1:");
		for (j = 0; j < MAX_TLV_PER_CMDU; ++j) {
			if (CMDU_VALIDATOR_TEMPLATE[i].profile1[j].tlv_type ==
			    0xFF)
				break;

			temp = map_stringify_tlv_type(
			      CMDU_VALIDATOR_TEMPLATE[i].profile1[j].tlv_type);

			temp2 = get_tlv_count_string(
				CMDU_VALIDATOR_TEMPLATE[i].profile1[j].count);
			PRINT_DBG("\t\tTLV:(%s) -> (%s)", temp, temp2);
		}
		PRINT_DBG("\t# Profile2:");
		for (j = 0; j < MAX_TLV_PER_CMDU; ++j) {
			if (CMDU_VALIDATOR_TEMPLATE[i].profile2[j].tlv_type ==
			    0xFF)
				break;

			temp = map_stringify_tlv_type(
			      CMDU_VALIDATOR_TEMPLATE[i].profile2[j].tlv_type);

			temp2 = get_tlv_count_string(
				CMDU_VALIDATOR_TEMPLATE[i].profile2[j].count);
			PRINT_DBG("\t\tTLV:(%s) -> (%s)", temp, temp2);
		}
	}
	PRINT_DBG("########## Dump MAP cmdu END ###########");
}

bool map_get_relay_bit(uint16_t cmdu_type)
{
	int8_t cmdu_index = get_map_cmdu_index(cmdu_type);

	if (cmdu_index == -1)
		return false;

	return CMDU_VALIDATOR_TEMPLATE[cmdu_index].relay;
}

bool map_validate_type(uint16_t cmdu_type)
{
	int8_t cmdu_index = get_map_cmdu_index(cmdu_type);

	if (cmdu_index == -1)
		return false;

	return true;
}

uint8_t map_get_transmission_type(uint16_t cmdu_type)
{
	int8_t cmdu_index = get_map_cmdu_index(cmdu_type);

	if (cmdu_index == -1)
		return TRANSMISSION_UNICAST;

	return CMDU_VALIDATOR_TEMPLATE[cmdu_index].transmission_type;
}

char *map_stringify_cmdu_type(uint16_t cmdu_type)
{
	switch (cmdu_type) {
		/* 1905 related cmdu extended to map */
	case CMDU_TYPE_TOPOLOGY_NOTIFICATION:
		return "CMDU_TYPE_TOPOLOGY_NOTIFICATION";
	case CMDU_TYPE_TOPOLOGY_QUERY:
		return "CMDU_TYPE_TOPOLOGY_QUERY";
	case CMDU_TYPE_TOPOLOGY_RESPONSE:
		return "CMDU_TYPE_TOPOLOGY_RESPONSE";
	case CMDU_TYPE_VENDOR_SPECIFIC:
		return "CMDU_TYPE_VENDOR_SPECIFIC";
	case CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH:
		return "CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH";
	case CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE:
		return "CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE";
	case CMDU_TYPE_AP_AUTOCONFIGURATION_WSC:
		return "CMDU_TYPE_AP_AUTOCONFIGURATION_WSC";
	case CMDU_TYPE_AP_AUTOCONFIGURATION_RENEW:
		return "CMDU_TYPE_AP_AUTOCONFIGURATION_RENEW";
		/* map profile-1 related cmdu */
	case CMDU_1905_ACK:
		return "CMDU_1905_ACK";
	case CMDU_AP_CAPABILITY_QUERY:
		return "CMDU_AP_CAPABILITY_QUERY";
	case CMDU_AP_CAPABILITY_REPORT:
		return "CMDU_AP_CAPABILITY_REPORT";
	case CMDU_POLICY_CONFIG_REQ:
		return "CMDU_POLICY_CONFIG_REQ";
	case CMDU_CHANNEL_PREFERENCE_QUERY:
		return "CMDU_CHANNEL_PREFERENCE_QUERY";
	case CMDU_CHANNEL_PREFERENCE_REPORT:
		return "CMDU_CHANNEL_PREFERENCE_REPORT";
	case CMDU_CHANNEL_SELECTION_REQ:
		return "CMDU_CHANNEL_SELECTION_REQ";
	case CMDU_CHANNEL_SELECTION_RESPONSE:
		return "CMDU_CHANNEL_SELECTION_RESPONSE";
	case CMDU_OPERATING_CHANNEL_REPORT:
		return "CMDU_OPERATING_CHANNEL_REPORT";
	case CMDU_CLIENT_CAPABILITY_QUERY:
		return "CMDU_CLIENT_CAPABILITY_QUERY";
	case CMDU_CLIENT_CAPABILITY_REPORT:
		return "CMDU_CLIENT_CAPABILITY_REPORT";
	case CMDU_AP_METRICS_QUERY:
		return "CMDU_AP_METRICS_QUERY";
	case CMDU_AP_METRICS_RESPONSE:
		return "CMDU_AP_METRICS_RESPONSE";
	case CMDU_ASSOC_STA_LINK_METRICS_QUERY:
		return "CMDU_ASSOC_STA_LINK_METRICS_QUERY";
	case CMDU_ASSOC_STA_LINK_METRICS_RESPONSE:
		return "CMDU_ASSOC_STA_LINK_METRICS_RESPONSE";
	case CMDU_UNASSOC_STA_LINK_METRIC_QUERY:
		return "CMDU_UNASSOC_STA_LINK_METRIC_QUERY";
	case CMDU_UNASSOC_STA_LINK_METRIC_RESPONSE:
		return "CMDU_UNASSOC_STA_LINK_METRIC_RESPONSE";
	case CMDU_BEACON_METRICS_QUERY:
		return "CMDU_BEACON_METRICS_QUERY";
	case CMDU_BEACON_METRICS_RESPONSE:
		return "CMDU_BEACON_METRICS_RESPONSE";
	case CMDU_COMBINED_INFRA_METRICS:
		return "CMDU_COMBINED_INFRA_METRICS";
	case CMDU_CLIENT_STEERING_REQUEST:
		return "CMDU_CLIENT_STEERING_REQUEST";
	case CMDU_CLIENT_STEERING_BTM_REPORT:
		return "CMDU_CLIENT_STEERING_BTM_REPORT";
	case CMDU_CLIENT_ASSOC_CONTROL_REQUEST:
		return "CMDU_CLIENT_ASSOC_CONTROL_REQUEST";
	case CMDU_STEERING_COMPLETED:
		return "CMDU_STEERING_COMPLETED";
	case CMDU_HIGHER_LAYER_DATA:
		return "CMDU_HIGHER_LAYER_DATA";
	case CMDU_BACKHAUL_STEER_REQUEST:
		return "CMDU_BACKHAUL_STEER_REQUEST";
	case CMDU_BACKHAUL_STEER_RESPONSE:
		return "CMDU_BACKHAUL_STEER_RESPONSE";
		/* map profile-2 related CMDUs */
	case CMDU_CHANNEL_SCAN_REQUEST:
		return "CMDU_CHANNEL_SCAN_REQUEST";
	case CMDU_CHANNEL_SCAN_REPORT:
		return "CMDU_CHANNEL_SCAN_REPORT";
	case CMDU_CAC_REQUEST:
		return "CMDU_CAC_REQUEST";
	case CMDU_CAC_TERMINATION:
		return "CMDU_CAC_TERMINATION";
	case CMDU_CLIENT_DISASSOCIATION_STATS:
		return "CMDU_CLIENT_DISASSOCIATION_STATS";
	case CMDU_ERROR_RESPONSE:
		return "CMDU_ERROR_RESPONSE";
	case CMDU_ASSOCIATION_STATUS_NOTIFICATION:
		return "CMDU_ASSOCIATION_STATUS_NOTIFICATION";
	case CMDU_TUNNELED:
		return "CMDU_TUNNELED";
	case CMDU_BACKHAUL_STA_CAPABILITY_QUERY:
		return "CMDU_BACKHAUL_STA_CAPABILITY_QUERY";
	case CMDU_BACKHAUL_STA_CAPABILITY_REPORT:
		return "CMDU_BACKHAUL_STA_CAPABILITY_REPORT";
	case CMDU_FAILED_CONNECTION:
		return "CMDU_FAILED_CONNECTION";
	}
	return "";
}

bool map_cmdu_string_to_type(char *cmdu_string, uint16_t *cmdu_type)
{
	if (cmdu_type == NULL)
		return false;

	// initialize with dummy value
	*cmdu_type = 0xFFFF;

	/* 1905 related cmdu extended to map */
	if (!strcmp(cmdu_string, "CMDU_TYPE_TOPOLOGY_NOTIFICATION"))
		*cmdu_type = CMDU_TYPE_TOPOLOGY_NOTIFICATION;
	else if (!strcmp(cmdu_string, "CMDU_TYPE_TOPOLOGY_QUERY"))
		*cmdu_type = CMDU_TYPE_TOPOLOGY_QUERY;
	else if (!strcmp(cmdu_string, "CMDU_TYPE_TOPOLOGY_RESPONSE"))
		*cmdu_type = CMDU_TYPE_TOPOLOGY_RESPONSE;
	else if (!strcmp(cmdu_string, "CMDU_TYPE_VENDOR_SPECIFIC"))
		*cmdu_type = CMDU_TYPE_VENDOR_SPECIFIC;
	else if (!strcmp(cmdu_string, "CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH"))
		*cmdu_type = CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH;
	else if (!strcmp(cmdu_string,
			 "CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE"))
		*cmdu_type = CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE;
	else if (!strcmp(cmdu_string, "CMDU_TYPE_AP_AUTOCONFIGURATION_WSC"))
		*cmdu_type = CMDU_TYPE_AP_AUTOCONFIGURATION_WSC;
	else if (!strcmp(cmdu_string, "CMDU_TYPE_AP_AUTOCONFIGURATION_RENEW"))
		*cmdu_type = CMDU_TYPE_AP_AUTOCONFIGURATION_RENEW;
	/* map profile-1 related cmdu */
	else if (!strcmp(cmdu_string, "CMDU_1905_ACK"))
		*cmdu_type = CMDU_1905_ACK;
	else if (!strcmp(cmdu_string, "CMDU_AP_CAPABILITY_QUERY"))
		*cmdu_type = CMDU_AP_CAPABILITY_QUERY;
	else if (!strcmp(cmdu_string, "CMDU_AP_CAPABILITY_REPORT"))
		*cmdu_type = CMDU_AP_CAPABILITY_REPORT;
	else if (!strcmp(cmdu_string, "CMDU_POLICY_CONFIG_REQ"))
		*cmdu_type = CMDU_POLICY_CONFIG_REQ;
	else if (!strcmp(cmdu_string, "CMDU_CHANNEL_PREFERENCE_QUERY"))
		*cmdu_type = CMDU_CHANNEL_PREFERENCE_QUERY;
	else if (!strcmp(cmdu_string, "CMDU_CHANNEL_PREFERENCE_REPORT"))
		*cmdu_type = CMDU_CHANNEL_PREFERENCE_REPORT;
	else if (!strcmp(cmdu_string, "CMDU_CHANNEL_SELECTION_REQ"))
		*cmdu_type = CMDU_CHANNEL_SELECTION_REQ;
	else if (!strcmp(cmdu_string, "CMDU_CHANNEL_SELECTION_RESPONSE"))
		*cmdu_type = CMDU_CHANNEL_SELECTION_RESPONSE;
	else if (!strcmp(cmdu_string, "CMDU_OPERATING_CHANNEL_REPORT"))
		*cmdu_type = CMDU_OPERATING_CHANNEL_REPORT;
	else if (!strcmp(cmdu_string, "CMDU_CLIENT_CAPABILITY_QUERY"))
		*cmdu_type = CMDU_CLIENT_CAPABILITY_QUERY;
	else if (!strcmp(cmdu_string, "CMDU_CLIENT_CAPABILITY_REPORT"))
		*cmdu_type = CMDU_CLIENT_CAPABILITY_REPORT;
	else if (!strcmp(cmdu_string, "CMDU_AP_METRICS_QUERY"))
		*cmdu_type = CMDU_AP_METRICS_QUERY;
	else if (!strcmp(cmdu_string, "CMDU_AP_METRICS_RESPONSE"))
		*cmdu_type = CMDU_AP_METRICS_RESPONSE;
	else if (!strcmp(cmdu_string, "CMDU_ASSOC_STA_LINK_METRICS_QUERY"))
		*cmdu_type = CMDU_ASSOC_STA_LINK_METRICS_QUERY;
	else if (!strcmp(cmdu_string, "CMDU_ASSOC_STA_LINK_METRICS_RESPONSE"))
		*cmdu_type = CMDU_ASSOC_STA_LINK_METRICS_RESPONSE;
	else if (!strcmp(cmdu_string, "CMDU_UNASSOC_STA_LINK_METRIC_QUERY"))
		*cmdu_type = CMDU_UNASSOC_STA_LINK_METRIC_QUERY;
	else if (!strcmp(cmdu_string, "CMDU_UNASSOC_STA_LINK_METRIC_RESPONSE"))
		*cmdu_type = CMDU_UNASSOC_STA_LINK_METRIC_RESPONSE;
	else if (!strcmp(cmdu_string, "CMDU_BEACON_METRICS_QUERY"))
		*cmdu_type = CMDU_BEACON_METRICS_QUERY;
	else if (!strcmp(cmdu_string, "CMDU_BEACON_METRICS_RESPONSE"))
		*cmdu_type = CMDU_BEACON_METRICS_RESPONSE;
	else if (!strcmp(cmdu_string, "CMDU_COMBINED_INFRA_METRICS"))
		*cmdu_type = CMDU_COMBINED_INFRA_METRICS;
	else if (!strcmp(cmdu_string, "CMDU_CLIENT_STEERING_REQUEST"))
		*cmdu_type = CMDU_CLIENT_STEERING_REQUEST;
	else if (!strcmp(cmdu_string, "CMDU_CLIENT_STEERING_BTM_REPORT"))
		*cmdu_type = CMDU_CLIENT_STEERING_BTM_REPORT;
	else if (!strcmp(cmdu_string, "CMDU_CLIENT_ASSOC_CONTROL_REQUEST"))
		*cmdu_type = CMDU_CLIENT_ASSOC_CONTROL_REQUEST;
	else if (!strcmp(cmdu_string, "CMDU_STEERING_COMPLETED"))
		*cmdu_type = CMDU_STEERING_COMPLETED;
	else if (!strcmp(cmdu_string, "CMDU_HIGHER_LAYER_DATA"))
		*cmdu_type = CMDU_HIGHER_LAYER_DATA;
	else if (!strcmp(cmdu_string, "CMDU_BACKHAUL_STEER_REQUEST"))
		*cmdu_type = CMDU_BACKHAUL_STEER_REQUEST;
	else if (!strcmp(cmdu_string, "CMDU_BACKHAUL_STEER_RESPONSE"))
		*cmdu_type = CMDU_BACKHAUL_STEER_RESPONSE;
	/* map profile-2 related CMDUs */
	else if (!strcmp(cmdu_string, "CMDU_CHANNEL_SCAN_REQUEST"))
		*cmdu_type = CMDU_CHANNEL_SCAN_REQUEST;
	else if (!strcmp(cmdu_string, "CMDU_CHANNEL_SCAN_REPORT"))
		*cmdu_type = CMDU_CHANNEL_SCAN_REPORT;
	else if (!strcmp(cmdu_string, "CMDU_CAC_REQUEST"))
		*cmdu_type = CMDU_CAC_REQUEST;
	else if (!strcmp(cmdu_string, "CMDU_CAC_TERMINATION"))
		*cmdu_type = CMDU_CAC_TERMINATION;
	else if (!strcmp(cmdu_string, "CMDU_CLIENT_DISASSOCIATION_STATS"))
		*cmdu_type = CMDU_CLIENT_DISASSOCIATION_STATS;
	else if (!strcmp(cmdu_string, "CMDU_ERROR_RESPONSE"))
		*cmdu_type = CMDU_ERROR_RESPONSE;
	else if (!strcmp(cmdu_string, "CMDU_ASSOCIATION_STATUS_NOTIFICATION"))
		*cmdu_type = CMDU_ASSOCIATION_STATUS_NOTIFICATION;
	else if (!strcmp(cmdu_string, "CMDU_TUNNELED"))
		*cmdu_type = CMDU_TUNNELED;
	else if (!strcmp(cmdu_string, "CMDU_BACKHAUL_STA_CAPABILITY_QUERY"))
		*cmdu_type = CMDU_BACKHAUL_STA_CAPABILITY_QUERY;
	else if (!strcmp(cmdu_string, "CMDU_BACKHAUL_STA_CAPABILITY_REPORT"))
		*cmdu_type = CMDU_BACKHAUL_STA_CAPABILITY_REPORT;
	else if (!strcmp(cmdu_string, "CMDU_FAILED_CONNECTION"))
		*cmdu_type = CMDU_FAILED_CONNECTION;
	else
		return false;

	return true;
}

// Input ==>
//     p -> cmdu stream structure
//
//     profile -> 1(MULTIAP_PROFILE_1), validate against profile 1 type TLVs
//             -> 2(MULTIAP_PROFILE_2), validate against profile 2 type of TLVs
//
//     controller -> 0(false), validation for map-agent
//                -> 1(true), validation for map-controller
//
// Output ==>
//     return  -> true, if cmdu has all the required tlvs as per
//                profile and controller
//             -> false, if cmdu has missing or extra tlvs as per
//                profile and controller
//
bool map_validate_cmdu_stream(const struct cmdu_stream *p,
				 uint8_t profile, bool controller)
{
	uint8_t i = 0;
	uint8_t counter[MAP_TLV_LAST] = {[0 ... MAP_TLV_LAST-1] = 0};
	int8_t cmdu_index = -1;
	bool ret = false;

	// Sanitize the output parameters
	if (profile > MULTIAP_PROFILE_2 || profile < MULTIAP_PROFILE_1) {
		PRINT_ERR("profile|%d| not supported", profile);
		return false;
	}

	// validate the input parameters
	if (p == NULL) {
		PRINT_ERR("Invalid CMDU structure");
		return false;
	}

	// Count the tlvs present in the cmdu
	i = 0;
	while (i < p->nbr_tlv_streams) {
		counter[*(p->tlv_stream_list[i])]++;
		i++;
	}

	cmdu_index = get_map_cmdu_index(p->message_type);
	if (cmdu_index == -1) {
		//PRINT_DBG("not a valid map cmdu|0x%x|", p->message_type);
		return false;
	}

	// Validation for mode {agent, controller}, this is only applicable
	// for WSC cmdu
	if (p->message_type == CMDU_TYPE_AP_AUTOCONFIGURATION_WSC) {
		// validation of controller
		if (controller) {
			if (profile == MULTIAP_PROFILE_1) {
				return (counter[MAP_TLV_AP_RADIO_IDENTIFIER] == 1 &&
					counter[TLV_TYPE_WSC] >= 1);
			} else if (profile == MULTIAP_PROFILE_2) {
				return (counter[MAP_TLV_AP_RADIO_IDENTIFIER] == 1 &&
					 counter[TLV_TYPE_WSC] >= 1 &&
					 (counter[MAP_TLV_DEFAULT_8021Q_SETTINGS] == 0 ||
					  counter[MAP_TLV_DEFAULT_8021Q_SETTINGS] == 1) &&
					 (counter[MAP_TLV_TRAFFIC_SEPARATION_POLICY] == 0 ||
					  counter[MAP_TLV_TRAFFIC_SEPARATION_POLICY] == 1));
				}
		} else {    // validation for map-agent
			if (profile == MULTIAP_PROFILE_1) {
				return (counter[MAP_TLV_AP_RADIO_BASIC_CAPABILITIES] == 1 &&
					counter[TLV_TYPE_WSC] == 1);
			} else if (profile == MULTIAP_PROFILE_2) {
				return (counter[MAP_TLV_AP_RADIO_BASIC_CAPABILITIES] == 1 &&
					counter[TLV_TYPE_WSC] == 1 &&
					counter[MAP_TLV_PROFILE2_AP_CAPABILITY] == 1 &&
					counter[MAP_TLV_AP_RADIO_ADV_CAPABILITY] == 1);
			}
		}
		return false;
	}

	// Validation for profile1
	if (profile == MULTIAP_PROFILE_1) {
		ret = true;
		for (i = 0; i < MAP_TLV_LAST; i++) {
			uint8_t tally = get_tlv_tally_profile1(cmdu_index, i);

			if (tally == TLIST_INVALID) {
				if (counter[i]) {
					PRINT_ERR("extra TLV|0x%02x| found |%d| times",
						  i, counter[i]);
					ret = false;
				}
				break;
			}

			switch (tally) {
			case TLIST_EMPTY:
				{
					if (counter[i] > 0) {
						PRINT_ERR("TLV(%s) Zero, but |%d| times",
							  map_stringify_tlv_type(i), counter[i]);
						ret = false;
					}
					break;
				}
			case TLIST_ZEROORMORE_TLVS:
				{
					break;
				}
			case TLIST_ZEROORONE_TLV:
				{
					if (counter[i] != 0 && counter[i] != 1) {
						PRINT_ERR("TLV(%s) ZeroOrOne, but %d times",
							  map_stringify_tlv_type(i), counter[i]);
						ret = false;
					}
					break;
				}
			case TLIST_ONE_TLV:
				{
					if (counter[i] != 1) {
						PRINT_ERR("TLV(%s) one, but %d times",
							  map_stringify_tlv_type(i), counter[i]);
						ret = false;
					}
					break;
				}
			case TLIST_ONEORMORE_TLVS:
				{
					if (counter[i] < 1) {
						PRINT_ERR("TLV(%s) OneorMore, but %d times",
							  map_stringify_tlv_type(i), counter[i]);
						ret = false;
					}
					break;
				}
			default:
				{
					// Not reachable case
					ret = false;
				}
			}
			if (ret == false)
				break;

		}
		return ret;
	} else if (profile == MULTIAP_PROFILE_2) {
		// now validate if cmdu has all the tlvs required for profile2
		ret = true;
		for (i = 0; i < MAP_TLV_LAST; i++) {
			uint8_t tally = get_tlv_tally_profile2(cmdu_index, i);

			if (tally == TLIST_INVALID) {
				if (counter[i]) {
					PRINT_ERR("extra TLV|0x%02x| found |%d| times",
						  i, counter[i]);
					ret = false;
				}
				break;
			}

			switch (tally) {
			case TLIST_EMPTY:
			case TLIST_INVALID:
				{
					if (counter[i] > 0) {
						PRINT_ERR("P2 TLV(%s) 0, but |%d| times",
							  map_stringify_tlv_type(i), counter[i]);
						ret = false;
					}
					break;
				}
			case TLIST_ZEROORMORE_TLVS:
				{
					break;
				}
			case TLIST_ZEROORONE_TLV:
				{
					if (counter[i] != 0 && counter[i] != 1) {
						PRINT_ERR("P2 TLV(%s) ZeroOrOne, but |%d| times",
							  map_stringify_tlv_type(i), counter[i]);
						ret = false;
					}
					break;
				}
			case TLIST_ONE_TLV:
				{
					if (counter[i] != 1) {
						PRINT_ERR("P2 TLV(%s) One, but |%d| times",
							  map_stringify_tlv_type(i), counter[i]);
						ret = false;
					}
					break;
				}
			case TLIST_ONEORMORE_TLVS:
				{
					if (counter[i] < 1) {
						PRINT_ERR("P2 TLV(%s) OneorMore, but |%d| times",
							  map_stringify_tlv_type(i), counter[i]);
						ret = false;
					}
					break;
				}
			default:
				{
					// not reachable case
					ret = false;
				}
			}
			if (ret == false)
				break;
		}
		return ret;
	}

	// logically un-reachable code
	return false;
}

// Input ==>
//     p -> cmdu structure
//
// Output ==>
//     return  -> true, if cmdu has all the required tlvs
//             -> false, if cmdu has missing or extra tlvs
//
//     profile -> NULL, if this info is not required by caller
//             -> 1, if cmdu has profile 1 type tlvs
//             -> 2, if cmdu has profile 2 type of tlvs
//             -> 3, profile not determined
//
//     mode    -> NULL, if this info is not required by caller
//             -> 0, if cmdu contains tlvs valid for map-agent
//             -> 1, if cmdu contains tlvs valid for map-controller
//             -> 2, mode not determined
bool map_get_profile_mode_cmdu_stream(const struct cmdu_stream *p,
					 uint8_t *profile,
					 uint8_t *mode)
{
	uint8_t i = 0;
	uint8_t counter[MAP_TLV_LAST] = {[0 ... MAP_TLV_LAST-1] = 0};
	int8_t cmdu_index = -1;
	bool ret = false, profile1_ret, profile2_ret;

	// Sanitize the output parameters
	if (profile != NULL)
		*profile = 0xff;

	if (mode != NULL)
		*mode = 0xff;

	// validate the input parameters
	if (p == NULL) {
		PRINT_ERR("Invalid CMDU structure");
		return false;
	}

	// Count the tlvs present in the cmdu
	for (i = 0; i < p->nbr_tlv_streams; i++)
		counter[*(p->tlv_stream_list[i])]++;


	cmdu_index = get_map_cmdu_index(p->message_type);
	if (cmdu_index == -1)
		return false;

	// Validation for mode {agent, controller}, this is only applicable for WSC cmdu
	if (p->message_type == CMDU_TYPE_AP_AUTOCONFIGURATION_WSC) {
		if (counter[MAP_TLV_AP_RADIO_BASIC_CAPABILITIES] == 1 &&
		    counter[TLV_TYPE_WSC] == 1) {
			ret = true;
			if (counter[MAP_TLV_PROFILE2_AP_CAPABILITY] == 1 &&
			    counter[MAP_TLV_AP_RADIO_ADV_CAPABILITY] == 1) {
				if (profile != NULL)
					*profile = MULTIAP_PROFILE_2;
			} else {
				if (profile != NULL)
					*profile = MULTIAP_PROFILE_1;
			}
			if (mode != NULL)
				*mode = 0;
		}

		if (counter[MAP_TLV_AP_RADIO_IDENTIFIER] == 1 &&
		    counter[TLV_TYPE_WSC] >= 1) {
			ret = true;
			if ((counter[MAP_TLV_DEFAULT_8021Q_SETTINGS] == 0 ||
			     counter[MAP_TLV_DEFAULT_8021Q_SETTINGS] == 1) &&
			    (counter[MAP_TLV_TRAFFIC_SEPARATION_POLICY] == 0 ||
			     counter[MAP_TLV_TRAFFIC_SEPARATION_POLICY] == 1)) {
				if (profile != NULL)
					*profile = MULTIAP_PROFILE_2;
			} else {
				if (profile != NULL)
					*profile = MULTIAP_PROFILE_1;
			}
			if (counter[MAP_TLV_DEFAULT_8021Q_SETTINGS] == 0 &&
			    counter[MAP_TLV_TRAFFIC_SEPARATION_POLICY] == 0) {
				if (profile != NULL)
					*profile = MULTIAP_PROFILE_1;
			}
			if (mode != NULL)
				*mode = 1;
		}
		if (ret)
			return ret;
	}

	// first validate if cmdu has all the tlvs required for profile1
	profile1_ret = true;
	for (i = 0; i < MAP_TLV_LAST; i++) {
		switch (get_tlv_tally_profile1(cmdu_index, i)) {
		case TLIST_EMPTY:
		case TLIST_INVALID:
			{
				if (counter[i] > 0) {
					PRINT_ERR("TLV(%s) 0, but |%d| times",
						map_stringify_tlv_type(i), counter[i]);
					profile1_ret = false;
				}
				break;
			}
		case TLIST_ZEROORMORE_TLVS:
			{
				break;
			}
		case TLIST_ZEROORONE_TLV:
			{
				if (counter[i] != 0 && counter[i] != 1) {
					PRINT_ERR("TLV(%s) 0or1, but %d times",
						map_stringify_tlv_type(i), counter[i]);
					profile1_ret = false;
				}
				break;
			}
		case TLIST_ONE_TLV:
			{
				if (counter[i] != 1) {
					PRINT_ERR("TLV(%s) one, but %d times",
						map_stringify_tlv_type(i), counter[i]);
					profile1_ret = false;
				}
				break;
			}
		case TLIST_ONEORMORE_TLVS:
			{
				if (counter[i] < 1) {
					PRINT_ERR("TLV(%s) 1orMore, but %d times",
						map_stringify_tlv_type(i), counter[i]);
					profile1_ret = false;
				}
				break;
			}
		default:
			{
				// Not reachable case
				profile1_ret = false;
			}
		}
	}

	// now validate if cmdu has all the tlvs required for profile2
	profile2_ret = true;
	for (i = 0; i < MAP_TLV_LAST; i++) {
		switch (get_tlv_tally_profile2(cmdu_index, i)) {
		case TLIST_EMPTY:
		case TLIST_INVALID:
			{
				if (counter[i] > 0) {
					PRINT_ERR("P2 TLV(%s) 0, but |%d| times",
						map_stringify_tlv_type(i), counter[i]);
					profile2_ret = false;
				}
				break;
			}
		case TLIST_ZEROORMORE_TLVS:
			{
				break;
			}
		case TLIST_ZEROORONE_TLV:
			{
				if (counter[i] != 0 && counter[i] != 1) {
					PRINT_ERR("P2 TLV(%s) 0or1, but |%d| times",
						map_stringify_tlv_type(i), counter[i]);
					profile2_ret = false;
				}
				break;
			}
		case TLIST_ONE_TLV:
			{
				if (counter[i] != 1) {
					PRINT_ERR("P2 TLV(%s) one, but |%d| times",
						map_stringify_tlv_type(i), counter[i]);
					profile2_ret = false;
				}
				break;
			}
		case TLIST_ONEORMORE_TLVS:
			{
				if (counter[i] < 1) {
					PRINT_ERR("P2 TLV(%s) 1orMore, but |%d| times",
						map_stringify_tlv_type(i), counter[i]);
					profile2_ret = false;
				}
				break;
			}
		default:
			{
				// not reachable case
				profile2_ret = false;
			}
		}
	}

	// ret val should be true if any profile matched
	ret = profile1_ret || profile2_ret;

	// set the profile
	if (profile != NULL && ret == true) {
		if (profile2_ret)
			*profile = MULTIAP_PROFILE_2;
		else
			*profile = MULTIAP_PROFILE_1;

		if (profile1_ret && profile2_ret)
			*profile = MULTIAP_PROFILE_1;
	}

	return ret;
}

void map_free_cmdu(struct cmdu_cstruct *c)
{
	uint8_t i;
	if (c == NULL)
		return;

	for (i = 0; i < c->num_tlvs; i++)
		map_free_tlv_cstruct(c->tlvs[i]);

	if (c->num_tlvs)
		free(c->tlvs);

	free(c);
}

void map_free_tlv_stream_list(uint8_t **packet_streams, uint8_t nbr_streams)
{
	uint8_t i;

	if (packet_streams == NULL)
		return;

	for (i = 0; i < nbr_streams; ++i)
		free(packet_streams[i]);

	free(packet_streams);
}

// Input:
//      cmdu1 -> cmdu structure 1
//      cmdu2 -> cmdu structure 2
// Output:
//      ret -> true, if both the cmdu's are same
//          -> false, if both the cmdu's are different
bool map_compare_cmdu_cstruct(const struct cmdu_cstruct *memory_structure_1,
				 const struct cmdu_cstruct *memory_structure_2)
{
	uint8_t i;
	uint8_t ret = true;

	if (memory_structure_1 == NULL ||
	    memory_structure_2 == NULL)
		return false;

	if (memory_structure_1->num_tlvs !=
	    memory_structure_2->num_tlvs)
		return false;

	// Passed the same memory structure twice
	if (memory_structure_1 == memory_structure_2)
		return true;

	if ((memory_structure_1->message_version != memory_structure_2->message_version) ||
	    (memory_structure_1->message_type != memory_structure_2->message_type) ||
	    (memory_structure_1->message_id != memory_structure_2->message_id) ||
	    (memory_structure_1->relay_indicator != memory_structure_2->relay_indicator))
		return false;

	// TODO: fix this logic, as tlv cstruct is an unordered list, so one-on-one
	// comparison is not valid
	for (i = 0; i < memory_structure_1->num_tlvs; i++) {
		ret = map_compare_tlv_cstructs(memory_structure_1->tlvs[i],
						 memory_structure_2->tlvs[i]);
		if (ret == false)
			return false;
	}

	return true;
}

/**
 * @brief API to determine if the cmdu_type is a response message. This is
 * useful to know as the mid will be same for response message, whereas for
 * notifications and queries new mid will be used.
 *
 * @param cmdu_type cmdu type as described in map specification
 *
 * @return boolean true, if cmdu is a response message, false otherwise
 */
bool map_is_response_cmdu(uint16_t cmdu_type)
{

	switch (cmdu_type) {
	case CMDU_TYPE_TOPOLOGY_RESPONSE:
	case CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE:
	case CMDU_AP_CAPABILITY_REPORT:
	case CMDU_CHANNEL_PREFERENCE_REPORT:
	case CMDU_CHANNEL_SELECTION_RESPONSE:
	case CMDU_OPERATING_CHANNEL_REPORT:
	case CMDU_CLIENT_CAPABILITY_REPORT:
	case CMDU_AP_METRICS_RESPONSE:
	case CMDU_ASSOC_STA_LINK_METRICS_RESPONSE:
	case CMDU_UNASSOC_STA_LINK_METRIC_RESPONSE:
	case CMDU_BEACON_METRICS_RESPONSE:
	case CMDU_COMBINED_INFRA_METRICS:
	case CMDU_CLIENT_STEERING_BTM_REPORT:
	case CMDU_STEERING_COMPLETED:
	case CMDU_HIGHER_LAYER_DATA:
	case CMDU_BACKHAUL_STEER_RESPONSE:
	case CMDU_CHANNEL_SCAN_REPORT:
	case CMDU_ERROR_RESPONSE:
	case CMDU_BACKHAUL_STA_CAPABILITY_REPORT:
		return true;
	}

	return false;
}

uint8_t map_count_tlvs_safe(uint8_t *tlv_stream, size_t stream_len)
{
	uint8_t *p = tlv_stream;
	uint8_t count =0;
	uint16_t tlv_len = 0;
	size_t parsed_len = 0;
	uint8_t tlv_type = 0;

	if (tlv_stream == NULL ||
	    stream_len == 0) {
		PRINT_ERR("Stream is empty");
		return 0;
	}

	while (parsed_len < stream_len) {
		_E1B(&p, &tlv_type);
		_E2B(&p, &tlv_len);

		// Assuming we reached the end of cmdu's tlv stream
		if (tlv_type == TLV_TYPE_END_OF_MESSAGE)
			break;

		p = p + tlv_len;
		parsed_len += tlv_len + 3;
		count++;
	}

	if (parsed_len > stream_len) {
		PRINT_ERR("Last tlv(0x%x) length(%d) is more than stream len(%zd)",
			  tlv_type, tlv_len, stream_len);
		// removing the last tlv
		count = (count) ? (count - 1) : 0;
	}

	return count;
}

uint8_t map_count_tlvs_in_stream(uint8_t *tlv_stream)
{
	uint8_t *p = tlv_stream;
	uint8_t count =0;
	uint16_t tlv_len = 0;
	uint8_t tlv_type;

	if (tlv_stream == NULL) {
		PRINT_ERR("Stream is empty");
		return 0;
	}

	do {
		_E1B(&p, &tlv_type);
		_E2B(&p, &tlv_len);
		printf("tlv(0x%x), len(0x%x:%d)\n", tlv_type, tlv_len, tlv_len);

		// Assuming we reached the end of cmdu's tlv stream
		if (tlv_type == TLV_TYPE_END_OF_MESSAGE)
			break;

		p = p + tlv_len;
		count++;
	} while(tlv_type != TLV_TYPE_END_OF_MESSAGE);

	return count;
}


/**
 * @brief Function to build a cmdu c-structure
 *
 * @param type cmdu message_type as defined in map specification
 * @param mid cmdu message_id as defined in map specs
 * @param rxif received interface name
 * @param srcmac source mac-address of received cmdu
 * @param tlvs Binary streams of tlvs associated with cmdu
 *
 * @return NULL if decoding fails
 * @return cmdu_cstruct if decoding succeeds
 */
 struct cmdu_cstruct *map_build_cmdu(uint16_t type, uint16_t mid,
				     const char *rxif, uint8_t *srcmac,
				     uint8_t *tlvs)
{
	uint8_t i, tlv_count;
	struct cmdu_cstruct *cmdu;
	uint8_t *cur_stream, *tlv = NULL;

	if (map_validate_type(type) == false) {
		PRINT_ERR("Not a valid map cmdu(0x%04x)", type);
		return NULL;
	}

	cmdu = calloc(1, sizeof(struct cmdu_cstruct));
	if (!cmdu) {
		PRINT_ERR("Insufficient memory!");
		return NULL;
	}

	cmdu->message_version = CMDU_MESSAGE_VERSION_1905_1_2013;
	cmdu->message_type = type;
	cmdu->message_id = mid;
	cmdu->relay_indicator = map_get_relay_bit(type);

	if (rxif)
		strcpy(cmdu->intf_name, rxif);

	if (srcmac)
		memcpy(cmdu->origin, srcmac, 6);

	cmdu->num_tlvs = map_count_tlvs_in_stream(tlvs);
	if (cmdu->num_tlvs) {
		cmdu->tlvs = calloc(cmdu->num_tlvs , sizeof(uint8_t *));
		if (!cmdu->tlvs) {
			free(cmdu);
			PRINT_ERR("Insufficient memory!");
			return NULL;
		}
	}

	tlv_count = 0;
	cur_stream = tlvs;
	for (i = 0; i < cmdu->num_tlvs; i++) {
		uint16_t len = map_get_first_tlv_length(cur_stream);

		tlv = map_get_tlv_cstruct(cur_stream);
		if (tlv)
			cmdu->tlvs[tlv_count++] =  tlv;

		// Move to the next tlv
		cur_stream = cur_stream + len;
	}

	if (cmdu->num_tlvs != tlv_count) {
		PRINT_ERR("Decoding failed for some of the tlv streams! |%d:%d|", cmdu->num_tlvs, tlv_count);
		cmdu->num_tlvs = tlv_count;
	}

	return cmdu;
}
