/*
 * map_module.c
 * interface to multi-ap module(s)
 *
 * Copyright (C) 2020 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>
#include <easy/easy.h>

#include <ieee1905/plugin.h>

#include "map2.h"
#include "maputils.h"
#include "map_module.h"
#include "map_common.h"

#define MULTICAST_ADDR_STR "01:80:c2:00:00:13"

enum {
	MAP_MODULE_NOTPRESENT,
	MAP_MODULE_REGISTERED,
	MAP_MODULE_STARTED,
	MAP_MODULE_PAUSED,
};

/* registered agent and/or cntlr - internal struct */
struct registered_map_module {
	struct map_module module;
	uint32_t state;
	void *handle;
	struct ubus_subscriber sub;
};

/* TODO: abi versioning */
#define module_sign	"f6dfa346957f45cb8d82dc6b77e2df61"

/* plugin private data struct */
struct map_plugin {
	uint8_t sign[32];
	struct registered_map_module regd_agent;
	struct registered_map_module regd_cntlr;
	/* others ... */

	struct ubus_context *ctx;
	struct ubus_object obj;
};

static int map_plugin_process_cmdu(struct map_plugin *map, char *out_intf,
				   char *dest, uint32_t type,
				   char *data, uint16_t *mid)
{
	bool ret = false;
	uint8_t *cur_stream = NULL;
	uint8_t stream_count = 0, i;
	uint16_t cmdu_type = type & 0xFFFF;
	uint16_t stream_len;
	struct plugin_process_data pdata;
	struct cmdu_stream cstream;

	if (!map)
		return 0;

	if ((map->regd_agent.state != MAP_MODULE_REGISTERED) ||
		(map->regd_cntlr.state != MAP_MODULE_REGISTERED)) {
		PRINT_DBG("Agent/Controller missing, skip error");
		//return 0;
	}

	memset(&pdata, 0, sizeof(struct plugin_process_data));

	if (data) {
		stream_len = strlen(data) / 2;
		cur_stream = malloc(sizeof(uint8_t) * stream_len);
		if (!cur_stream) {
			PRINT_ERR("Insufficient memory!");
			return 0;
		}
		_strtob(cur_stream, data);
	} else {
		stream_len = 0;
	}

	if (cur_stream)
		stream_count = map_count_tlvs_safe(cur_stream, stream_len);

	memset(&cstream, 0, sizeof(struct cmdu_stream));

	cstream.message_version = CMDU_MESSAGE_VERSION_1905_1_2013;
	cstream.message_type = cmdu_type;
	cstream.relay_indicator = map_get_relay_bit(cmdu_type);
	cstream.message_id = *mid;
	cstream.nbr_tlv_streams = stream_count;

	if (stream_count) {
		cstream.tlv_stream_list = malloc(sizeof(uint8_t *) * stream_count);
		if (!cstream.tlv_stream_list) {
			PRINT_ERR("Insufficient memory!");
			free(cur_stream);
			return 0;
		}
	}

	uint8_t *p = cur_stream;
	for (i = 0; i < stream_count; i++) {
		if (stream_len <= 0)
			break;

		uint16_t tlv_len = map_get_first_tlv_length(p);

		cstream.tlv_stream_list[i] =  p;
		p = p + tlv_len;
		stream_len -= tlv_len;
	}

	if (out_intf)
		strncpy(pdata.intf_name, out_intf, IFNAMESIZE);
	else
		strncpy(pdata.intf_name, "lo", IFNAMESIZE);

	if (dest)
		strncpy(pdata.mac_ascii, dest, 18);
	else
		strcpy(pdata.mac_ascii, "00:00:00:00:00:00");


	pdata.cstream = &cstream;

	// flag to determine that mid should be generated or not
	if (map_is_response_cmdu(cmdu_type) || cstream.message_id > 0)
		pdata.flag = false;
	else
		pdata.flag = true;

	pdata.rel_multicast_flag = false;
	// set multicast flag if the destination address is multicast address
	// and cmdu support reliable multicast
	if(strncasecmp(MULTICAST_ADDR_STR, pdata.mac_ascii, 17) == 0) {
		uint8_t ret_trans_type = map_get_transmission_type(cmdu_type);

		pdata.rel_multicast_flag = (ret_trans_type == TRANSMISSION_RELIABLE_MULTICAST);
	}

	// Send the packet accross
	ret = cmdu_dispacher(&pdata);

	PRINT_DBG("## Map mid of sent msg |%d|, ctype|%x| type|%x|", cstream.message_id, cmdu_type, type);
	if (ret && mid)
		*mid = cstream.message_id;

	// Free the tlv allocations
	free(cur_stream);
	free(cstream.tlv_stream_list);

	return 0;
}

static inline int is_same_uid(void *a, void *b)
{
	if (!a || !b)
		return 0;

	return !memcmp(a, b, 16);
}

static int unregister_multiap_module(struct map_plugin *p, struct map_module *m)
{
	bool is_agent, is_cntlr;

	if (!m)
		return -1;

	is_agent = !!(m->role & MAP_ROLE_AGENT);
	is_cntlr = !!(m->role & MAP_ROLE_CONTROLLER);

	if ((is_agent && p->regd_agent.state < MAP_MODULE_REGISTERED) ||
		(is_cntlr && p->regd_cntlr.state < MAP_MODULE_REGISTERED)) {

		PRINT_ERR("module with role = %s is not registered",
			  is_agent ? "Agent" : "Controller");
		return -1;
	}

	update_module_db(false, m->role, m->profile);
	if (is_agent && is_same_uid(p->regd_agent.module.uid, m->uid)) {
		//if (regd_agent->state != MAP_MODULE_PAUSED)
		//	return -1;

		memset(&p->regd_agent.module, 0, sizeof(struct map_module));
		p->regd_agent.state = MAP_MODULE_NOTPRESENT;
		PRINT_ERR("cleanup mod");
		return 0;
	}

	if (is_cntlr && is_same_uid(p->regd_cntlr.module.uid, m->uid)) {
		//if (regd_cntlr->state != MAP_MODULE_PAUSED)
		//	return -1;

		memset(&p->regd_cntlr.module, 0, sizeof(struct map_module));
		p->regd_cntlr.state = MAP_MODULE_NOTPRESENT;
		PRINT_ERR("cleanup mod");
		return 0;
	}

	return -1;
}

static int sub_cb(struct ubus_context *ctx __attribute__((unused)),
		  struct ubus_object *obj __attribute__((unused)),
		  struct ubus_request_data *req __attribute__((unused)),
		  const char *method,
		  struct blob_attr *msg)
{
	char *str;

	str = blobmsg_format_json(msg, true);
	PRINT_ERR("Received notification '%s': %s", method, str);
	free(str);

	return 0;
}

static void sub_remove_cb(struct ubus_context *ctx __attribute__((unused)),
		   struct ubus_subscriber *sub,
		   uint32_t obj)
{
	struct registered_map_module *regd_mod;
	struct map_module *mod;
	struct map_plugin *map = NULL;

	PRINT_ERR("Object %x disappeared", obj);
	regd_mod = container_of(sub, struct registered_map_module, sub);
	mod = &regd_mod->module;

	if (!!(mod->role & MAP_ROLE_AGENT))
		map = container_of(regd_mod, struct map_plugin, regd_agent);
	else if (!!(mod->role & MAP_ROLE_CONTROLLER))
		map = container_of(regd_mod, struct map_plugin, regd_cntlr);

	unregister_multiap_module(map, mod);
}

static int subscription_handler(struct map_module *m, bool subscribe)
{
	struct registered_map_module *regd_mod;
	struct ubus_subscriber *sub;
	struct map_plugin *map;
	uint32_t id;
	int ret;

	if (!m)
		return -1;

	id = *((uint32_t *)m->uid);
	regd_mod = container_of(m, struct registered_map_module, module);

	if (m->role == MAP_ROLE_AGENT)
		map = container_of(regd_mod, struct map_plugin, regd_agent);
	else
		map = container_of(regd_mod, struct map_plugin, regd_cntlr);

	sub = &regd_mod->sub;

	if (subscribe)
		ret = ubus_subscribe(map->ctx, sub, id);
	else
		ret = ubus_unsubscribe(map->ctx, sub, id);

	if (ret)
		PRINT_ERR("Failed subscribe(%d): %s", subscribe, ubus_strerror(ret));

	return 0;
}

static int register_multiap_module(struct map_plugin *p, struct map_module *m)
{
	struct registered_map_module *regd_mod;
	bool is_agent, is_cntlr;
	int ret = 0;

	if (!m)
		return -1;

	is_agent = !!(m->role & MAP_ROLE_AGENT);
	is_cntlr = !!(m->role & MAP_ROLE_CONTROLLER);

	PRINT_DBG("agent = %d   cntlr = %d", is_agent, is_cntlr);

	if ((is_agent && p->regd_agent.state >= MAP_MODULE_REGISTERED) ||
		(is_cntlr && p->regd_cntlr.state >= MAP_MODULE_REGISTERED)) {

		PRINT_ERR("module with role = %s already registered",
				is_agent ? "Agent" : "Controller");
		return -1;
	}

	update_module_db(true, m->role, m->profile);
	regd_mod = is_agent ? &p->regd_agent : &p->regd_cntlr;
	memcpy(&regd_mod->module, m, sizeof(*m));
	regd_mod->state = MAP_MODULE_REGISTERED;

	subscription_handler(&regd_mod->module, true);

	return ret;
}

#if 0	//TODO - maybe later
int start_multiap_module(struct map_module *m)
{
	bool is_agent, is_cntlr;

	if (!m)
		return -1;

	is_agent = !!(m->role & MAP_ROLE_AGENT);
	is_cntlr = !!(m->role & MAP_ROLE_CONTROLLER);

	if (is_agent && is_same_uid(m->uid, regd_agent->module.uid) &&
			regd_agent->state != MAP_MODULE_STARTED) {
		regd_agent->state = MAP_MODULE_STARTED;
		return 0;
	}

	if (is_cntlr && is_same_uid(m->uid, regd_cntlr->module.uid) &&
			regd_cntlr->state != MAP_MODULE_STARTED) {
		regd_cntlr->state = MAP_MODULE_STARTED;
		return 0;
	}

	return -1;
}

int stop_multiap_module(struct map_module *m)
{
	bool is_agent, is_cntlr;

	if (!m)
		return -1;

	is_agent = !!(m->role & MAP_ROLE_AGENT);
	is_cntlr = !!(m->role & MAP_ROLE_CONTROLLER);

	if (is_agent && is_same_uid(m->uid, regd_agent->module.uid) &&
			regd_agent->state != MAP_MODULE_PAUSED) {
		regd_agent->state = MAP_MODULE_PAUSED;
		return 0;
	}

	if (is_cntlr && is_same_uid(m->uid, regd_cntlr->module.uid) &&
			regd_cntlr->state != MAP_MODULE_PAUSED) {
		regd_cntlr->state = MAP_MODULE_PAUSED;
		return 0;
	}

	return -1;
}
#endif


enum {
	REGISTER_MODULE_ID,
	REGISTER_MODULE_DATA,
	_REGISTER_POLICY_MAX,
};

static const struct blobmsg_policy register_policy[_REGISTER_POLICY_MAX] = {
	[REGISTER_MODULE_ID] = { .name = "module", .type = BLOBMSG_TYPE_INT32 },
	[REGISTER_MODULE_DATA] = { .name = "data", .type = BLOBMSG_TYPE_STRING },
};


static int register_map_module(struct ubus_context *ctx __attribute__((unused)),
			       struct ubus_object *obj,
			       struct ubus_request_data *req __attribute__((unused)),
			       const char *method __attribute__((unused)),
			       struct blob_attr *msg)
{
	int dlen = 2 * sizeof(struct map_module) + 1;
	struct blob_attr *tb[_REGISTER_POLICY_MAX];
	char module_data[512] = {0};
	uint32_t module_id = 0;
	int datalen;
	struct map_module m;
	uint8_t module_sig[16] = {0};
	int ret;
	struct map_plugin *map;

	map = container_of(obj, struct map_plugin, obj);

	PRINT_DBG("mapplugin: received 'register'");
	blobmsg_parse(register_policy, _REGISTER_POLICY_MAX, tb,
					blob_data(msg), blob_len(msg));

	if (!tb[REGISTER_MODULE_ID] || !tb[REGISTER_MODULE_DATA]) {
		PRINT_ERR("module or data missing!");
		return UBUS_STATUS_INVALID_ARGUMENT;
	}

	module_id = blobmsg_get_u32(tb[REGISTER_MODULE_ID]);
	datalen = blobmsg_data_len(tb[REGISTER_MODULE_DATA]);
	memcpy(module_data, blobmsg_data(tb[REGISTER_MODULE_DATA]), datalen);

	if (datalen != dlen) {
		PRINT_ERR("module check 1 failed!");
		return UBUS_STATUS_INVALID_ARGUMENT;
	}

	memset(&m, 0, sizeof(struct map_module));

	PRINT_DBG("%s", module_data);
	strtob(module_data, datalen, (unsigned char *)&m);
	strtob(module_sign, strlen(module_sign) + 1, module_sig);

	if (memcmp(m.sign, module_sig, sizeof(module_sig))) {
		PRINT_ERR("module check 2 failed!");
		return UBUS_STATUS_INVALID_ARGUMENT;
	}

	PRINT_DBG("mapplugin: registering module %x ...", module_id);

	*((uint32_t *)m.uid) = module_id;
	ret = register_multiap_module(map, &m);
	if (ret)
		return UBUS_STATUS_UNKNOWN_ERROR;

	PRINT_DBG("mapplugin: module registered OK");
	return UBUS_STATUS_OK;
}

enum {
	SEND_CMDU_TYPE,		/* cmdu type */
	SEND_OUT_INTF,
	SEND_CMDU_DEST,
	SEND_CMDU_MID,
	SEND_CMDU_DATA,		/* tlv stream */
	_SEND_POLICY_MAX,
};

static const struct blobmsg_policy send_policy[_SEND_POLICY_MAX] = {
	[SEND_CMDU_TYPE] = { .name = "type", .type = BLOBMSG_TYPE_INT32 },
	[SEND_OUT_INTF] = { .name = "egress", .type = BLOBMSG_TYPE_STRING },
	[SEND_CMDU_DEST] = { .name = "dst_macaddr", .type = BLOBMSG_TYPE_STRING },
	[SEND_CMDU_MID] = { .name = "mid", .type = BLOBMSG_TYPE_INT32 },
	[SEND_CMDU_DATA] = { .name = "data", .type = BLOBMSG_TYPE_STRING },
};

static const struct blobmsg_policy send_array_policy[_SEND_POLICY_MAX] = {
	[SEND_CMDU_TYPE] = { .name = "type", .type = BLOBMSG_TYPE_INT32 },
	[SEND_OUT_INTF] = { .name = "egress", .type = BLOBMSG_TYPE_STRING },
	[SEND_CMDU_DEST] = { .name = "dst_macaddr", .type = BLOBMSG_TYPE_STRING },
	[SEND_CMDU_MID] = { .name = "mid", .type = BLOBMSG_TYPE_INT32 },
	[SEND_CMDU_DATA] = { .name = "tlv_array", .type = BLOBMSG_TYPE_ARRAY },
};

static int send_tlvstream(struct ubus_context *ctx,
			  struct ubus_object *obj,
			  struct ubus_request_data *req,
			  const char *method,
			  struct blob_attr *msg)
{
	struct blob_attr *tb[_SEND_POLICY_MAX];
	struct map_plugin *map;
	uint32_t cmdu_type = 0;
	char *cmdu_data = NULL, *out_intf = NULL, *dest = NULL;
	struct blob_buf bb = { };
	int ret;
	uint16_t mid = 0;
	size_t buff_len;

	PRINT_DBG("mapplugin: received 'data' from map module");

	map = container_of(obj, struct map_plugin, obj);

	if (strcmp(method, "send") == 0) {
		ret = blobmsg_parse(send_policy, _SEND_POLICY_MAX, tb,
			      blob_data(msg), blob_len(msg));

		if (ret != UBUS_STATUS_OK)
			return UBUS_STATUS_INVALID_ARGUMENT;

		if (tb[SEND_CMDU_DATA])
			cmdu_data = strdup(blobmsg_get_string(tb[SEND_CMDU_DATA]));
	} else {
		ret = blobmsg_parse(send_array_policy, _SEND_POLICY_MAX, tb,
			      blob_data(msg), blob_len(msg));

		if (ret != UBUS_STATUS_OK)
			return UBUS_STATUS_INVALID_ARGUMENT;

		if (tb[SEND_CMDU_DATA]) {
			struct blob_attr *cur;
			size_t rem;

			buff_len = 0;
			blobmsg_for_each_attr(cur, tb[SEND_CMDU_DATA], rem)
				buff_len += blobmsg_len(cur);

			cmdu_data =  calloc(buff_len + 1, sizeof(char));
			if (!cmdu_data) {
				PRINT_ERR("Insufficient memory!");
				return UBUS_STATUS_UNKNOWN_ERROR;
			}

			blobmsg_for_each_attr(cur, tb[SEND_CMDU_DATA], rem)
				strncat(cmdu_data, blobmsg_get_string(cur), blobmsg_len(cur));
		}
	}

	if (!tb[SEND_CMDU_TYPE]) {
		PRINT_ERR("Missing cmdu type !!");
		free(cmdu_data);
		return UBUS_STATUS_INVALID_ARGUMENT;
	}
	cmdu_type = blobmsg_get_u32(tb[SEND_CMDU_TYPE]);

	if (tb[SEND_CMDU_MID])
	    mid = (uint16_t) blobmsg_get_u32(tb[SEND_CMDU_MID]);

	if (tb[SEND_OUT_INTF])
		out_intf = blobmsg_get_string(tb[SEND_OUT_INTF]);

	if (tb[SEND_CMDU_DEST])
		dest = blobmsg_get_string(tb[SEND_CMDU_DEST]);

	ret = map_plugin_process_cmdu(map, out_intf, dest, cmdu_type,
				      cmdu_data, &mid);
	if (ret) {
		free(cmdu_data);
		return UBUS_STATUS_UNKNOWN_ERROR;
	}

	free(cmdu_data);

	blob_buf_init(&bb, 0);
	blobmsg_add_u32(&bb, "mid", mid);
	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	PRINT_DBG("mapplugin: Send CMDU data OK.");
	return UBUS_STATUS_OK;
}

static int map_plugin_create_object(struct map_plugin *map)
{
	struct ubus_object *obj;
	struct ubus_object_type *obj_type;
	struct ubus_method *obj_methods;
	int ret;
	struct ubus_method m[3] = {
		UBUS_METHOD("register", register_map_module, register_policy),
		UBUS_METHOD("send", send_tlvstream, send_policy),
		UBUS_METHOD("send_array", send_tlvstream, send_array_policy)
	};
	//int num_methods = sizeof(m)/sizeof(m[0]);
	int num_methods = ARRAY_SIZE(m);


	obj = &map->obj;
	memset(obj, 0, sizeof(*obj));

	obj_type = calloc(1, sizeof(struct ubus_object_type));
	if (!obj_type)
		return -1;

	obj_methods = calloc(num_methods, sizeof(struct ubus_method));
	if (!obj_methods) {
		free(obj_type);
		return -1;
	}

	obj->name = "map.1905";
	memcpy(obj_methods, m, num_methods * sizeof(struct ubus_method));
	obj->methods = obj_methods;
	obj->n_methods = num_methods;

	obj_type->name = obj->name;
	obj_type->n_methods = obj->n_methods;
	obj_type->methods = obj->methods;
	obj->type = obj_type;

	ret = ubus_add_object(map->ctx, obj);
	if (ret) {
		PRINT_ERR("Failed to add 'obj' err = %s", ubus_strerror(ret));
		free(obj_methods);
		free(obj_type);

		return ret;
	}

	PRINT_DBG("Added 'map.1905'");

	/* init sub for map-modules. This watches for notifications from the
	 * map-modules.
	 */
	map->regd_agent.sub.cb = sub_cb;
	map->regd_agent.sub.remove_cb = sub_remove_cb;
	map->regd_cntlr.sub.cb = sub_cb;
	map->regd_cntlr.sub.remove_cb = sub_remove_cb;
	ret = ubus_register_subscriber(map->ctx, &map->regd_agent.sub);
	if (ret)
		PRINT_ERR("Failed to register agent sub: %s", ubus_strerror(ret));

	ret = ubus_register_subscriber(map->ctx, &map->regd_cntlr.sub);
	if (ret)
		PRINT_ERR("Failed to register cntlr sub: %s", ubus_strerror(ret));

	return 0;
}

static void plugin_init(struct ubus_context *ctx)
{
	struct map_plugin *map = NULL;

	map = calloc(1, sizeof(struct map_plugin));
	if (!map)
		return;

	map->ctx = ctx;
	if (!map->ctx) {
		PRINT_ERR("Failed to connect to ubus");
		free(map);
		return;
	}

	map_plugin_create_object(map);
}

int initialize_map(void *priv)
{
	struct ubus_context *ctx = (struct ubus_context *)priv;

	plugin_init(ctx);
	//dump_map_cmdu_validator();
	return 0;
}

int map_cleanup(void *priv __attribute__((unused)))
{
	// do the cleanup of the modules
	// un-register from ubus
	PRINT_DBG("Cleanup called");
	return 0;
}

int map_validate(void *data)
{
	struct plugin_process_data *pdata;

	pdata = (struct plugin_process_data *) data;

	return map_validate_cmdu_stream(pdata->cstream, pdata->profile, pdata->flag);
}

int map_process(void *data)
{
	struct ubus_context *ctx;
	struct plugin_process_data *pdata;
	uint8_t i;
	struct cmdu_stream *c_stream;
	char *tlv_str_buff, *tlvs;
	uint16_t tlv_len = 0;
	size_t buff_len;

	pdata = (struct plugin_process_data *) data;

	if (pdata == NULL) {
		PRINT_ERR("Invalid input params");
		return false;
	}

	c_stream = pdata->cstream;
	if (c_stream == NULL) {
		PRINT_ERR("## Invalid input cmdu stream ##");
		return false;
	}

	if (false == map_validate_cmdu_stream(c_stream, pdata->profile, pdata->flag))
		return false;

	ctx = ubus_connect(NULL);
	if (!ctx) {
		PRINT_ERR("Failed to connect with ubus");
		return false;
	}

	struct blob_buf bb;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_u32(&bb, "type", c_stream->message_type);
	blobmsg_add_u32(&bb, "mid", c_stream->message_id);
	blobmsg_add_string(&bb, "ingress", pdata->intf_name);
	blobmsg_add_string(&bb, "source", pdata->mac_ascii);

	buff_len = 0;
	for (i = 0; i < c_stream->nbr_tlv_streams; ++i) {
		tlv_len = map_get_first_tlv_length(c_stream->tlv_stream_list[i]);
		buff_len = buff_len + tlv_len;
	}

	tlvs = calloc(sizeof(char), (1 + 6 + buff_len * 2));
	if (!tlvs) {
		PRINT_ERR("Insufficient memory!");
		return false;
	}

	for (i = 0; i < c_stream->nbr_tlv_streams; ++i) {
		tlv_len = map_get_first_tlv_length(c_stream->tlv_stream_list[i]);

		tlv_str_buff = _btostr(c_stream->tlv_stream_list[i], tlv_len);
		strcat(tlvs, tlv_str_buff);
		free(tlv_str_buff);
	}

	// append end-of-message tlv
	strcat(tlvs, "000000");

	blobmsg_add_string(&bb, "tlvs", tlvs);
	ubus_send_event(ctx, "ieee1905.cmdu", bb.head);
	free(tlvs);

	blob_buf_free(&bb);
	ubus_free(ctx);
	return false;
}

struct plugin libwifimap2 = {
	.id = "\x76\x2e\x64\x78",
	.name = "libwifimap2",
	.init = initialize_map,
	.exit = map_cleanup,
	.ops = {
		[PLUGIN_FUNC_PROCESS] = map_process,
		[PLUGIN_FUNC_VALIDATE] = map_validate,
	},
};
