/*
 * map_tlvs.c: MultiAP tlv
 *
 * Author: Yashvardhan <y.yashvardhan@iopsys.eu>
 * Author: Vivek Dutta <vivek.dutta@iopsys.eu>
 *
 * Copyright (C) 2020, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "map2.h"
#include "map_common.h"

uint8_t *map_get_tlv_cstruct(uint8_t *packet_stream)
{
	uint8_t *p;
	uint16_t tlvLen;
	size_t i, j, k;
	uint8_t temp = 0;

	if (packet_stream == NULL)
		return NULL;

	// Skip the tlvType field
	p = packet_stream + 1;

	// Get len
	tlvLen = 0;
	_E2B(&p, &tlvLen);

	switch (*packet_stream) {
		/* 1905 specific TLVs*/
	case TLV_TYPE_END_OF_MESSAGE:
		{
			struct tlv_eom *ret;

			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			// According to the standard, the length *must* be 0
			if (tlvLen != 0) {
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;

			return (uint8_t *) ret;
		}
	case TLV_TYPE_AL_MAC_ADDRESS_TYPE:
		{
			struct tlv_al_mac *ret;

			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			// According to the standard, the length *must* be 6
			if (tlvLen != 6) {
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;

			_EnB(&p, ret->al_mac_address, 6);

			return (uint8_t *) ret;
		}
	case TLV_TYPE_MAC_ADDRESS_TYPE:
		{
			struct tlv_mac *ret;

			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			// According to the standard, the length *must* be 6
			if (tlvLen != 6) {
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;

			_EnB(&p, ret->mac_address, 6);

			return (uint8_t *) ret;
		}
	case TLV_TYPE_DEVICE_INFORMATION_TYPE:
		{
			struct tlv_device_info *ret;

			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;

			_EnB(&p, ret->al_mac_address, 6);
			_E1B(&p, &ret->local_interfaces_nr);

			ret->local_interfaces = malloc(sizeof(*ret->local_interfaces) * (ret->local_interfaces_nr));
			if (!ret->local_interfaces) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}


			for (i = 0; i < ret->local_interfaces_nr; i++) {
				_EnB(&p, ret->local_interfaces[i].mac_address, 6);
				_E2B(&p, &ret->local_interfaces[i].media_type);
				_E1B(&p, &ret->local_interfaces[i].media_specific_data_size);

				if ((ret->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11B_2_4_GHZ) ||
				    (ret->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11G_2_4_GHZ) ||
				    (ret->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11A_5_GHZ) ||
				    (ret->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11N_2_4_GHZ) ||
				    (ret->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11N_5_GHZ) ||
				    (ret->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11AC_5_GHZ) ||
				    (ret->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11AD_60_GHZ) ||
				    (ret->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11AF_GHZ)
				   ) {
					if (ret->local_interfaces[i].media_specific_data_size != 10) {
						// Malformed packet
						free(ret->local_interfaces);
						free(ret);
						return NULL;
					}

					_EnB(&p,
					     ret->local_interfaces[i].media_specific_data.ieee80211.network_membership,
					     6);

					_E1B(&p, &temp);
					ret->local_interfaces[i].media_specific_data.ieee80211.role = temp >> 4;
					_E1B(&p,
					     &ret->local_interfaces[i].media_specific_data.ieee80211.ap_channel_band);

					_E1B(&p,
					     &ret->local_interfaces[i].media_specific_data.ieee80211.ap_ch_freq_1);

					_E1B(&p, &ret->local_interfaces[i].media_specific_data.ieee80211.ap_ch_freq_2);

				} else if ((ret->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_1901_WAVELET) ||
					   (ret->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_1901_FFT)
					  ) {
					if (ret->local_interfaces[i].media_specific_data_size != 7) {
						// Malformed packet
						//
						free(ret->local_interfaces);
						free(ret);
						return NULL;
					}
					_EnB(&p,
					     ret->local_interfaces[i].media_specific_data.ieee1901.network_identifier,
					     7);

				} else {
					if (ret->local_interfaces[i].media_specific_data_size != 0) {
						// Malformed packet
						//
						free(ret->local_interfaces);
						free(ret);
						return NULL;
					}
				}
			}

			if (p - (packet_stream + 3) != tlvLen) {
				// Malformed packet
				//
				free(ret->local_interfaces);
				free(ret);
				return NULL;
			}

			return (uint8_t *) ret;
		}

	case TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES:
		{
			struct tlv_device_bridge_cap *ret;

			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;

			_E1B(&p, &ret->bridging_tuples_nr);

			if (ret->bridging_tuples_nr == 0) {
				ret->bridging_tuples = NULL;
				return (uint8_t *) ret;

			}

			ret->bridging_tuples = malloc(sizeof(*ret->bridging_tuples) * ret->bridging_tuples_nr);
			if (!ret->bridging_tuples) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}


			for (i = 0; i < ret->bridging_tuples_nr; i++) {
				_E1B(&p, &ret->bridging_tuples[i].bridging_tuple_macs_nr);

				if (ret->bridging_tuples[i].bridging_tuple_macs_nr > 0) {
					ret->bridging_tuples[i].bridging_tuple_macs =
						malloc(sizeof(*ret->bridging_tuples[i].bridging_tuple_macs) *
						       ret->bridging_tuples[i].bridging_tuple_macs_nr);

					if (!ret->bridging_tuples[i].bridging_tuple_macs) {
						for (size_t ii = 0; ii < i; ii++) {
							free(ret->bridging_tuples[ii].bridging_tuple_macs);
						}
						free(ret->bridging_tuples);
						free(ret);
						PRINT_ERR("Insufficient memory!");
						return NULL;
					}


					for (j = 0; j < ret->bridging_tuples[i].bridging_tuple_macs_nr; j++) {
						_EnB(&p,
						     ret->bridging_tuples[i].bridging_tuple_macs[j].mac_address,
						     6);
					}
				}
			}

			if (p - (packet_stream + 3) != tlvLen) {
				// Malformed packet
				for (i = 0; i < ret->bridging_tuples_nr; i++)
					free(ret->bridging_tuples[i].bridging_tuple_macs);

				free(ret->bridging_tuples);
				free(ret);
				return NULL;
			}

			return (uint8_t *) ret;
		}

	case TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST:
		{
			struct tlv_non1905_neighbor *ret;

			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (0 != ((tlvLen - 6) % 6)) {
				free(ret);
				return NULL;
			}
			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;

			_EnB(&p, ret->local_mac_address, 6);

			ret->non_1905_neighbors_nr = (tlvLen - 6) / 6;

			if (ret->non_1905_neighbors_nr == 0) {
				ret->non_1905_neighbors = NULL;
				return (uint8_t *) ret;
			}

			ret->non_1905_neighbors = malloc(sizeof(*ret->non_1905_neighbors) *
							 ret->non_1905_neighbors_nr);

			if (!ret->non_1905_neighbors) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			for (i = 0; i < ret->non_1905_neighbors_nr; i++)
				_EnB(&p, ret->non_1905_neighbors[i].mac_address, 6);

			return (uint8_t *) ret;
		}

	case TLV_TYPE_NEIGHBOR_DEVICE_LIST:
		{
			struct tlv_neighbor_device *ret;

			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (0 != ((tlvLen - 6) % 7)) {
				free(ret);
				return NULL;
			}
			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;

			_EnB(&p, ret->local_mac_address, 6);

			ret->neighbors_nr = (tlvLen - 6) / 7;
			if (ret->neighbors_nr == 0) {
				ret->neighbors = NULL;
				return (uint8_t *) ret;
			}

			ret->neighbors = malloc(sizeof(*ret->neighbors) *
						ret->neighbors_nr);

			if (!ret->neighbors) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}


			for (i = 0; i < ret->neighbors_nr; i++) {
				_EnB(&p, ret->neighbors[i].mac_address, 6);
				_E1B(&p, &temp);
				ret->neighbors[i].bridge_flag = BIT(temp, 7);
			}

			return (uint8_t *) ret;
		}
	case TLV_TYPE_TRANSMITTER_LINK_METRIC:
		{
			struct tlv_tx_link_metric *ret;

			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if ((12 + 29 * 1) > tlvLen) {
				free(ret);
				return NULL;
			}
			if (0 != (tlvLen - 12) % 29) {
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;

			_EnB(&p, ret->local_al_address, 6);
			_EnB(&p, ret->neighbor_al_address, 6);

			ret->transmitter_link_metrics_nr = (tlvLen - 12) / 29;

			if (ret->transmitter_link_metrics_nr == 0) {
				ret->transmitter_link_metrics = NULL;
				return (uint8_t *) ret;
			}

			ret->transmitter_link_metrics = malloc(sizeof(*ret->transmitter_link_metrics) *
							       ret->transmitter_link_metrics_nr);
			if (!ret->transmitter_link_metrics) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}


			for (i = 0; i < ret->transmitter_link_metrics_nr; i++) {
				_EnB(&p, ret->transmitter_link_metrics[i].local_interface_address, 6);
				_EnB(&p, ret->transmitter_link_metrics[i].neighbor_interface_address, 6);

				_E2B(&p, &ret->transmitter_link_metrics[i].intf_type);
				_E1B(&p, &ret->transmitter_link_metrics[i].bridge_flag);
				_E4B(&p, &ret->transmitter_link_metrics[i].packet_errors);
				_E4B(&p, &ret->transmitter_link_metrics[i].transmitted_packets);
				_E2B(&p, &ret->transmitter_link_metrics[i].mac_throughput_capacity);
				_E2B(&p, &ret->transmitter_link_metrics[i].link_availability);
				_E2B(&p, &ret->transmitter_link_metrics[i].phy_rate);
			}

			if (p - (packet_stream + 3) != tlvLen) {
				free(ret->transmitter_link_metrics);
				free(ret);
				return NULL;
			}

			return (uint8_t *) ret;
		}

	case TLV_TYPE_RECEIVER_LINK_METRIC:
		{
			struct tlv_rx_link_metric *ret;

			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = packet_stream + 1;
			_E2B(&p, &tlvLen);

			if ((12 + 23 * 1) > tlvLen) {
				free(ret);
				return NULL;
			}
			if (0 != (tlvLen - 12) % 23) {
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;

			_EnB(&p, ret->local_al_address, 6);
			_EnB(&p, ret->neighbor_al_address, 6);

			ret->receiver_link_metrics_nr = (tlvLen - 12) / 23;

			if (ret->receiver_link_metrics_nr == 0) {
				ret->receiver_link_metrics = NULL;
				return (uint8_t *)ret;
			}

			ret->receiver_link_metrics = malloc(sizeof(*ret->receiver_link_metrics) *
							    ret->receiver_link_metrics_nr);
			if (!ret->receiver_link_metrics) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}


			for (i = 0; i < ret->receiver_link_metrics_nr; i++) {
				_EnB(&p, ret->receiver_link_metrics[i].local_interface_address, 6);
				_EnB(&p, ret->receiver_link_metrics[i].neighbor_interface_address, 6);

				_E2B(&p, &ret->receiver_link_metrics[i].intf_type);
				_E4B(&p, &ret->receiver_link_metrics[i].packet_errors);
				_E4B(&p, &ret->receiver_link_metrics[i].packets_received);
				_E1B(&p, &ret->receiver_link_metrics[i].rssi);
			}

			if (p - (packet_stream + 3) != tlvLen) {
				free(ret->receiver_link_metrics);
				free(ret);
				return NULL;
			}

			return (uint8_t *) ret;
		}

	case TLV_TYPE_VENDOR_SPECIFIC:
		{
			struct tlv_vendor_specific *ret;

			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = packet_stream + 1;

			_E2B(&p, &tlvLen);
			_EnB(&p, ret->vendor_oui, 3);
			_E2B(&p, &ret->m_nr);

			if ((3 + 2 + ret->m_nr) > tlvLen) {
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;

			ret->m = calloc(1, ret->m_nr);
			if (!ret->m) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			_EnB(&p, ret->m, ret->m_nr);
			return (uint8_t *) ret;
		}

	case TLV_TYPE_SEARCHED_ROLE:
		{
			struct tlv_searched_role *ret;

			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			// According to the standard, the length *must* be 1
			if (tlvLen != 1) {
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;

			_E1B(&p, &ret->role);

			return (uint8_t *) ret;
		}

	case TLV_TYPE_AUTOCONFIG_FREQ_BAND:
		{
			struct tlv_autoconf_freq_band *ret;

			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			// According to the standard, the length *must* be 1
			if (tlvLen != 1) {
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;

			_E1B(&p, &ret->freq_band);

			return (uint8_t *) ret;
		}

	case TLV_TYPE_SUPPORTED_ROLE:
		{
			struct tlv_supported_role *ret;

			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			// According to the standard, the length *must* be 1
			if (tlvLen != 1) {
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;

			_E1B(&p, &ret->role);

			return (uint8_t *) ret;
		}

	case TLV_TYPE_SUPPORTED_FREQ_BAND:
		{
			struct tlv_supported_freq_band *ret;

			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			// According to the standard, the length *must* be 1
			if (tlvLen != 1) {
				// Malformed packet
				//
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;

			_E1B(&p, &ret->freq_band);

			return (uint8_t *) ret;
		}

	case TLV_TYPE_WSC:
		{
			struct tlv_wsc *ret;

			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			ret->wsc_frame_size = tlvLen;

			if (tlvLen > 0) {
				ret->wsc_frame = malloc(sizeof(*ret->wsc_frame) *
							tlvLen);
				if (!ret->wsc_frame) {
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}
				_EnB(&p, ret->wsc_frame, tlvLen);
			}

			return (uint8_t *) ret;
		}

		/* map specific tlvs*/
	case MAP_TLV_SUPPORTED_SERVICE:
		{
			struct tlv_supp_service *ret;

			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;

			if (tlvLen == 0) {
				free(ret);
				return NULL;
			}

			_E1B(&p, &ret->supported_services_list);

			if (ret->supported_services_list == 0) {
				ret->supported_services = NULL;
				return (uint8_t *) ret;
			}

			ret->supported_services = malloc(sizeof(*ret->supported_services) *
							 ret->supported_services_list);
			if (!ret->supported_services) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			for (i = 0; i < ret->supported_services_list; i++)
				_E1B(&p, &ret->supported_services[i].service);

			return (uint8_t *) ret;
		}
	case MAP_TLV_SEARCHED_SERVICE:
		{
			// tlv structure pointer
			struct tlv_searched_service *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E1B(&p, &ret->searched_services_list);

			if (ret->searched_services_list == 0) {
				ret->service = NULL;
				return (uint8_t *) ret;
			}

			ret->service = malloc(sizeof(*ret->service) *
					      ret->searched_services_list);
			if (!ret->service) {
				free(ret->service);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			for (i = 0; i < ret->searched_services_list; ++i)
				_E1B(&p, &ret->service[i]);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_AP_RADIO_IDENTIFIER:
		{
			// tlv structure pointer
			struct tlv_ap_radio_identifier *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			// tlv length check from standard
			if (tlvLen != 6) {
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->radio_id, 6);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_AP_OPERATIONAL_BSS:
		{
			struct tlv_ap_oper_bss *ret;

			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;

			if (tlvLen == 0) {
				free(ret);
				return NULL;
			}

			_E1B(&p, &ret->radios_nr);

			if (ret->radios_nr == 0) {
				ret->radio = NULL;
				return (uint8_t *) ret;
			}

			ret->radio = malloc(sizeof(*ret->radio) *
					    ret->radios_nr);
			if (!ret->radio) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}


			for (i = 0; i < ret->radios_nr; i++) {
				_EnB(&p, ret->radio[i].radio_id, 6);
				_E1B(&p, &ret->radio[i].bss_nr);

				if (ret->radio[i].bss_nr == 0)
					continue;

				ret->radio[i].bss = malloc(sizeof(*ret->radio[i].bss) *
							   ret->radio[i].bss_nr);
				if (!ret->radio[i].bss) {
					for(size_t ii = 0; ii < i; ii++)
						free(ret->radio[ii].bss);

					free(ret->radio);
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}

				for (j = 0; j < ret->radio[i].bss_nr; j++) {
					_EnB(&p, ret->radio[i].bss[j].bssid, 6);
					_E1B(&p, &ret->radio[i].bss[j].ssid_len);
					_EnB(&p, ret->radio[i].bss[j].ssid,
					     ret->radio[i].bss[j].ssid_len);
				}

			}

			return (uint8_t *) ret;
		}
	case MAP_TLV_ASSOCIATED_CLIENTS:
		{
			struct tlv_assoc_client *ret;

			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;

			if (tlvLen == 0) {
				free(ret);
				return NULL;
			}

			_E1B(&p, &ret->bss_nr);

			if (ret->bss_nr == 0)
				return (uint8_t *) ret;

			ret->bss = malloc(sizeof(*ret->bss) *
					  ret->bss_nr);
			if (!ret->bss) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			for (i = 0; i < ret->bss_nr; i++) {
				_EnB(&p, ret->bss[i].bssid, 6);
				_E2B(&p, &ret->bss[i].assoc_clients_nr);
				if (ret->bss[i].assoc_clients_nr == 0)
					continue;

				ret->bss[i].clients = malloc(sizeof(*ret->bss[i].clients) *
							     ret->bss[i].assoc_clients_nr);
				if (!ret->bss[i].clients) {
					for (size_t ii = 0; ii < i; ii++) {
						if (ret->bss[ii].assoc_clients_nr)
							free(ret->bss[ii].clients);
					}
					free(ret->bss);
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}

				for (j = 0; j < ret->bss[i].assoc_clients_nr; j++) {
					_EnB(&p, ret->bss[i].clients[j].client_addr, 6);
					_E2B(&p, &ret->bss[i].clients[j].uptime);
				}
			}

			return (uint8_t *) ret;
		}
	case MAP_TLV_AP_CAPABILITY:
		{
			// Multiap spec Section 17.2.6
			struct tlv_ap_cap *ret;

			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen != 1) {
				free(ret);
				return NULL;
			}

			ret->tlv_type = MAP_TLV_AP_CAPABILITY;
			ret->tlv_len = tlvLen;
			_E1B(&p, &temp);

			ret->op_ch_metric_reporting = BIT(7, temp);
			ret->non_op_ch_metric_reporting = BIT(6, temp);
			ret->agent_init_rcpi_steering = BIT(5, temp);

			return (uint8_t *) ret;
		}
	case MAP_TLV_AP_RADIO_BASIC_CAPABILITIES:
		{
			// tlv structure pointer
			struct tlv_ap_radio_basic_cap *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->radio_id, 6);
			_E1B(&p, &ret->max_bss_nr);
			_E1B(&p, &ret->operating_classes_nr);

			if (ret->operating_classes_nr == 0)
				return (uint8_t *) ret;

			ret->operating_class = malloc(sizeof(*ret->operating_class) *
						      ret->operating_classes_nr);
			if (!ret->operating_class) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}


			for (i = 0; i < ret->operating_classes_nr; ++i) {
				_E1B(&p, &ret->operating_class[i].op_class);
				_E1B(&p, &ret->operating_class[i].max_tx_power);
				_E1B(&p, &ret->operating_class[i].non_op_ch_nr);

				if (ret->operating_class[i].non_op_ch_nr == 0)
					continue;

				ret->operating_class[i].channel =
					malloc(sizeof(*ret->operating_class[i].channel) *
					       ret->operating_class[i].non_op_ch_nr);
				if (!ret->operating_class[i].channel) {
					PRINT_ERR("Insufficient memory!");
					for (size_t ii = 0; ii < i; ii++) {
						if (ret->operating_class[ii].non_op_ch_nr)
							free(ret->operating_class[ii].channel);
					}
					free(ret->operating_class);
					free(ret);
					return NULL;
				}

				for (j = 0; j < ret->operating_class[i].non_op_ch_nr; ++j)
					_E1B(&p, &ret->operating_class[i].channel[j]);
			}

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_AP_HT_CAPABILITIES:
		{
			// tlv structure pointer
			struct tlv_ap_ht_cap *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			// tlv length check from standard
			if (tlvLen != 7) {
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->radio_id, 6);

			_E1B(&p, &temp);
			ret->max_tx_streams_supported = BITS(7, 6, temp);
			ret->max_rx_streams_supported = BITS(5, 4, temp);
			ret->gi_20_support = BIT(3, temp);
			ret->gi_40_support = BIT(2, temp);
			ret->ht_40_support = BIT(1, temp);
			ret->reserved = BIT(0, temp);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_AP_VHT_CAPABILITIES:
		{
			// tlv structure pointer
			struct tlv_ap_vht_cap *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			// tlv length check from standard
			if (tlvLen != 12) {
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->radio_id, 6);
			_E2B(&p, &ret->vht_tx_mcs_supported);
			_E2B(&p, &ret->vht_rx_mcs_supported);

			_E1B(&p, &temp);
			ret->max_tx_streams_supported = BITS(7, 5, temp);
			ret->max_rx_streams_supported = BITS(4, 2, temp);
			ret->gi_80_support = BIT(1, temp);
			ret->gi_160_support = BIT(0, temp);

			_E1B(&p, &temp);
			ret->vht_8080_support = BIT(7, temp);
			ret->vht_160_support = BIT(6, temp);
			ret->su_beamformer_capable = BIT(5, temp);
			ret->mu_beamformer_capable = BIT(4, temp);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_AP_HE_CAPABILITIES:
		{
			// tlv structure pointer
			struct tlv_ap_he_cap *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->radio_id, 6);
			_E1B(&p, &ret->he_mcs_len);

			if (ret->he_mcs_len) {
				ret->he_mcs_supported =
					malloc(sizeof(*ret->he_mcs_supported) *
					       ret->he_mcs_len);

				if (!ret->he_mcs_supported) {
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}

				for (i = 0; i < ret->he_mcs_len; ++i)
					_E1B(&p, &ret->he_mcs_supported[i]);
			}

			_E1B(&p, &temp);
			ret->max_tx_streams_supported = BITS(7, 5, temp);
			ret->max_rx_streams_supported = BITS(4, 2, temp);
			ret->he_8080_support = BIT(1, temp);
			ret->he_160_support = BIT(0, temp);

			_E1B(&p, &temp);
			ret->su_beamformer_capable = BIT(7, temp);
			ret->mu_beamformer_capable = BIT(6, temp);
			ret->ul_mu_mimo_capable = BIT(5, temp);
			ret->ul_mu_mimo_ofdma_capable = BIT(4, temp);
			ret->dl_mu_mimo_ofdma_capable = BIT(3, temp);
			ret->ul_ofdma_capable = BIT(2, temp);
			ret->dl_ofdma_capable = BIT(1, temp);
			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_STEERING_POLICY:
		{
			// tlv structure pointer
			struct tlv_steering_policy *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E1B(&p, &ret->local_disallowed_sta_nr);
			if (ret->local_disallowed_sta_nr) {
				ret->local_disallowed_sta_macs =
					malloc(sizeof(*ret->local_disallowed_sta_macs) *
					       ret->local_disallowed_sta_nr);
				if (!ret->local_disallowed_sta_macs) {
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}

				for (i = 0; i < ret->local_disallowed_sta_nr; ++i)
					_EnB(&p, &ret->local_disallowed_sta_macs[i].addr, 6);
			}

			_E1B(&p, &ret->btm_disallowed_sta_nr);
			if (ret->btm_disallowed_sta_nr) {
				ret->btm_disallowed_sta_macs =
					malloc(sizeof(*ret->btm_disallowed_sta_macs) *
					       ret->btm_disallowed_sta_nr);
				if (!ret->btm_disallowed_sta_macs) {
					free(ret->local_disallowed_sta_macs);
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}

				for (i = 0; i < ret->btm_disallowed_sta_nr; ++i)
					_EnB(&p, &ret->btm_disallowed_sta_macs[i].addr, 6);
			}

			_E1B(&p, &ret->control_policy_radio_nr);
			if (ret->control_policy_radio_nr) {
				ret->control_policy =
					malloc(sizeof(*ret->control_policy) *
					       ret->control_policy_radio_nr);

				if (!ret->control_policy) {
					free(ret->btm_disallowed_sta_macs);
					free(ret->local_disallowed_sta_macs);
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}

				for (i = 0; i < ret->control_policy_radio_nr; ++i) {
					_EnB(&p, &ret->control_policy[i].radio_id, 6);
					_E1B(&p, &ret->control_policy[i].steering_policy);
					_E1B(&p, &ret->control_policy[i].channel_utilization_thres);
					_E1B(&p, &ret->control_policy[i].rcpi_steering_thres);
				}
			}
			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_METRIC_REPORTING_POLICY:
		{
			// tlv structure pointer
			struct tlv_metric_report_policy *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E1B(&p, &ret->interval);
			_E1B(&p, &ret->metric_reporting_policy_radio_nr);

			if (ret->metric_reporting_policy_radio_nr == 0)
				return (uint8_t *) ret;

			ret->metric_reporting_policy = malloc(sizeof(*ret->metric_reporting_policy) *
							      ret->metric_reporting_policy_radio_nr);
			if (!ret->metric_reporting_policy) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			for (i = 0; i < ret->metric_reporting_policy_radio_nr; ++i) {
				_EnB(&p, &ret->metric_reporting_policy[i].radio_id, 6);
				_E1B(&p, &ret->metric_reporting_policy[i].rcpi_thres);
				_E1B(&p, &ret->metric_reporting_policy[i].rcpi_hysteresis_margin);
				_E1B(&p, &ret->metric_reporting_policy[i].channel_utilization_thres);

				_E1B(&p, &temp);
				ret->metric_reporting_policy[i].is_assoc_sta_traffic_stats = BIT(7, temp);
				ret->metric_reporting_policy[i].is_assoc_sta_link_metrics = BIT(6, temp);
			}
			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_CHANNEL_PREFERENCE:
		{
			// tlv structure pointer
			struct tlv_channel_pref *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->radio_id, 6);
			_E1B(&p, &ret->ch_preference_op_class_nr);

			if (ret->ch_preference_op_class_nr == 0)
				return (uint8_t *) ret;

			ret->op_class =
				malloc(sizeof(*ret->op_class) *
				       ret->ch_preference_op_class_nr);
			if (!ret->op_class) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			for (i = 0; i < ret->ch_preference_op_class_nr; ++i) {
				_E1B(&p, &ret->op_class[i].op_class);
				_E1B(&p, &ret->op_class[i].channel_nr);

				if (ret->op_class[i].channel_nr) {
					ret->op_class[i].channel_list =
						malloc(sizeof(*ret->op_class[i].channel_list) *
						       ret->op_class[i].channel_nr);
					if (!ret->op_class[i].channel_list) {
						free(ret->op_class);
						free(ret);
						PRINT_ERR("Insufficient memory!");
						return NULL;
					}

					for (j = 0; j < ret->op_class[i].channel_nr; ++j)
						_E1B(&p, &ret->op_class[i].channel_list[j]);
				}

				_E1B(&p, &temp);
				ret->op_class[i].preference = BITS(7, 4, temp);
				ret->op_class[i].preference_reason = BITS(3, 0, temp);
			}
			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_RADIO_OPERATION_RESTRICTION:
		{
			// tlv structure pointer
			struct tlv_radio_oper_restrict *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->radio_id, 6);
			_E1B(&p, &ret->op_restricted_op_class_nr);

			if (ret->op_restricted_op_class_nr == 0)
				return (uint8_t *) ret;

			ret->restricted_op_class =
				malloc(sizeof(*ret->restricted_op_class) *
				       ret->op_restricted_op_class_nr);
			if (!ret->restricted_op_class) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			for (i = 0; i < ret->op_restricted_op_class_nr; ++i) {
				_E1B(&p, &ret->restricted_op_class[i].op_class);
				_E1B(&p, &ret->restricted_op_class[i].channel_nr);

				if (ret->restricted_op_class[i].channel_nr == 0)
					continue;

				ret->restricted_op_class[i].restricted_channels =
					malloc(sizeof(*ret->restricted_op_class[i].restricted_channels) *
					       ret->restricted_op_class[i].channel_nr);
				if (!ret->restricted_op_class[i].restricted_channels) {
					free(ret->restricted_op_class);
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}

				for (j = 0; j < ret->restricted_op_class[i].channel_nr; ++j) {
					_E1B(&p, &ret->restricted_op_class[i].restricted_channels[j].channel);
					_E1B(&p, &ret->restricted_op_class[i].restricted_channels[j].min_freq_sep);
				}
			}
			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_TRANSMIT_POWER_LIMIT:
		{
			// tlv structure pointer
			struct tlv_tx_power_limit *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			// tlv length check from standard
			if (tlvLen != 7) {
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->radio_id, 6);
			_E1B(&p, &ret->tx_power_limit);
			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_CHANNEL_SELECTION_RESPONSE:
		{
			// tlv structure pointer
			struct tlv_ch_selection_resp *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			// tlv length check from standard
			if (tlvLen != 7) {
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->radio_id, 6);
			_E1B(&p, &ret->response_code);
			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_OPERATING_CHANNEL_REPORT:
		{
			// tlv structure pointer
			struct tlv_oper_ch_report *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->radio_id, 6);
			_E1B(&p, &ret->op_ch_op_class_nr);

			if (ret->op_ch_op_class_nr != 0)
			{
				ret->op_ch_op_class = malloc(sizeof(*ret->op_ch_op_class) *
						     ret->op_ch_op_class_nr);
				if (!ret->op_ch_op_class) {
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}

				for (i = 0; i < ret->op_ch_op_class_nr; ++i) {
					_E1B(&p, &ret->op_ch_op_class[i].op_class);
					_E1B(&p, &ret->op_ch_op_class[i].channel);
				}
			}
			_E1B(&p, &ret->curr_tx_power);
			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_CLIENT_INFO:
		{
			// tlv structure pointer
			struct tlv_client_info *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			// tlv length check from standard
			if (tlvLen != 12) {
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->bssid, 6);
			_EnB(&p, &ret->client_addr, 6);
			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_CLIENT_CAPABILITY_REPORT:
		{
			// tlv structure pointer
			struct tlv_client_cap_report *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E1B(&p, &ret->result_code);
			if(ret->result_code == 0x00) {
				uint16_t body_len = *p;
				ret->frame_body = malloc(body_len * sizeof(*ret->frame_body));
				memset(ret->frame_body, 0, body_len);
				_EnB(&p, ret->frame_body, body_len);
			}

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_CLIENT_ASSOCIATION_EVENT:
		{
			// Multiap spec Section 17.2.6
			struct tlv_client_assoc_event *ret;

			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen != 13) {
				free(ret);
				return NULL;
			}
			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			_EnB(&p, ret->client_addr, 6);
			_EnB(&p, ret->bssid, 6);

			_E1B(&p, &temp);

			ret->assoc_event = BIT(7, temp);
			return (uint8_t *) ret;
		}
	case MAP_TLV_AP_METRIC_QUERY:
		{
			// tlv structure pointer
			struct tlv_ap_metric_query *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E1B(&p, &ret->bssid_nr);
			if (ret->bssid_nr == 0)
				return (uint8_t *) ret;

			ret->ap_metric_query_bssid =
				malloc(sizeof(*ret->ap_metric_query_bssid) *
				       ret->bssid_nr);

			if (!ret->ap_metric_query_bssid) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			for (i = 0; i < ret->bssid_nr; ++i)
				_EnB(&p, &ret->ap_metric_query_bssid[i].bssid, 6);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_AP_METRICS:
		{
			// tlv structure pointer
			struct tlv_ap_metrics *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->bssid, 6);
			_E1B(&p, &ret->channel_utilization);
			_E2B(&p, &ret->total_assoc_stas);

			_E1B(&p, &temp);
			ret->is_ac_be = BIT(7, temp);
			ret->is_ac_bk = BIT(6, temp);
			ret->is_ac_vo = BIT(5, temp);
			ret->is_ac_vi = BIT(4, temp);
			_EnB(&p, &ret->service_param_info_be, 3);

			if (ret->is_ac_bk)
				_EnB(&p, &ret->service_param_info_bk, 3);

			if (ret->is_ac_vo)
				_EnB(&p, &ret->service_param_info_vo, 3);

			if (ret->is_ac_vi)
				_EnB(&p, &ret->service_param_info_vi, 3);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_STA_MAC_ADDRESS:
		{
			// tlv structure pointer
			struct tlv_sta_mac *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			// tlv length check from standard
			if (tlvLen != 6) {
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->addr, 6);
			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_ASSOCIATED_STA_LINK_METRICS:
		{
			// tlv structure pointer
			struct tlv_assoc_sta_link_metrics *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->addr, 6);
			_E1B(&p, &ret->bssid_nr);
			if (ret->bssid_nr == 0)
				return (uint8_t *) ret;

			ret->sta_link_metrics_bssid =
				malloc(sizeof(*ret->sta_link_metrics_bssid) *
				       ret->bssid_nr);

			if (!ret->sta_link_metrics_bssid) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			for (i = 0; i < ret->bssid_nr; ++i) {
				_EnB(&p, &ret->sta_link_metrics_bssid[i].bssid, 6);
				_E4B(&p, &ret->sta_link_metrics_bssid[i].time_delta);
				_E4B(&p, &ret->sta_link_metrics_bssid[i].dl_mac_data_rate);
				_E4B(&p, &ret->sta_link_metrics_bssid[i].ul_mac_data_rate);
				_E1B(&p, &ret->sta_link_metrics_bssid[i].ul_rcpi);
			}
			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_UNASSOCIATED_STA_LINK_METRICS_QUERY:
		{
			// tlv structure pointer
			struct tlv_unassoc_sta_link_metrics_query *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E1B(&p, &ret->op_class);
			_E1B(&p, &ret->ch_nr);

			if (ret->ch_nr == 0)
				return (uint8_t *) ret;

			ret->unassoc_sta_link_metrics_query_ch =
				malloc(sizeof(*ret->unassoc_sta_link_metrics_query_ch) *
				       ret->ch_nr);
			if (!ret->unassoc_sta_link_metrics_query_ch) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			for (i = 0; i < ret->ch_nr; ++i) {
				_E1B(&p, &ret->unassoc_sta_link_metrics_query_ch[i].channel);
				_E1B(&p, &ret->unassoc_sta_link_metrics_query_ch[i].sta_mac_nr);

				if (ret->unassoc_sta_link_metrics_query_ch[i].sta_mac_nr == 0)
					continue;

				ret->unassoc_sta_link_metrics_query_ch[i].sta_mac =
					malloc(sizeof(*ret->unassoc_sta_link_metrics_query_ch[i].sta_mac) *
					       ret->unassoc_sta_link_metrics_query_ch[i].sta_mac_nr);

				if (!ret->unassoc_sta_link_metrics_query_ch[i].sta_mac) {
					free(ret->unassoc_sta_link_metrics_query_ch);
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}

				for (j = 0; j < ret->unassoc_sta_link_metrics_query_ch[i].sta_mac_nr; ++j)
					_EnB(&p, &ret->unassoc_sta_link_metrics_query_ch[i].sta_mac[j].addr, 6);

			}
			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_UNASSOCIATED_STA_LINK_METRICS_RESPONSE:
		{
			// tlv structure pointer
			struct tlv_unassoc_sta_link_metrics_resp *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E1B(&p, &ret->op_class);
			_E1B(&p, &ret->sta_nr);
			
			if(ret->sta_nr == 0)
				return (uint8_t *) ret;

			ret->unassoc_sta_link_metrics_sta = malloc(sizeof(*ret->unassoc_sta_link_metrics_sta) *
								   ret->sta_nr);
			if (!ret->unassoc_sta_link_metrics_sta) {
				free(ret);
				return NULL;
			}

			for (i = 0; i < ret->sta_nr; ++i) {
				_EnB(&p, &ret->unassoc_sta_link_metrics_sta[i].addr, 6);
				_E1B(&p, &ret->unassoc_sta_link_metrics_sta[i].channel);
				_E4B(&p, &ret->unassoc_sta_link_metrics_sta[i].time_delta);
				_E1B(&p, &ret->unassoc_sta_link_metrics_sta[i].ul_rcpi);
			}
			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_BEACON_METRICS_QUERY:
		{
			// tlv structure pointer
			struct tlv_beacon_metrics_query *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->sta_addr, 6);
			_E1B(&p, &ret->op_class);
			_E1B(&p, &ret->channel);
			_EnB(&p, &ret->bssid, 6);
			_E1B(&p, &ret->reporting_detail);
			_E1B(&p, &ret->ssid_len);
			memset(&ret->ssid, 0x0, 33);
			_EnB(&p, ret->ssid, ret->ssid_len);

			_E1B(&p, &ret->ch_reports_nr);

			if (ret->ch_reports_nr == 0)
				return (uint8_t *) ret;

			ret->ch_report =
				malloc(sizeof(*ret->ch_report) *
				       ret->ch_reports_nr);

			if (!ret->ch_report) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			for (i = 0; i < ret->ch_reports_nr; ++i) {
				_E1B(&p, &ret->ch_report[i].report_len);
				_E1B(&p, &ret->ch_report[i].op_class);
				ret->ch_report[i].ch_list =
					malloc(sizeof(*ret->ch_report[i].ch_list) *
					       (ret->ch_report[i].report_len - 1));

				if (!ret->ch_report[i].ch_list) {
					free(ret->ch_report);
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}

				size_t chLen = (size_t) ret->ch_report[i].report_len - 1;

				for (j = 0; j < chLen; ++j)
					_E1B(&p, &ret->ch_report[i].ch_list[j]);

			}
			_E1B(&p, &ret->elem_ids_nr);
			if (ret->elem_ids_nr == 0)
				return (uint8_t *) ret;

			ret->elem_list =
				malloc(sizeof(*ret->elem_list) *
				       ret->elem_ids_nr);
			if (!ret->elem_list) {
				for (i = 0; i < ret->ch_reports_nr; ++i)
					free(ret->ch_report[i].ch_list);

				free(ret->ch_report);
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			for (i = 0; i < ret->elem_ids_nr; ++i)
				_E1B(&p, &ret->elem_list[i]);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_BEACON_METRICS_RESPONSE:
		{
			// tlv structure pointer
			struct tlv_beacon_metrics_resp *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->sta_addr, 6);
			_E1B(&p, &ret->reserved);
			_E1B(&p, &ret->report_elems_nr);

			if (ret->report_elems_nr == 0)
				return (uint8_t *) ret;

			ret->report_elem =
				malloc(sizeof(*ret->report_elem) *
				       ret->report_elems_nr);
			if (!ret->report_elem) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			for (i = 0; i < ret->report_elems_nr; ++i)
				_E1B(&p, &ret->report_elem[i]);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_STEERING_REQUEST:
		{
			// tlv structure pointer
			struct tlv_steer_Req *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->bssid, 6);

			_E1B(&p, &temp);
			ret->request_mode = BIT(7, temp);
			ret->btm_disassoc_imminent = BIT(6, temp);
			ret->btm_abridged = BIT(5, temp);
			_E2B(&p, &ret->steer_opp_window);
			_E2B(&p, &ret->btm_disassoc_timer);
			_E1B(&p, &ret->sta_list_cnt);

			if (ret->sta_list_cnt) {
				ret->steering_req_macs =
					malloc(sizeof(*ret->steering_req_macs) *
					       ret->sta_list_cnt);
				if (!ret->steering_req_macs) {
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}

				for (i = 0; i < ret->sta_list_cnt; ++i)
					_EnB(&p, &ret->steering_req_macs[i].addr, 6);
			}

			_E1B(&p, &ret->target_bssid_list_cnt);
			if (ret->target_bssid_list_cnt) {
				ret->steering_req_target_bssids =
					malloc(sizeof(*ret->steering_req_target_bssids) *
					       ret->target_bssid_list_cnt);
				if (!ret->steering_req_target_bssids) {
					free(ret->steering_req_macs);
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}

				for (i = 0; i < ret->target_bssid_list_cnt; ++i) {
					_EnB(&p, &ret->steering_req_target_bssids[i].bssid, 6);
					_E1B(&p, &ret->steering_req_target_bssids[i].op_class);
					_E1B(&p, &ret->steering_req_target_bssids[i].channel);
				}
			}
			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_STEERING_BTM_REPORT:
		{
			// tlv structure pointer
			struct tlv_steer_btm_report *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->bssid, 6);
			_EnB(&p, &ret->sta_mac_addr, 6);
			_E1B(&p, &ret->status_code);
			// zero value means success
			if (ret->status_code == 0x00)
				_EnB(&p, &ret->target_bssid, 6);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_CLIENT_ASSOCIATION_CONTROL_REQUEST:
		{
			// tlv structure pointer
			struct tlv_client_assoc_control_req *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->bssid, 6);
			_E1B(&p, &ret->assoc_control);
			_E2B(&p, &ret->validity_period);
			_E1B(&p, &ret->sta_list_cnt);

			if (ret->sta_list_cnt == 0)
				return (uint8_t *) ret;

			ret->client_assoc_ctrl_req_stas =
				malloc(sizeof(*ret->client_assoc_ctrl_req_stas) *
				       ret->sta_list_cnt);

			if (!ret->client_assoc_ctrl_req_stas) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}


			for (i = 0; i < ret->sta_list_cnt; ++i)
				_EnB(&p, &ret->client_assoc_ctrl_req_stas[i].addr, 6);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_BACKHAUL_STEERING_REQUEST:
		{
			// tlv structure pointer
			struct tlv_backhaul_steer_req *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			// tlv length check from standard
			if (tlvLen != 14) {
				free(ret);
				return NULL;
			}
			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->addr, 6);
			_EnB(&p, &ret->bssid, 6);
			_E1B(&p, &ret->op_class);
			_E1B(&p, &ret->channel);
			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_BACKHAUL_STEERING_RESPONSE:
		{
			// tlv structure pointer
			struct tlv_backhaul_steer_resp *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			// tlv length check from standard
			if (tlvLen != 13) {
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->addr, 6);
			_EnB(&p, &ret->bssid, 6);
			_E1B(&p, &ret->res_code);
			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_HIGHER_LAYER_DATA:
		{
			// tlv structure pointer
			struct tlv_higher_layer_data *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E1B(&p, &ret->protocol);

			ret->payload = malloc(sizeof(*ret->payload) *
					      (tlvLen - 1));
			if (!ret->payload) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			memset(&ret->payload, 0x0, tlvLen - 1);
			_EnB(&p, &ret->payload, tlvLen - 2);
			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_ASSOCIATED_STA_TRAFFIC_STATS:
		{
			// tlv structure pointer
			struct tlv_assoc_sta_traffic_stats *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->addr, 6);
			_E4B(&p, &ret->bytes_sent);
			_E4B(&p, &ret->bytes_received);
			_E4B(&p, &ret->packets_sent);
			_E4B(&p, &ret->packets_received);
			_E4B(&p, &ret->tx_packets_err);
			_E4B(&p, &ret->rx_packets_err);
			_E4B(&p, &ret->retransmission_cnt);
			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_ERROR_CODE:
		{
			// tlv structure pointer
			struct tlv_error_code *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			// tlv length check from standard
			if (tlvLen != 7) {
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E1B(&p, &ret->reason_code);
			_EnB(&p, &ret->addr, 6);
			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_CHANNEL_SCAN_REPORTING_POLICY:
		{
			// tlv structure pointer
			struct tlv_ch_scan_rep_policy *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E1B(&p, &temp);
			ret->ch_scans = BIT(7, temp);
			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_CHANNEL_SCAN_CAPABILITY:
		{
			// tlv structure pointer
			struct tlv_ch_scan_cap *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E1B(&p, &ret->nbr_radios);

			if (ret->nbr_radios == 0)
				return (uint8_t *) ret;

			ret->radio =
				malloc(sizeof(*ret->radio) *
				       ret->nbr_radios);
			if (!ret->radio) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			for (i = 0; i < ret->nbr_radios; ++i) {

				_EnB(&p, &ret->radio[i].radio_id, 6);
				_E1B(&p, &temp);
				ret->radio[i].on_boot = BIT(7, temp);
				ret->radio[i].scan_impact = BITS(6, 5, temp);
				_E4B(&p, &ret->radio[i].min_scan_interval);
				_E1B(&p, &ret->radio[i].nbr_op_classes);

				if (ret->radio[i].nbr_op_classes == 0)
					continue;

				ret->radio[i].op_class =
					malloc(sizeof(*ret->radio[i].op_class) *
					       ret->radio[i].nbr_op_classes);
				if (!ret->radio[i].op_class) {
					free(ret->radio);
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}

				for (j = 0; j < ret->radio[i].nbr_op_classes; ++j) {
					_E1B(&p, &ret->radio[i].op_class[j].op_class);
					_E1B(&p, &ret->radio[i].op_class[j].nbr_ch);

					if (ret->radio[i].op_class[j].nbr_ch == 0)
						continue;

					ret->radio[i].op_class[j].ch =
						malloc(sizeof(*ret->radio[i].op_class[j].ch) *
						       ret->radio[i].op_class[j].nbr_ch);

					if (!ret->radio[i].op_class[j].ch) {
						free(ret->radio);
						free(ret);
						PRINT_ERR("Insufficient memory!");
						return NULL;
					}

					_EnB(&p, ret->radio[i].op_class[j].ch, ret->radio[i].op_class[j].nbr_ch);
				}
			}

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_CHANNEL_SCAN_REQ:
		{
			// tlv structure pointer
			struct tlv_ch_scan_req *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E1B(&p, &ret->fresh_scan);
			_E1B(&p, &ret->nbr_radios);

			if (ret->nbr_radios == 0)
				return (uint8_t *) ret;

			ret->radio_data =
				malloc(sizeof(*ret->radio_data) *
				       ret->nbr_radios);
			if (!ret->radio_data) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			for (i = 0; i < ret->nbr_radios; ++i) {
				_EnB(&p, &ret->radio_data[i].radio_id, 6);
				_E1B(&p, &ret->radio_data[i].nbr_op_class);
				if (ret->radio_data[i].nbr_op_class == 0)
					continue;

				ret->radio_data[i].op_class_data =
					malloc(sizeof(*ret->radio_data[i].op_class_data) *
					       ret->radio_data[i].nbr_op_class);
				if (!ret->radio_data[i].op_class_data) {
					free(ret->radio_data);
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}

				for (j = 0; j < ret->radio_data[i].nbr_op_class; ++j) {
					_E1B(&p, &ret->radio_data[i].op_class_data[j].op_class);
					_E1B(&p, &ret->radio_data[i].op_class_data[j].nbr_ch);

					if (ret->radio_data[i].op_class_data[j].nbr_ch == 0)
						continue;

					ret->radio_data[i].op_class_data[j].ch =
						malloc(sizeof(*ret->radio_data[i].op_class_data[j].ch) *
						       ret->radio_data[i].op_class_data[j].nbr_ch);
					if (!ret->radio_data[i].op_class_data[j].ch) {
						free(ret->radio_data);
						free(ret);
						PRINT_ERR("Insufficient memory!");
						return NULL;
					}

					_EnB(&p, ret->radio_data[i].op_class_data[j].ch,
					     ret->radio_data[i].op_class_data[j].nbr_ch);
				}
			}

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_CHANNEL_SCAN_RES:
		{
			// tlv structure pointer
			struct tlv_ch_scan_res *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->radio_id, 6);
			_E1B(&p, &ret->op_class);
			_E1B(&p, &ret->ch);
			_E1B(&p, &ret->scan_res);

			if (ret->scan_res != CH_SCAN_STATUS_SUCCESS)
				return (uint8_t *) ret;

			_E1B(&p, &ret->time_len);

			if (ret->time_len) {
				ret->timestamp = malloc(sizeof(*ret->timestamp) *
							ret->time_len);
				if (!ret->timestamp) {
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}
				_EnB(&p, ret->timestamp, ret->time_len);
			}

			_E1B(&p, &ret->utilization);
			_E1B(&p, &ret->noise);
			_E2B(&p, &ret->nbr_neighbors);

			if (ret->nbr_neighbors) {

				ret->neighbor_data =
					malloc(sizeof(*ret->neighbor_data) *
					       ret->nbr_neighbors);
				if (!ret->neighbor_data) {
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}

				for (i = 0; i < ret->nbr_neighbors; ++i) {
					size_t len = 0;

					_EnB(&p, &ret->neighbor_data[i].bssid, 6);

					_E1B(&p, &ret->neighbor_data[i].ssid_len);
					len = ret->neighbor_data[i].ssid_len;

					ret->neighbor_data[i].ssid =
						malloc(sizeof(*ret->neighbor_data[i].ssid) *
						       (len+1));
					if (!ret->neighbor_data[i].ssid) {
						free(ret->neighbor_data);
						free(ret);
						PRINT_ERR("Insufficient memory!");
						return NULL;
					}

					_EnB(&p, ret->neighbor_data[i].ssid, len);
					ret->neighbor_data[i].ssid[len] = 0x0;

					_E1B(&p, &ret->neighbor_data[i].signal_strength);
					_E1B(&p, &ret->neighbor_data[i].ch_band_len);
					len = ret->neighbor_data[i].ch_band_len;

					ret->neighbor_data[i].ch_bandwidth =
						malloc(sizeof(*ret->neighbor_data[i].ch_bandwidth) *
						       (len+1));
					if (!ret->neighbor_data[i].ch_bandwidth) {
						free(ret->neighbor_data);
						free(ret);
						PRINT_ERR("Insufficient memory!");
						return NULL;
					}

					_EnB(&p, ret->neighbor_data[i].ch_bandwidth, len);
					ret->neighbor_data[i].ch_bandwidth[len] = 0x0;

					_E1B(&p, &ret->neighbor_data[i].bss_load_elm_present);
					if (ret->neighbor_data[i].bss_load_elm_present) {
						_E1B(&p, &ret->neighbor_data[i].ch_util);
						_E2B(&p, &ret->neighbor_data[i].sta_count);
					}
				}
			}

			_E4B(&p, &ret->agg_scan_duration);

			_E1B(&p, &temp);
			ret->scan_type = BIT(7, temp);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_TIMESTAMP:
		{
			// tlv structure pointer
			struct tlv_timestamp *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E1B(&p, &ret->time_len);

			if (ret->time_len == 0) {
				PRINT_ERR("Invalid time len!");
				free(ret);
				return NULL;
			}

			ret->time = malloc(sizeof(*ret->time) *
					   (ret->time_len+1));
			if (!ret->time) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			_EnB(&p, ret->time, ret->time_len);
			ret->time[ret->time_len] = 0x0;

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_CAC_REQ:
		{
			// tlv structure pointer
			struct tlv_cac_req *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;

			_E1B(&p, &ret->nbr_radios);
			if (ret->nbr_radios == 0)
				return (uint8_t *) ret;

			ret->radio_data = malloc(sizeof(*ret->radio_data) *
						 ret->nbr_radios);
			if (!ret->radio_data) {
				PRINT_ERR("Insufficient memory!");
				free(ret);
				return NULL;
			}

			for (i = 0; i < ret->nbr_radios; ++i) {
				_EnB(&p, &ret->radio_data[i].radio_id, 6);
				_E1B(&p, &ret->radio_data[i].op_class);
				_E1B(&p, &ret->radio_data[i].ch);

				_E1B(&p, &temp);
				ret->radio_data[i].cac_method = BITS(7, 5, temp);
				ret->radio_data[i].cac_comp_action = BITS(4, 3, temp);
			}

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_CAC_TERMINATION:
		{
			// tlv structure pointer
			struct tlv_cac_term *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E1B(&p, &ret->nbr_radio);

			if (ret->nbr_radio == 0)
				return (uint8_t *) ret;

			ret->radio_data = malloc(sizeof(*ret->radio_data) *
						 ret->nbr_radio);
			if (!ret->radio_data) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			for (i = 0; i < ret->nbr_radio; ++i) {
				_EnB(&p, &ret->radio_data[i].radio_id, 6);
				_E1B(&p, &ret->radio_data[i].op_class);
				_E1B(&p, &ret->radio_data[i].ch);
			}

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_CAC_COMPLETION_REPORT:
		{
			// tlv structure pointer
			struct tlv_cac_comp_report *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory !");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv !");
				free(ret);
				return NULL;
			}

			ret->tlv_len = tlvLen;
			ret->tlv_type = *packet_stream;

			_E1B(&p, &ret->nbr_radios);
			if (ret->nbr_radios == 0)
				return (uint8_t *) ret;

			ret->radio_data =
				malloc(sizeof(*ret->radio_data) *
				       ret->nbr_radios);
			if (!ret->radio_data) {
				PRINT_ERR("Insufficient memory");
				free(ret);
				return NULL;
			}

			for (i = 0; i < ret->nbr_radios; ++i) {
				_EnB(&p, &ret->radio_data[i].radio_id, 6);
				_E1B(&p, &ret->radio_data[i].op_class);
				_E1B(&p, &ret->radio_data[i].ch);
				_E1B(&p, &ret->radio_data[i].completion_status);
				_E1B(&p, &ret->radio_data[i].nbr_pairs);

				if (ret->radio_data[i].nbr_pairs == 0)
					continue;

				ret->radio_data[i].pair_data =
					malloc(sizeof(*ret->radio_data[i].pair_data) *
					       ret->radio_data[i].nbr_pairs);
				if (!ret->radio_data[i].pair_data) {
					free(ret->radio_data);
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}

				for (j = 0; j < ret->radio_data[i].nbr_pairs; ++j) {
					_E1B(&p, &ret->radio_data[i].pair_data[j].op_class_detected);
					_E1B(&p, &ret->radio_data[i].pair_data[j].ch_detected);
				}
			}

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_CAC_STATUS_REPORT:
		{
			// tlv structure pointer
			struct tlv_cac_status_report *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			size_t nbr = 0;

			_E1B(&p, &ret->nbr_available_ch);
			nbr = ret->nbr_available_ch;

			if (nbr) {
				ret->ch_data =
					malloc(sizeof(*ret->ch_data) *
					       nbr);
				if (!ret->ch_data) {
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}

				for (i = 0; i < nbr; ++i) {
					_E1B(&p, &ret->ch_data[i].op_class);
					_E1B(&p, &ret->ch_data[i].ch);
					_E2B(&p, &ret->ch_data[i].time);
				}
			}

			_E1B(&p, &ret->nbr_pairs_duration);
			nbr = ret->nbr_pairs_duration;

			if (nbr) {
				ret->duration_pair_data =
					malloc(sizeof(*ret->duration_pair_data) *
					       nbr);
				if (!ret->duration_pair_data) {
					free(ret->ch_data);
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}

				for (i = 0; i < nbr; ++i) {
					_E1B(&p, &ret->duration_pair_data[i].op_class);
					_E1B(&p, &ret->duration_pair_data[i].ch);
					_E2B(&p, &ret->duration_pair_data[i].time);
				}
			}

			_E1B(&p, &ret->nbr_pairs_coundown);
			nbr = ret->nbr_pairs_coundown;
			if (nbr) {
				ret->count_pair_data =
					malloc(sizeof(*ret->count_pair_data) *
					       nbr);
				if (!ret->count_pair_data) {
					free(ret->duration_pair_data);
					free(ret->ch_data);
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}

				for (i = 0; i < nbr; ++i) {
					_E1B(&p, &ret->count_pair_data[i].op_class);
					_E1B(&p, &ret->count_pair_data[i].ch);
					_E2B(&p, &ret->count_pair_data[i].time);
				}
			}

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_CAC_CAPABILITY:
		{
			// tlv structure pointer
			struct tlv_cac_cap *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, ret->country_code, 2);
			_E1B(&p, &ret->nbr_radios);

			if (ret->nbr_radios == 0)
				return (uint8_t *) ret;

			ret->radio_data = malloc(sizeof(*ret->radio_data) *
						 ret->nbr_radios);
			if (!ret->radio_data) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			for (i = 0; i < ret->nbr_radios; ++i) {
				typeof(*ret->radio_data) *cp = &ret->radio_data[i];

				_EnB(&p, &cp->radio_id, 6);
				_E1B(&p, &cp->nbr_cac_types);
				if (cp->nbr_cac_types == 0)
					continue;

				cp->cac_data =
					malloc(sizeof(*cp->cac_data) *
					       cp->nbr_cac_types);
				if (!cp->cac_data) {
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}


				for (j = 0; j < cp->nbr_cac_types; ++j) {
					typeof(*cp->cac_data) *dp = &cp->cac_data[j];

					_E1B(&p, &dp->cac_method_supported);

					dp->duration = 0;
					_EnB(&p, &dp->duration, 3);
					_E1B(&p, &dp->nbr_op_classes);

					if (dp->nbr_op_classes == 0)
						continue;

					dp->op_class_data =
						malloc(sizeof(*dp->op_class_data) *
						       dp->nbr_op_classes);
					if (!dp->op_class_data) {
						free(ret);
						PRINT_ERR("Insufficient memory!");
						return NULL;
					}

					for (k = 0; k < dp->nbr_op_classes; ++k) {
						typeof(*dp->op_class_data) *op = &dp->op_class_data[k];

						_E1B(&p, &op->op_class);
						_E1B(&p, &op->nbr_ch);
						if (op->nbr_ch == 0)
							continue;

						op->ch =malloc(sizeof(*op->ch) *
							       op->nbr_ch);
						if (!op->ch) {
							free(ret);
							PRINT_ERR("Insufficient memory!");
							return NULL;
						}

						_EnB(&p, op->ch, op->nbr_ch);
					}
				}
			}

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_MULTIAP_PROFILE:
		{
			// tlv structure pointer
			struct tlv_map_profile *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			// tlv length check from standard
			if (tlvLen != 1) {
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E1B(&p, &ret->profile);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_PROFILE2_AP_CAPABILITY:
		{
			// tlv structure pointer
			struct tlv_profile2_ap_cap *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E2B(&p, &ret->reserved);
			_E1B(&p, &temp);
			ret->byte_counter = BITS(7, 6, temp);
			_E1B(&p, &ret->max_vids);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_DEFAULT_8021Q_SETTINGS:
		{
			// tlv structure pointer
			struct tlv_default_8021q_settings *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E2B(&p, &ret->primary_vid);

			_E1B(&p, &temp);
			ret->pcp = BITS(7, 5, temp);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_TRAFFIC_SEPARATION_POLICY:
		{
			// tlv structure pointer
			struct tlv_traffic_sep_policy *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E1B(&p, &ret->nbr_ssid);

			if(ret->nbr_ssid == 0)
				return (uint8_t *) ret;

			ret->data = malloc(sizeof(*ret->data) *
					   ret->nbr_ssid);
			if (!ret->data) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}


			for (i = 0; i < ret->nbr_ssid; ++i) {
				size_t len = 0;

				_E1B(&p, &ret->data[i].ssid_len);
				len = ret->data[i].ssid_len;

				ret->data[i].ssid = malloc(sizeof(*ret->data[i].ssid) *
							   (len+1));
				if (!ret->data[i].ssid) {
					for (size_t ii = 0; ii < i; ii++) {
						free(ret->data[ii].ssid);
					}
					free(ret->data);
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}


				_EnB(&p, ret->data[i].ssid, len);
				ret->data[i].ssid[len] = 0x0;

				_E2B(&p, &ret->data[i].vid);
			}

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_PROFILE2_ERR_CODE:
		{
			// tlv structure pointer
			struct tlv_profile2_err_code *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E1B(&p, &ret->reason_code);

			//Here the bssid is only included when the reason code 
			// is set to '0x07' or '0x08'.
			if(ret->reason_code == 0x07 || ret->reason_code == 0x08)
				_EnB(&p, &ret->bssid, 6);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_AP_RADIO_ADV_CAPABILITY:
		{
			// tlv structure pointer
			struct tlv_ap_radio_adv_cap *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->radio_id, 6);

			_E1B(&p, &temp);
			ret->combined_front_back = BIT(7, temp);
			ret->combined_p1_p2 = BIT(6, temp);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_ASSOCIATION_STATUS_NOTIF:
		{
			// tlv structure pointer
			struct tlv_assoc_status_notif *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E1B(&p, &ret->nbr_bssid);
			if(ret->nbr_bssid == 0)
				return (uint8_t *) ret;

			ret->data = malloc(sizeof(*ret->data) *
					   ret->nbr_bssid);
			if (!ret->data) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}


			for (i = 0; i < ret->nbr_bssid; ++i) {
				_EnB(&p, &ret->data[i].bssid, 6);
				_E1B(&p, &ret->data[i].association_allowance);
			}

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_SOURCE_INFO:
		{
			// tlv structure pointer
			struct tlv_source_info *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			// tlv length check from standard
			if (tlvLen != 6) {
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->mac, 6);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_TUNNELED_MSG_TYPE:
		{
			// tlv structure pointer
			struct tlv_tunnel_msg_type *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			// tlv length check from standard
			if (tlvLen != 1) {
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E1B(&p, &ret->tunnel_protocol_type);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_TUNNELED:
		{
			// tlv structure pointer
			struct tlv_tunneled *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			ret->frame_body = malloc(sizeof(*ret->frame_body) *
						 tlvLen);
			if (!ret->frame_body) {
				free(ret);
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			_EnB(&p, ret->frame_body, tlvLen);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_PROFILE2_STEERING_REQ:
		{
			// tlv structure pointer
			struct tlv_profile2_steer_req *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->bssid, 6);
			_E1B(&p, &ret->req_mode);

			_E1B(&p, &temp);

			ret->req_mode = BIT(7, temp);
			ret->btm_imminent_bit = BIT(6, temp);
			ret->btm_abridged_bit = BIT(5, temp);

			_E2B(&p, &ret->steering_opt_win);
			_E2B(&p, &ret->btm_dis_timer);
			_E1B(&p, &ret->sta_list_count);

			//Here the sta_mac field is repeated k – 1 times
			// (if k = 0, the sta_mac field is omitted). 
                        if(ret->sta_list_count > 0)
			{
				ret->sta_mac = malloc(sizeof(*ret->sta_mac) *
					      ret->sta_list_count);
				if (!ret->sta_mac) {
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}

				for (i = 0; i < ret->sta_list_count; ++i)
					_EnB(&p, &ret->sta_mac[i], 6);
			}
			_E1B(&p, &ret->target_bssid_count);

			//Here the below fileds are repeated m – 1 times 
			// (if m = 0, these fields are omitted).
			if(ret->target_bssid_count > 0)
			{
				ret->target_data = malloc(sizeof(*ret->target_data) *
						  ret->target_bssid_count);
				if (!ret->target_data) {
					free(ret->sta_mac);
					free(ret);
					PRINT_ERR("Insufficient memory!");
					return NULL;
				}
			

				for (i = 0; i < ret->target_bssid_count; ++i) {
					_EnB(&p, &ret->target_data[i].bssid, 6);
					_E1B(&p, &ret->target_data[i].bss_op_class);
					_E1B(&p, &ret->target_data[i].bss_ch);
					_E1B(&p, &ret->target_data[i].reason_code);
				}
			}
			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_UNSUCCESS_ASSOCIATION_POLICY:
		{
			// tlv structure pointer
			struct tlv_unsuccess_assoc_policy *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure

			_E1B(&p, &temp);
			ret->report = BIT(7, temp);
			_E4B(&p, &ret->max_reporting_rate);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_METRIC_COLLECTION_INTERVAL:
		{
			// tlv structure pointer
			struct tlv_metric_collection_interval *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E4B(&p, &ret->collection_interval);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_RADIO_METRICS:
		{
			// tlv structure pointer
			struct tlv_radio_metrics *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->radio_id, 6);
			_E1B(&p, &ret->noise);
			_E1B(&p, &ret->transmit);
			_E1B(&p, &ret->receive_self);
			_E1B(&p, &ret->receive_other);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_AP_EXTENDED_METRICS:
		{
			// tlv structure pointer
			struct tlv_ap_ext_metrics *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->bssid, 6);
			_E4B(&p, &ret->uni_bytes_sent);
			_E4B(&p, &ret->uni_bytes_recv);
			_E4B(&p, &ret->multi_bytes_sent);
			_E4B(&p, &ret->multi_bytes_recv);
			_E4B(&p, &ret->bro_bytes_sent);
			_E4B(&p, &ret->bro_bytes_recv);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_ASSOCIATED_STA_EXT_LINK_METRICS:
		{
			// tlv structure pointer
			struct tlv_assoc_sta_ext_link_metric *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->mac, 6);
			_E1B(&p, &ret->nbr_bssid);

			if(ret->nbr_bssid == 0)
				return (uint8_t *) ret;

			ret->data = malloc(ret->nbr_bssid *
					   sizeof(*ret->data));
			if (!ret->data) {
				PRINT_ERR("Insufficient memory!");
				free(ret);
				return NULL;
			}

			for (i = 0; i < ret->nbr_bssid; ++i) {
				_EnB(&p, &ret->data[i].bssid, 6);
				_E4B(&p, &ret->data[i].last_data_dwn_rate);
				_E4B(&p, &ret->data[i].last_data_up_rate);
				_E4B(&p, &ret->data[i].uti_recv);
				_E4B(&p, &ret->data[i].uti_transmit);
			}

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_STATUS_CODE:
		{
			// tlv structure pointer
			struct tlv_status_code *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E2B(&p, &ret->status_code);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_REASON_CODE:
		{
			// tlv structure pointer
			struct tlv_reason_code *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_E2B(&p, &ret->reason_code);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_BACKHAUL_STA_RADIO_CAPABILITY:
		{
			// tlv structure pointer
			struct tlv_backhaul_sta_radio_cap *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->radio_id, 6);

			_E1B(&p, &temp);
			ret->mac_include = BIT(7, temp);

			if (ret->mac_include)
				_EnB(&p, &ret->mac, 6);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	case MAP_TLV_BACKHAUL_BSS_CONFIG:
		{
			// tlv structure pointer
			struct tlv_backhaul_bss_config *ret;

			// allocate memory to tlv structure
			ret = malloc(sizeof(*ret));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			if (tlvLen == 0) {
				PRINT_ERR("Empty tlv");
				free(ret);
				return NULL;
			}

			ret->tlv_type = *packet_stream;
			ret->tlv_len = tlvLen;
			// Parse stream and fill the structure
			_EnB(&p, &ret->bssid, 6);

			_E1B(&p, &temp);
			ret->p1 = BIT(7, temp);
			ret->p2 = BIT(6, temp);

			// Return the tlv structure
			return (uint8_t *) ret;
		}
	default:
		{
			PRINT_ERR("Unknown TLV|0x%02x|", *packet_stream);
		}

	}

	return NULL;
}

uint8_t *map_put_tlv_cstruct(uint8_t *memory_structure, uint16_t *len)
{
	size_t i, j, k;
	uint8_t *ret, *p;
	uint8_t temp = 0;
	uint16_t tlv_length = 0;

	if (len)
		*len = 0;

	if (memory_structure == NULL)
		return NULL;

	switch (*memory_structure) {
		/* 1905 related tlvs */
	case TLV_TYPE_END_OF_MESSAGE:
		{
			// "IEEE Std 1905.1-2013 Section 6.4.1"

			struct tlv_eom *m;

			m = (struct tlv_eom *)memory_structure;

			tlv_length = 0;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}

	case TLV_TYPE_AL_MAC_ADDRESS_TYPE:
		{
			// "IEEE Std 1905.1-2013 Section 6.4.3"
			struct tlv_al_mac *m;

			m = (struct tlv_al_mac *)memory_structure;

			tlv_length = 6;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_InB(m->al_mac_address, &p, 6);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}

	case TLV_TYPE_MAC_ADDRESS_TYPE:
		{
			// "IEEE Std 1905.1-2013 Section 6.4.4"
			struct tlv_mac *m;

			m = (struct tlv_mac *)memory_structure;

			tlv_length = 6;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_InB(m->mac_address, &p, 6);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}

	case TLV_TYPE_DEVICE_INFORMATION_TYPE:
		{
			// "IEEE Std 1905.1-2013 Section 6.4.5"

			struct tlv_device_info *m;

			m = (struct tlv_device_info *)memory_structure;

			tlv_length = 7;	// AL MAC (6 bytes) + number of ifaces (1 bytes)
			for (i = 0; i < m->local_interfaces_nr; i++) {
				tlv_length += 6 + 2 + 1;	// MAC (6 bytes) + media type (2
				// bytes) + number of octets (1 byte)

				tlv_length += m->local_interfaces[i].media_specific_data_size;
			}

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_InB(m->al_mac_address, &p, 6);
			_I1B(&m->local_interfaces_nr, &p);

			for (i = 0; i < m->local_interfaces_nr; i++) {
				_InB(m->local_interfaces[i].mac_address, &p, 6);
				_I2B(&m->local_interfaces[i].media_type, &p);
				_I1B(&m->local_interfaces[i].media_specific_data_size, &p);

				if ((m->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11B_2_4_GHZ) ||
				    (m->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11G_2_4_GHZ) ||
				    (m->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11A_5_GHZ) ||
				    (m->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11N_2_4_GHZ) ||
				    (m->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11N_5_GHZ) ||
				    (m->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11AC_5_GHZ) ||
				    (m->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11AD_60_GHZ) ||
				    (m->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11AF_GHZ)
				   ) {
					if (m->local_interfaces[i].media_specific_data_size != 10) {
						// Malformed structure
						free(ret);
						return NULL;
					}

					_InB(m->local_interfaces[i].media_specific_data.ieee80211.network_membership,
					     &p, 6);

					temp = m->local_interfaces[i].media_specific_data.ieee80211.role << 4;
					_I1B(&temp, &p);
					_I1B(&m->local_interfaces[i].media_specific_data.ieee80211.ap_channel_band, &p);

					_I1B(&m->local_interfaces[i].media_specific_data.ieee80211.ap_ch_freq_1,
					     &p);

					_I1B(&m->local_interfaces[i].media_specific_data.ieee80211.ap_ch_freq_2, &p);

				} else if ((m->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_1901_WAVELET) ||
					   (m->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_1901_FFT)
					  ) {
					if (m->local_interfaces[i].media_specific_data_size != 7) {
						// Malformed structure
						//
						free(ret);
						return NULL;
					}
					_InB(m->local_interfaces[i].media_specific_data.ieee1901.network_identifier,
					     &p, 7);
				} else {
					if (m->local_interfaces[i].media_specific_data_size != 0) {
						// Malformed structure
						//
						free(ret);
						return NULL;
					}
				}
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}

	case TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES:
		{
			// "IEEE Std 1905.1-2013 Section 6.4.6"

			struct tlv_device_bridge_cap *m;

			m = (struct tlv_device_bridge_cap *)memory_structure;

			tlv_length = 1;	// number of bridging tuples (1 bytes)
			for (i = 0; i < m->bridging_tuples_nr; i++) {
				tlv_length += 1;	// number of MAC addresses (1 bytes)
				tlv_length += 6 * m->bridging_tuples[i].bridging_tuple_macs_nr;
			}

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_I1B(&m->bridging_tuples_nr, &p);

			for (i = 0; i < m->bridging_tuples_nr; i++) {
				_I1B(&m->bridging_tuples[i].bridging_tuple_macs_nr, &p);

				for (j = 0; j < m->bridging_tuples[i].bridging_tuple_macs_nr; j++)
					_InB(m->bridging_tuples[i].bridging_tuple_macs[j].mac_address, &p, 6);
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}

	case TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST:
		{
			// "IEEE Std 1905.1-2013 Section 6.4.8"

			struct tlv_non1905_neighbor *m;

			m = (struct tlv_non1905_neighbor *)memory_structure;

			tlv_length = 6 + 6 * m->non_1905_neighbors_nr;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_InB(m->local_mac_address, &p, 6);

			for (i = 0; i < m->non_1905_neighbors_nr; i++)
				_InB(m->non_1905_neighbors[i].mac_address, &p, 6);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}

	case TLV_TYPE_NEIGHBOR_DEVICE_LIST:
		{
			// "IEEE Std 1905.1-2013 Section 6.4.9"

			struct tlv_neighbor_device *m;

			m = (struct tlv_neighbor_device *)memory_structure;

			tlv_length = 6 + 7 * m->neighbors_nr;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_InB(m->local_mac_address, &p, 6);

			for (i = 0; i < m->neighbors_nr; i++) {
				_InB(m->neighbors[i].mac_address, &p, 6);

				if (m->neighbors[i].bridge_flag == 1) {
					temp = 1 << 7;
					_I1B(&temp, &p);
				} else {
					temp = 0;
					_I1B(&temp, &p);
				}
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}

	case TLV_TYPE_TRANSMITTER_LINK_METRIC:
		{
			// "IEEE Std 1905.1-2013 Section 6.4.11"
			struct tlv_tx_link_metric *m;

			m = (struct tlv_tx_link_metric *)memory_structure;

			tlv_length = 12 + 29 * m->transmitter_link_metrics_nr;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_InB(m->local_al_address, &p, 6);
			_InB(m->neighbor_al_address, &p, 6);

			for (i = 0; i < m->transmitter_link_metrics_nr; i++) {
				_InB(m->transmitter_link_metrics[i].local_interface_address, &p, 6);
				_InB(m->transmitter_link_metrics[i].neighbor_interface_address, &p, 6);
				_I2B(&m->transmitter_link_metrics[i].intf_type, &p);
				_I1B(&m->transmitter_link_metrics[i].bridge_flag, &p);
				_I4B(&m->transmitter_link_metrics[i].packet_errors, &p);
				_I4B(&m->transmitter_link_metrics[i].transmitted_packets, &p);
				_I2B(&m->transmitter_link_metrics[i].mac_throughput_capacity, &p);
				_I2B(&m->transmitter_link_metrics[i].link_availability, &p);
				_I2B(&m->transmitter_link_metrics[i].phy_rate, &p);
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}

	case TLV_TYPE_RECEIVER_LINK_METRIC:
		{
			// "IEEE Std 1905.1-2013 Section 6.4.12"

			struct tlv_rx_link_metric *m;

			m = (struct tlv_rx_link_metric *)memory_structure;

			tlv_length = 12 + 23 * m->receiver_link_metrics_nr;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory !");
				return NULL;
			}

			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_InB(m->local_al_address, &p, 6);
			_InB(m->neighbor_al_address, &p, 6);

			for (i = 0; i < m->receiver_link_metrics_nr; i++) {
				_InB(m->receiver_link_metrics[i].local_interface_address, &p, 6);
				_InB(m->receiver_link_metrics[i].neighbor_interface_address, &p, 6);
				_I2B(&m->receiver_link_metrics[i].intf_type, &p);
				_I4B(&m->receiver_link_metrics[i].packet_errors, &p);
				_I4B(&m->receiver_link_metrics[i].packets_received, &p);
				_I1B(&m->receiver_link_metrics[i].rssi, &p);
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}

	case TLV_TYPE_VENDOR_SPECIFIC:
		{
			// "IEEE Std 1905.1-2013 Section 6.4.13"
			struct tlv_vendor_specific *m;

			m = (struct tlv_vendor_specific *)memory_structure;

			tlv_length = 3 + 2 + m->m_nr;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory !");
				return NULL;
			}

			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_InB(m->vendor_oui, &p, 3);
			_I2B(&m->m_nr, &p);

			if (m->m_nr)
				_InB(m->m, &p, m->m_nr);

			if (len)
				*len = 1 + 2 + tlv_length;

			return (uint8_t *) ret;
		}

	case TLV_TYPE_SEARCHED_ROLE:
		{
			// "IEEE Std 1905.1-2013 Section 6.4.14"
			struct tlv_searched_role *m;

			m = (struct tlv_searched_role *)memory_structure;

			tlv_length = 1;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			if (m->role != IEEE80211_ROLE_REGISTRAR) {
				free(ret);
				return NULL;
			}

			_I1B(&m->role, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}

	case TLV_TYPE_AUTOCONFIG_FREQ_BAND:
		{
			// "IEEE Std 1905.1-2013 Section 6.4.14"

			struct tlv_autoconf_freq_band *m;

			m = (struct tlv_autoconf_freq_band *)memory_structure;

			tlv_length = 1;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			if ((m->freq_band != IEEE80211_FREQUENCY_BAND_2_4_GHZ) &&
			    (m->freq_band != IEEE80211_FREQUENCY_BAND_5_GHZ) &&
			    (m->freq_band != IEEE80211_FREQUENCY_BAND_60_GHZ)
			   ) {
				// Malformed structure
				free(ret);
				return NULL;
			}

			_I1B(&m->freq_band, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}

	case TLV_TYPE_SUPPORTED_ROLE:
		{
			// "IEEE Std 1905.1-2013 Section 6.4.16"

			struct tlv_supported_role *m;

			m = (struct tlv_supported_role *)memory_structure;

			tlv_length = 1;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory !");
				return NULL;
			}

			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			if (m->role != IEEE80211_ROLE_REGISTRAR) {
				// Malformed structure
				free(ret);
				return NULL;
			}

			_I1B(&m->role, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}

	case TLV_TYPE_SUPPORTED_FREQ_BAND:
		{
			// "IEEE Std 1905.1-2013 Section 6.4.17"

			struct tlv_supported_freq_band *m;

			m = (struct tlv_supported_freq_band *)memory_structure;

			tlv_length = 1;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			switch (m->freq_band) {
			case IEEE80211_FREQUENCY_BAND_2_4_GHZ:
			case IEEE80211_FREQUENCY_BAND_5_GHZ:
			case IEEE80211_FREQUENCY_BAND_60_GHZ:
				break;
			default:
				{
					// Malformed structure
					free(ret);
					return NULL;
				}
			}

			_I1B(&m->freq_band, &p);

			*len = tlv_length + 1 + 2;
			return (uint8_t *) ret;
		}

	case TLV_TYPE_WSC:
		{
			// "IEEE Std 1905.1-2013 Section 6.4.18"

			struct tlv_wsc *m;

			m = (struct tlv_wsc *)memory_structure;

			tlv_length = m->wsc_frame_size;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_InB(m->wsc_frame, &p, m->wsc_frame_size);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
		/* map related tlvs */
	case MAP_TLV_SUPPORTED_SERVICE:
		{
			// Multiap spec Section 17.2.1
			struct tlv_supp_service *m;

			m = (struct tlv_supp_service *)memory_structure;

			tlv_length = 1 + m->supported_services_list;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_I1B(&m->supported_services_list, &p);

			for (i = 0; i < m->supported_services_list; i++)
				_I1B(&m->supported_services[i].service, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_SEARCHED_SERVICE:
		{
			// Multiap spec Section 17.2.2
			struct tlv_searched_service *m;

			m = (struct tlv_searched_service *)memory_structure;
			tlv_length = 1 + m->searched_services_list;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_I1B(&m->searched_services_list, &p);
			for (i = 0; i < m->searched_services_list; ++i)
				_I1B(&m->service[i], &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_AP_RADIO_IDENTIFIER:
		{
			// Multiap spec Section 17.2.3
			struct tlv_ap_radio_identifier *m;

			m = (struct tlv_ap_radio_identifier *)memory_structure;
			tlv_length = 6;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->radio_id, &p, 6);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_AP_OPERATIONAL_BSS:
		{
			// Multiap spec Section 17.2.4
			struct tlv_ap_oper_bss *m;

			m = (struct tlv_ap_oper_bss *)memory_structure;

			tlv_length = 1;	// number of radios (1 bytes)
			for (i = 0; i < m->radios_nr; i++) {
				tlv_length += 6;	// radio_id (6 bytes)
				tlv_length += 1;	// number of bss (1 bytes)
				for (j = 0; j < m->radio[i].bss_nr; j++)
					tlv_length += 6 + 1 + m->radio[i].bss[j].ssid_len;	// ssid

			}

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_I1B(&m->radios_nr, &p);

			for (i = 0; i < m->radios_nr; i++) {
				_InB(&m->radio[i].radio_id, &p, 6);
				_I1B(&m->radio[i].bss_nr, &p);

				for (j = 0; j < m->radio[i].bss_nr; j++) {
					_InB(&m->radio[i].bss[j].bssid, &p, 6);
					_I1B(&m->radio[i].bss[j].ssid_len, &p);
					_InB(&m->radio[i].bss[j].ssid, &p, m->radio[i].bss[j].ssid_len);
				}
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}

	case MAP_TLV_ASSOCIATED_CLIENTS:
		{
			// Multiap spec Section 17.2.5
			struct tlv_assoc_client *m;

			m = (struct tlv_assoc_client *)memory_structure;

			tlv_length = 1;	// number of bsss (1 bytes)
			for (i = 0; i < m->bss_nr; i++) {
				tlv_length += 6;	// bssid (6 bytes)
				tlv_length += 2;	// number of clients (1 bytes)
				tlv_length += 6 * m->bss[i].assoc_clients_nr;	// client_addr
				tlv_length += 2 * m->bss[i].assoc_clients_nr;	// uptime
			}

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_I1B(&m->bss_nr, &p);

			for (i = 0; i < m->bss_nr; i++) {
				_InB(&m->bss[i].bssid, &p, 6);
				_I2B(&m->bss[i].assoc_clients_nr, &p);

				for (j = 0; j < m->bss[i].assoc_clients_nr; j++) {
					_InB(&m->bss[i].clients[j].client_addr, &p, 6);
					_I2B(&m->bss[i].clients[j].uptime, &p);
				}
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_AP_CAPABILITY:
		{
			// Multiap spec Section 17.2.6
			struct tlv_ap_cap *m;

			m = (struct tlv_ap_cap *)memory_structure;

			tlv_length = 1;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			MODIFY_BIT(7, temp, m->op_ch_metric_reporting);
			MODIFY_BIT(6, temp, m->non_op_ch_metric_reporting);
			MODIFY_BIT(5, temp, m->agent_init_rcpi_steering);
			_I1B(&temp, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_AP_RADIO_BASIC_CAPABILITIES:
		{
			// Multiap spec Section 17.2.7
			struct tlv_ap_radio_basic_cap *m;

			m = (struct tlv_ap_radio_basic_cap *)memory_structure;
			tlv_length = 6 + 1 + 1;

			for (i = 0; i < m->operating_classes_nr; ++i)
				tlv_length += 1 + 1 + 1 + m->operating_class[i].non_op_ch_nr;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->radio_id, &p, 6);
			_I1B(&m->max_bss_nr, &p);
			_I1B(&m->operating_classes_nr, &p);
			for (i = 0; i < m->operating_classes_nr; ++i) {
				_I1B(&m->operating_class[i].op_class, &p);
				_I1B(&m->operating_class[i].max_tx_power, &p);
				_I1B(&m->operating_class[i].non_op_ch_nr, &p);
				for (j = 0; j < m->operating_class[i].non_op_ch_nr; ++j)
					_I1B(&m->operating_class[i].channel[j], &p);
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_AP_HT_CAPABILITIES:
		{
			// Multiap spec Section 17.2.8
			struct tlv_ap_ht_cap *m;

			m = (struct tlv_ap_ht_cap *)memory_structure;
			tlv_length = 7;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->radio_id, &p, 6);


			MODIFY_BITS(7, 6, temp, m->max_tx_streams_supported);
			MODIFY_BITS(5, 4, temp, m->max_rx_streams_supported);
			MODIFY_BIT(3, temp, m->gi_20_support);
			MODIFY_BIT(2, temp, m->gi_40_support);
			MODIFY_BIT(1, temp, m->ht_40_support);
			_I1B(&temp, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_AP_VHT_CAPABILITIES:
		{
			// Multiap spec Section 17.2.9
			struct tlv_ap_vht_cap *m;

			m = (struct tlv_ap_vht_cap *)memory_structure;
			tlv_length = 12;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->radio_id, &p, 6);
			_I2B(&m->vht_tx_mcs_supported, &p);
			_I2B(&m->vht_rx_mcs_supported, &p);

			MODIFY_BITS(7, 5, temp, m->max_tx_streams_supported);
			MODIFY_BITS(4, 2, temp, m->max_rx_streams_supported);
			MODIFY_BIT(1, temp, m->gi_80_support);
			MODIFY_BIT(0, temp, m->gi_160_support);
			_I1B(&temp, &p);

			MODIFY_BIT(7, temp, m->vht_8080_support);
			MODIFY_BIT(6, temp, m->vht_160_support);
			MODIFY_BIT(5, temp, m->su_beamformer_capable);
			MODIFY_BIT(4, temp, m->mu_beamformer_capable);
			_I1B(&temp, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_AP_HE_CAPABILITIES:
		{
			// Multiap spec Section 17.2.10
			struct tlv_ap_he_cap *m;

			m = (struct tlv_ap_he_cap *)memory_structure;
			tlv_length = 6 + 1 + m->he_mcs_len + 1 + 1;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->radio_id, &p, 6);
			_I1B(&m->he_mcs_len, &p);
			for (i = 0; i < m->he_mcs_len; ++i)
				_I1B(&m->he_mcs_supported[i], &p);

			MODIFY_BITS(7, 5, temp, m->max_tx_streams_supported);
			MODIFY_BITS(4, 2, temp, m->max_rx_streams_supported);
			MODIFY_BIT(1, temp, m->he_8080_support);
			MODIFY_BIT(0, temp, m->he_160_support);
			_I1B(&temp, &p);

			temp = 0;
			MODIFY_BIT(7, temp, m->su_beamformer_capable);
			MODIFY_BIT(6, temp, m->mu_beamformer_capable);
			MODIFY_BIT(5, temp, m->ul_mu_mimo_capable);
			MODIFY_BIT(4, temp, m->ul_mu_mimo_ofdma_capable);
			MODIFY_BIT(3, temp, m->dl_mu_mimo_ofdma_capable);
			MODIFY_BIT(2, temp, m->ul_ofdma_capable);
			MODIFY_BIT(1, temp, m->dl_ofdma_capable);
			_I1B(&temp, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_STEERING_POLICY:
		{
			// Multiap spec Section 17.2.11
			struct tlv_steering_policy *m;

			m = (struct tlv_steering_policy *)memory_structure;
			tlv_length = 1 + m->local_disallowed_sta_nr * 6;
			tlv_length += 1 + m->btm_disallowed_sta_nr * 6;
			tlv_length += 1 + m->control_policy_radio_nr * (6 + 1 + 1 + 1);

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_I1B(&m->local_disallowed_sta_nr, &p);
			for (i = 0; i < m->local_disallowed_sta_nr; ++i)
				_InB(&m->local_disallowed_sta_macs[i].addr, &p, 6);

			_I1B(&m->btm_disallowed_sta_nr, &p);
			for (i = 0; i < m->btm_disallowed_sta_nr; ++i)
				_InB(&m->btm_disallowed_sta_macs[i].addr, &p, 6);

			_I1B(&m->control_policy_radio_nr, &p);
			for (i = 0; i < m->control_policy_radio_nr; ++i) {
				_InB(&m->control_policy[i].radio_id, &p, 6);
				_I1B(&m->control_policy[i].steering_policy, &p);
				_I1B(&m->control_policy[i].channel_utilization_thres, &p);
				_I1B(&m->control_policy[i].rcpi_steering_thres, &p);
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_METRIC_REPORTING_POLICY:
		{
			// Multiap spec Section 17.2.12
			struct tlv_metric_report_policy *m;

			m = (struct tlv_metric_report_policy *)memory_structure;
			tlv_length = 1 + 1 + m->metric_reporting_policy_radio_nr * (6 + 1 + 1 + 1 + 1);

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_I1B(&m->interval, &p);
			_I1B(&m->metric_reporting_policy_radio_nr, &p);

			for (i = 0; i < m->metric_reporting_policy_radio_nr; ++i) {
				_InB(&m->metric_reporting_policy[i].radio_id, &p, 6);
				_I1B(&m->metric_reporting_policy[i].rcpi_thres, &p);
				_I1B(&m->metric_reporting_policy[i].rcpi_hysteresis_margin, &p);
				_I1B(&m->metric_reporting_policy[i].channel_utilization_thres, &p);

				MODIFY_BIT(7, temp, m->metric_reporting_policy[i].is_assoc_sta_traffic_stats);
				MODIFY_BIT(6, temp, m->metric_reporting_policy[i].is_assoc_sta_link_metrics);
				_I1B(&temp, &p);
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_CHANNEL_PREFERENCE:
		{
			// Multiap spec Section 17.2.13
			struct tlv_channel_pref *m;

			m = (struct tlv_channel_pref *)memory_structure;
			tlv_length = 6 + 1;

			for (i = 0; i < m->ch_preference_op_class_nr; ++i)
				tlv_length += 1 + 1 + m->op_class[i].channel_nr + 1;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->radio_id, &p, 6);
			_I1B(&m->ch_preference_op_class_nr, &p);
			for (i = 0; i < m->ch_preference_op_class_nr; ++i) {
				_I1B(&m->op_class[i].op_class, &p);
				_I1B(&m->op_class[i].channel_nr, &p);
				for (j = 0; j < m->op_class[i].channel_nr; ++j)
					_I1B(&m->op_class[i].channel_list[j], &p);

				MODIFY_BITS(7, 4, temp, m->op_class[i].preference);
				MODIFY_BITS(3, 0, temp, m->op_class[i].preference_reason);
				_I1B(&temp, &p);
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_RADIO_OPERATION_RESTRICTION:
		{
			// Multiap spec Section 17.2.14
			struct tlv_radio_oper_restrict *m;

			m = (struct tlv_radio_oper_restrict *)memory_structure;
			tlv_length = 6 + 1;
			for (i = 0; i < m->op_restricted_op_class_nr; ++i)
				tlv_length += 1 + 1 + m->restricted_op_class[i].channel_nr * 2;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->radio_id, &p, 6);
			_I1B(&m->op_restricted_op_class_nr, &p);
			for (i = 0; i < m->op_restricted_op_class_nr; ++i) {
				_I1B(&m->restricted_op_class[i].op_class, &p);
				_I1B(&m->restricted_op_class[i].channel_nr, &p);
				for (j = 0; j < m->restricted_op_class[i].channel_nr; ++j) {
					_I1B(&m->restricted_op_class[i].restricted_channels[j].channel, &p);
					_I1B(&m->restricted_op_class[i].restricted_channels[j].min_freq_sep, &p);
				}
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_TRANSMIT_POWER_LIMIT:
		{
			// Multiap spec Section 17.2.15
			struct tlv_tx_power_limit *m;

			m = (struct tlv_tx_power_limit *)memory_structure;
			tlv_length = 7;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->radio_id, &p, 6);
			_I1B(&m->tx_power_limit, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_CHANNEL_SELECTION_RESPONSE:
		{
			// Multiap spec Section 17.2.16
			struct tlv_ch_selection_resp *m;

			m = (struct tlv_ch_selection_resp *)memory_structure;
			tlv_length = 7;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->radio_id, &p, 6);
			_I1B(&m->response_code, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_OPERATING_CHANNEL_REPORT:
		{
			// Multiap spec Section 17.2.17
			struct tlv_oper_ch_report *m;

			m = (struct tlv_oper_ch_report *)memory_structure;
			tlv_length = 6 + 1 + m->op_ch_op_class_nr * 2 + 1;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->radio_id, &p, 6);
			_I1B(&m->op_ch_op_class_nr, &p);
			for (i = 0; i < m->op_ch_op_class_nr; ++i) {
				_I1B(&m->op_ch_op_class[i].op_class, &p);
				_I1B(&m->op_ch_op_class[i].channel, &p);
			}
			_I1B(&m->curr_tx_power, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_CLIENT_INFO:
		{
			// Multiap spec Section 17.2.18
			struct tlv_client_info *m;

			m = (struct tlv_client_info *)memory_structure;
			tlv_length = 12;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->bssid, &p, 6);
			_InB(&m->client_addr, &p, 6);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_CLIENT_CAPABILITY_REPORT:
		{
			// Multiap spec Section 17.2.19
			struct tlv_client_cap_report *m;
			uint16_t body_len = 0;

			m = (struct tlv_client_cap_report *)memory_structure;

			tlv_length = 1;
			if(m->result_code == 0x00) {
				tlv_length += *(m->frame_body);
				body_len = *(m->frame_body);
			}

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_I1B(&m->result_code, &p);
			if(m->result_code == 0x00) {
				_InB(m->frame_body, &p, body_len);
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_CLIENT_ASSOCIATION_EVENT:
		{
			// Multiap spec Section 17.2.20
			struct tlv_client_assoc_event *m;

			m = (struct tlv_client_assoc_event *)memory_structure;

			tlv_length = 13;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;


			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_InB(&m->client_addr, &p, 6);
			_InB(&m->bssid, &p, 6);

			MODIFY_BIT(7, temp, m->assoc_event);
			_I1B(&temp, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_AP_METRIC_QUERY:
		{
			// Multiap spec Section 17.2.21
			struct tlv_ap_metric_query *m;

			m = (struct tlv_ap_metric_query *)memory_structure;
			tlv_length = 1 + m->bssid_nr * 6;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_I1B(&m->bssid_nr, &p);
			for (i = 0; i < m->bssid_nr; ++i)
				_InB(&m->ap_metric_query_bssid[i].bssid, &p, 6);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_AP_METRICS:
		{
			// Multiap spec Section 17.2.22
			struct tlv_ap_metrics *m;

			m = (struct tlv_ap_metrics *)memory_structure;
			tlv_length = 6 + 1 + 2 + 1 + 3;
			if (m->is_ac_bk)
				tlv_length += 3;

			if (m->is_ac_vo)
				tlv_length += 3;

			if (m->is_ac_vi)
				tlv_length += 3;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->bssid, &p, 6);
			_I1B(&m->channel_utilization, &p);
			_I2B(&m->total_assoc_stas, &p);

			MODIFY_BIT(7, temp, m->is_ac_be);
			MODIFY_BIT(6, temp, m->is_ac_bk);
			MODIFY_BIT(5, temp, m->is_ac_vo);
			MODIFY_BIT(4, temp, m->is_ac_vi);
			_I1B(&temp, &p);
			_InB(&m->service_param_info_be, &p, 3);

			if (m->is_ac_bk)
				_InB(&m->service_param_info_bk, &p, 3);

			if (m->is_ac_vo)
				_InB(&m->service_param_info_vo, &p, 3);

			if (m->is_ac_vi)
				_InB(&m->service_param_info_vi, &p, 3);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_STA_MAC_ADDRESS:
		{
			// Multiap spec Section 17.2.23
			struct tlv_sta_mac *m;

			m = (struct tlv_sta_mac *)memory_structure;
			tlv_length = 6;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->addr, &p, 6);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_ASSOCIATED_STA_LINK_METRICS:
		{
			// Multiap spec Section 17.2.24
			struct tlv_assoc_sta_link_metrics *m;

			m = (struct tlv_assoc_sta_link_metrics *)memory_structure;
			tlv_length = 6 + 1 + m->bssid_nr * (6 + 4 + 4 + 4 + 1);

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->addr, &p, 6);
			_I1B(&m->bssid_nr, &p);
			for (i = 0; i < m->bssid_nr; ++i) {
				_InB(&m->sta_link_metrics_bssid[i].bssid, &p, 6);
				_I4B(&m->sta_link_metrics_bssid[i].time_delta, &p);
				_I4B(&m->sta_link_metrics_bssid[i].dl_mac_data_rate, &p);
				_I4B(&m->sta_link_metrics_bssid[i].ul_mac_data_rate, &p);
				_I1B(&m->sta_link_metrics_bssid[i].ul_rcpi, &p);
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_UNASSOCIATED_STA_LINK_METRICS_QUERY:
		{
			// Multiap spec Section 17.2.25
			struct tlv_unassoc_sta_link_metrics_query *m;

			m = (struct tlv_unassoc_sta_link_metrics_query *)memory_structure;
			tlv_length = 1 + 1;
			for (i = 0; i < m->ch_nr; ++i)
				tlv_length += 1 + 1 + m->unassoc_sta_link_metrics_query_ch[i].sta_mac_nr * 6;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_I1B(&m->op_class, &p);
			_I1B(&m->ch_nr, &p);
			for (i = 0; i < m->ch_nr; ++i) {
				_I1B(&m->unassoc_sta_link_metrics_query_ch[i].channel, &p);
				_I1B(&m->unassoc_sta_link_metrics_query_ch[i].sta_mac_nr, &p);

				for (j = 0; j < m->unassoc_sta_link_metrics_query_ch[i].sta_mac_nr; ++j)
					_InB(&m->unassoc_sta_link_metrics_query_ch[i].sta_mac[j].addr, &p, 6);
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_UNASSOCIATED_STA_LINK_METRICS_RESPONSE:
		{
			// Multiap spec Section 17.2.26
			struct tlv_unassoc_sta_link_metrics_resp *m;

			m = (struct tlv_unassoc_sta_link_metrics_resp *)memory_structure;
			tlv_length = 1 + 1 + m->sta_nr * (6 + 1 + 4 + 1);

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_I1B(&m->op_class, &p);
			_I1B(&m->sta_nr, &p);
			for (i = 0; i < m->sta_nr; ++i) {
				_InB(&m->unassoc_sta_link_metrics_sta[i].addr, &p, 6);
				_I1B(&m->unassoc_sta_link_metrics_sta[i].channel, &p);
				_I4B(&m->unassoc_sta_link_metrics_sta[i].time_delta, &p);
				_I1B(&m->unassoc_sta_link_metrics_sta[i].ul_rcpi, &p);
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_BEACON_METRICS_QUERY:
		{
			// Multiap spec Section 17.2.27
			struct tlv_beacon_metrics_query *m;

			m = (struct tlv_beacon_metrics_query *)memory_structure;
			tlv_length = 6 + 1 + 1 + 6 + 1 + 1 + m->ssid_len + 1;
			for (i = 0; i < m->ch_reports_nr; ++i)
				tlv_length += 1 + 1 + m->ch_report[i].report_len - 1;

			tlv_length += 1 + m->elem_ids_nr;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->sta_addr, &p, 6);
			_I1B(&m->op_class, &p);
			_I1B(&m->channel, &p);
			_InB(&m->bssid, &p, 6);
			_I1B(&m->reporting_detail, &p);
			_I1B(&m->ssid_len, &p);
			_InB(&m->ssid, &p, m->ssid_len);

			_I1B(&m->ch_reports_nr, &p);
			for (i = 0; i < m->ch_reports_nr; ++i) {
				_I1B(&m->ch_report[i].report_len, &p);
				_I1B(&m->ch_report[i].op_class, &p);
				size_t chLen = (size_t) m->ch_report[i].report_len - 1;

				for (j = 0; j < chLen; ++j)
					_I1B(&m->ch_report[i].ch_list[j], &p);
			}
			_I1B(&m->elem_ids_nr, &p);
			for (i = 0; i < m->elem_ids_nr; ++i)
				_I1B(&m->elem_list[i], &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_BEACON_METRICS_RESPONSE:
		{
			// Multiap spec Section 17.2.28
			struct tlv_beacon_metrics_resp *m;

			m = (struct tlv_beacon_metrics_resp *)memory_structure;
			tlv_length = 6 + 1 + 1 + m->report_elems_nr;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->sta_addr, &p, 6);
			_I1B(&m->reserved, &p);
			_I1B(&m->report_elems_nr, &p);
			for (i = 0; i < m->report_elems_nr; ++i)
				_I1B(&m->report_elem[i], &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_STEERING_REQUEST:
		{
			// Multiap spec Section 17.2.29
			struct tlv_steer_Req *m;

			m = (struct tlv_steer_Req *)memory_structure;
			tlv_length = 6 + 1 + 2 + 2 + 1 + m->sta_list_cnt * 6;
			tlv_length += 1 + m->target_bssid_list_cnt * (6 + 1 + 1);

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->bssid, &p, 6);

			MODIFY_BIT(7, temp, m->request_mode);
			MODIFY_BIT(6, temp, m->btm_disassoc_imminent);
			MODIFY_BIT(5, temp, m->btm_abridged);
			_I1B(&temp, &p);

			_I2B(&m->steer_opp_window, &p);
			_I2B(&m->btm_disassoc_timer, &p);
			_I1B(&m->sta_list_cnt, &p);
			for (i = 0; i < m->sta_list_cnt; ++i)
				_InB(&m->steering_req_macs[i].addr, &p, 6);

			_I1B(&m->target_bssid_list_cnt, &p);
			for (i = 0; i < m->target_bssid_list_cnt; ++i) {
				_InB(&m->steering_req_target_bssids[i].bssid, &p, 6);
				_I1B(&m->steering_req_target_bssids[i].op_class, &p);
				_I1B(&m->steering_req_target_bssids[i].channel, &p);
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_STEERING_BTM_REPORT:
		{
			// Multiap spec Section 17.2.30
			struct tlv_steer_btm_report *m;

			m = (struct tlv_steer_btm_report *)memory_structure;
			tlv_length = 6 + 6 + 1;
			if (m->status_code == 0x00)
				tlv_length += 6;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->bssid, &p, 6);
			_InB(&m->sta_mac_addr, &p, 6);
			_I1B(&m->status_code, &p);
			// zero value means success
			if (m->status_code == 0x00)
				_InB(&m->target_bssid, &p, 6);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_CLIENT_ASSOCIATION_CONTROL_REQUEST:
		{
			// Multiap spec Section 17.2.31
			struct tlv_client_assoc_control_req *m;

			m = (struct tlv_client_assoc_control_req *)memory_structure;
			tlv_length = 6 + 1 + 2 + 1 + m->sta_list_cnt * 6;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->bssid, &p, 6);
			_I1B(&m->assoc_control, &p);
			_I2B(&m->validity_period, &p);
			_I1B(&m->sta_list_cnt, &p);
			for (i = 0; i < m->sta_list_cnt; ++i)
				_InB(&m->client_assoc_ctrl_req_stas[i].addr, &p, 6);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_BACKHAUL_STEERING_REQUEST:
		{
			// Multiap spec Section 17.2.32
			struct tlv_backhaul_steer_req *m;

			m = (struct tlv_backhaul_steer_req *)memory_structure;
			tlv_length = 14;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->addr, &p, 6);
			_InB(&m->bssid, &p, 6);
			_I1B(&m->op_class, &p);
			_I1B(&m->channel, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_BACKHAUL_STEERING_RESPONSE:
		{
			// Multiap spec Section 17.2.33
			struct tlv_backhaul_steer_resp *m;

			m = (struct tlv_backhaul_steer_resp *)memory_structure;
			tlv_length = 13;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->addr, &p, 6);
			_InB(&m->bssid, &p, 6);
			_I1B(&m->res_code, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_HIGHER_LAYER_DATA:
		{
			// Multiap spec Section 17.2.34
			struct tlv_higher_layer_data *m;

			m = (struct tlv_higher_layer_data *)memory_structure;
			tlv_length = 1;
			size_t payload_len = 0;

			while (m->payload[payload_len])
				payload_len++;

			tlv_length += payload_len;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_I1B(&m->protocol, &p);
			_InB(&m->payload, &p, payload_len);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_ASSOCIATED_STA_TRAFFIC_STATS:
		{
			// Multiap spec Section 17.2.35
			struct tlv_assoc_sta_traffic_stats *m;

			m = (struct tlv_assoc_sta_traffic_stats *)memory_structure;
			tlv_length = 6 + 4 + 4 + 4 + 4 + 4 + 4 + 4;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->addr, &p, 6);
			_I4B(&m->bytes_sent, &p);
			_I4B(&m->bytes_received, &p);
			_I4B(&m->packets_sent, &p);
			_I4B(&m->packets_received, &p);
			_I4B(&m->tx_packets_err, &p);
			_I4B(&m->rx_packets_err, &p);
			_I4B(&m->retransmission_cnt, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_ERROR_CODE:
		{
			// Multiap spec Section 17.2.36
			struct tlv_error_code *m;

			m = (struct tlv_error_code *)memory_structure;
			tlv_length = 7;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_I1B(&m->reason_code, &p);
			_InB(&m->addr, &p, 6);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_CHANNEL_SCAN_REPORTING_POLICY:
		{
			// tlv structure pointer
			// Multiap spec Section 17.2.37
			struct tlv_ch_scan_rep_policy *m;

			// calculate length of tlv structure
			m = (struct tlv_ch_scan_rep_policy *)memory_structure;
			tlv_length = 1;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			MODIFY_BIT(7, temp, m->ch_scans);
			_I1B(&temp, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_CHANNEL_SCAN_CAPABILITY:
		{
			// Multiap spec Section 17.2.38
			// tlv structure pointer
			struct tlv_ch_scan_cap *m;

			// calculate length of tlv structure
			m = (struct tlv_ch_scan_cap *)memory_structure;
			tlv_length = 1;
			for (i = 0; i < m->nbr_radios; ++i) {
				tlv_length += 6 + 1 + 4 + 1;
				for (j = 0; j < m->radio[i].nbr_op_classes; ++j)
					tlv_length += 1 + 1 + m->radio[i].op_class[j].nbr_ch;

			}

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);


			_I1B(&m->nbr_radios, &p);
			for (i = 0; i < m->nbr_radios; ++i) {
				_InB(&m->radio[i].radio_id, &p, 6);

				MODIFY_BIT(7, temp, m->radio[i].on_boot);
				MODIFY_BITS(6, 5, temp, m->radio[i].scan_impact);
				_I1B(&temp, &p);

				_I4B(&m->radio[i].min_scan_interval, &p);
				_I1B(&m->radio[i].nbr_op_classes, &p);

				for (j = 0; j < m->radio[i].nbr_op_classes; ++j) {
					_I1B(&m->radio[i].op_class[j].op_class, &p);
					_I1B(&m->radio[i].op_class[j].nbr_ch, &p);
					_InB(m->radio[i].op_class[j].ch, &p, m->radio[i].op_class[j].nbr_ch);
				}
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_CHANNEL_SCAN_REQ:
		{
			// Multiap spec Section 17.2.39
			// tlv structure pointer
			struct tlv_ch_scan_req *m;

			// calculate length of tlv structure
			m = (struct tlv_ch_scan_req *)memory_structure;
			tlv_length = 1 + 1;
			for (i = 0; i < m->nbr_radios; ++i) {
				tlv_length += 6 + 1;
				for (j = 0; j < m->radio_data[i].nbr_op_class; ++j)
					tlv_length += 1 + 1 + m->radio_data[i].op_class_data[j].nbr_ch;

			}

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_I1B(&m->fresh_scan, &p);
			_I1B(&m->nbr_radios, &p);

			for (i = 0; i < m->nbr_radios; ++i) {
				_InB(&m->radio_data[i].radio_id, &p, 6);
				_I1B(&m->radio_data[i].nbr_op_class, &p);
				for (j = 0; j < m->radio_data[i].nbr_op_class; ++j) {
					_I1B(&m->radio_data[i].op_class_data[j].op_class, &p);
					_I1B(&m->radio_data[i].op_class_data[j].nbr_ch, &p);
					_InB(m->radio_data[i].op_class_data[j].ch, &p,
					     m->radio_data[i].op_class_data[j].nbr_ch);
				}
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_CHANNEL_SCAN_RES:
		{
			// Multiap spec Section 17.2.40
			// tlv structure pointer
			struct tlv_ch_scan_res *m;

			// calculate length of tlv structure
			m = (struct tlv_ch_scan_res *)memory_structure;
			tlv_length = 6 + 1 + 1 + 1;

			if (m->scan_res == CH_SCAN_STATUS_SUCCESS) {
				tlv_length += 1;
				tlv_length += m->time_len + 1 + 1 + 2;
				for (i = 0; i < m->nbr_neighbors; ++i) {
					tlv_length += 6 + 1 + m->neighbor_data[i].ssid_len;
					tlv_length += 1 + 1 + m->neighbor_data[i].ch_band_len;
					tlv_length += 1;
					if (m->neighbor_data[i].bss_load_elm_present)
						tlv_length += 1 + 2;
				}
				tlv_length += 4 + 1;
			}

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->radio_id, &p, 6);
			_I1B(&m->op_class, &p);
			_I1B(&m->ch, &p);
			_I1B(&m->scan_res, &p);

			if (m->scan_res == CH_SCAN_STATUS_SUCCESS) {
				_I1B(&m->time_len, &p);
				_InB(m->timestamp, &p, m->time_len);

				_I1B(&m->utilization, &p);
				_I1B(&m->noise, &p);
				_I2B(&m->nbr_neighbors, &p);
				for (i = 0; i < m->nbr_neighbors; ++i) {
					size_t len = 0;

					_InB(&m->neighbor_data[i].bssid, &p, 6);

					_I1B(&m->neighbor_data[i].ssid_len, &p);
					len = m->neighbor_data[i].ssid_len;
					_InB(m->neighbor_data[i].ssid, &p, len);

					_I1B(&m->neighbor_data[i].signal_strength, &p);
					_I1B(&m->neighbor_data[i].ch_band_len, &p);
					len = m->neighbor_data[i].ch_band_len;
					_InB(m->neighbor_data[i].ch_bandwidth, &p, len);

					_I1B(&m->neighbor_data[i].bss_load_elm_present, &p);
					if (m->neighbor_data[i].bss_load_elm_present) {
						_I1B(&m->neighbor_data[i].ch_util, &p);
						_I2B(&m->neighbor_data[i].sta_count, &p);
					}
				}

				_I4B(&m->agg_scan_duration, &p);

				MODIFY_BIT(7, temp, m->scan_type);
				_I1B(&temp, &p);
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_TIMESTAMP:
		{
			// Multiap spec Section 17.2.41
			// tlv structure pointer
			struct tlv_timestamp *m;

			// calculate length of tlv structure
			m = (struct tlv_timestamp *)memory_structure;
			tlv_length = 1 + m->time_len;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_I1B(&m->time_len, &p);

			_InB(m->time, &p, m->time_len);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_CAC_REQ:
		{
			// Multiap spec Section 17.2.42
			// tlv structure pointer
			struct tlv_cac_req *m;

			// calculate length of tlv structure
			m = (struct tlv_cac_req *)memory_structure;
			tlv_length = 1;
			for (i = 0; i < m->nbr_radios; ++i)
				tlv_length += 6 + 1 + 1 + 1;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_I1B(&m->nbr_radios, &p);

			for (i = 0; i < m->nbr_radios; ++i) {
				_InB(&m->radio_data[i].radio_id, &p, 6);
				_I1B(&m->radio_data[i].op_class, &p);
				_I1B(&m->radio_data[i].ch, &p);

				MODIFY_BITS(7, 5, temp, m->radio_data[i].cac_method);
				MODIFY_BITS(4, 3, temp, m->radio_data[i].cac_comp_action);
				_I1B(&temp, &p);
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_CAC_TERMINATION:
		{
			// Multiap spec Section 17.2.43
			// tlv structure pointer
			struct tlv_cac_term *m;

			// calculate length of tlv structure
			m = (struct tlv_cac_term *)memory_structure;
			tlv_length = 1 + m->nbr_radio * (6 + 1 + 1);

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_I1B(&m->nbr_radio, &p);

			for (i = 0; i < m->nbr_radio; ++i) {
				_InB(&m->radio_data[i].radio_id, &p, 6);
				_I1B(&m->radio_data[i].op_class, &p);
				_I1B(&m->radio_data[i].ch, &p);
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_CAC_COMPLETION_REPORT:
		{
			// Multiap spec Section 17.2.44
			// tlv structure pointer
			struct tlv_cac_comp_report *m;

			// calculate length of tlv structure
			m = (struct tlv_cac_comp_report *)memory_structure;
			tlv_length = 1;
			for (i = 0; i < m->nbr_radios; ++i) {
				tlv_length += 6 + 1 + 1 + 1 + 1;
				for (j = 0; j < m->radio_data[i].nbr_pairs; ++j)
					tlv_length += 1 + 1;
			}

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory !");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_I1B(&m->nbr_radios, &p);
			PRINT_DBG("skip cpd");
			for (i = 0; i < m->nbr_radios; ++i) {
				_InB(&m->radio_data[i].radio_id, &p, 6);
				_I1B(&m->radio_data[i].op_class, &p);
				_I1B(&m->radio_data[i].ch, &p);
				_I1B(&m->radio_data[i].completion_status, &p);
				_I1B(&m->radio_data[i].nbr_pairs, &p);

				for (j = 0; j < m->radio_data[i].nbr_pairs; ++j) {
					_I1B(&m->radio_data[i].pair_data[j].op_class_detected, &p);
					_I1B(&m->radio_data[i].pair_data[j].ch_detected, &p);
				}
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_CAC_STATUS_REPORT:
		{
			// Multiap spec Section 17.2.45
			// tlv structure pointer
			struct tlv_cac_status_report *m;

			// calculate length of tlv structure
			m = (struct tlv_cac_status_report *)memory_structure;
			tlv_length = 1 + m->nbr_available_ch * (1 + 1 + 2);
			tlv_length += 1 + m->nbr_pairs_duration * (1 + 1 + 2);
			tlv_length += 1 + m->nbr_pairs_coundown * (1 + 1 + 3);

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			size_t nbr = 0;

			_I1B(&m->nbr_available_ch, &p);
			nbr = m->nbr_available_ch;
			for (i = 0; i < nbr; ++i) {
				_I1B(&m->ch_data[i].op_class, &p);
				_I1B(&m->ch_data[i].ch, &p);
				_I2B(&m->ch_data[i].time, &p);
			}

			_I1B(&m->nbr_pairs_duration, &p);
			nbr = m->nbr_pairs_duration;
			for (i = 0; i < nbr; ++i) {
				_I1B(&m->duration_pair_data[i].op_class, &p);
				_I1B(&m->duration_pair_data[i].ch, &p);
				_I2B(&m->duration_pair_data[i].time, &p);
			}

			_I1B(&m->nbr_pairs_coundown, &p);
			nbr = m->nbr_pairs_coundown;
			for (i = 0; i < nbr; ++i) {
				_I1B(&m->count_pair_data[i].op_class, &p);
				_I1B(&m->count_pair_data[i].ch, &p);
				_InB(&m->count_pair_data[i].time, &p, 3);
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_CAC_CAPABILITY:
		{
			// Multiap spec Section 17.2.46
			// tlv structure pointer
			struct tlv_cac_cap *m;

			// calculate length of tlv structure
			m = (struct tlv_cac_cap *)memory_structure;
			tlv_length = 2 + 1;
			for (i = 0; i < m->nbr_radios; ++i) {
				typeof(*m->radio_data) *cp = &m->radio_data[i];

				tlv_length += 6 + 1;

				for (j = 0; j < cp->nbr_cac_types; ++j) {
					typeof(*cp->cac_data) *dp = &cp->cac_data[j];

					tlv_length += 1 + 3 + 1;

					for (k = 0; k < dp->nbr_op_classes; ++k) {
						typeof(*dp->op_class_data) *op = &dp->op_class_data[k];
						tlv_length += 1 + 1 + op->nbr_ch;
					}
				}
			}

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(m->country_code, &p, 2);
			_I1B(&m->nbr_radios, &p);
			for (i = 0; i < m->nbr_radios; ++i) {
				typeof(*m->radio_data) *cp = &m->radio_data[i];

				_InB(&cp->radio_id, &p, 6);
				_I1B(&cp->nbr_cac_types, &p);

				for (j = 0; j < cp->nbr_cac_types; ++j) {
					typeof(*cp->cac_data) *dp = &cp->cac_data[j];

					_I1B(&dp->cac_method_supported, &p);
					_InB(&dp->duration, &p, 3);
					_I1B(&dp->nbr_op_classes, &p);

					for (k = 0; k < dp->nbr_op_classes; ++k) {
						typeof(*dp->op_class_data) *op = &dp->op_class_data[k];

						_I1B(&op->op_class, &p);
						_I1B(&op->nbr_ch, &p);
						_InB(op->ch, &p, op->nbr_ch);
					}
				}
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_MULTIAP_PROFILE:
		{
			// Multiap spec Section 17.2.47
			// tlv structure pointer
			struct tlv_map_profile *m;

			// calculate length of tlv structure
			m = (struct tlv_map_profile *)memory_structure;
			tlv_length = 1;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_I1B(&m->profile, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_PROFILE2_AP_CAPABILITY:
		{
			// Multiap spec Section 17.2.48
			// tlv structure pointer
			struct tlv_profile2_ap_cap *m;

			// calculate length of tlv structure
			m = (struct tlv_profile2_ap_cap *)memory_structure;
			tlv_length = 2 + 1 + 1;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_I2B(&m->reserved, &p);
			temp = 0;
			MODIFY_BITS(7, 6, temp, m->byte_counter);
			_I1B(&temp, &p);
			_I1B(&m->max_vids, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_DEFAULT_8021Q_SETTINGS:
		{
			// Multiap spec Section 17.2.49
			// tlv structure pointer
			struct tlv_default_8021q_settings *m;

			// calculate length of tlv structure
			m = (struct tlv_default_8021q_settings *)memory_structure;
			tlv_length = 2 + 1;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_I2B(&m->primary_vid, &p);
			temp = 0;
			MODIFY_BITS(7, 5, temp, m->pcp);
			_I1B(&temp, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_TRAFFIC_SEPARATION_POLICY:
		{
			// Multiap spec Section 17.2.50
			// tlv structure pointer
			struct tlv_traffic_sep_policy *m;

			// calculate length of tlv structure
			m = (struct tlv_traffic_sep_policy *)memory_structure;
			tlv_length = 1;
			for (i = 0; i < m->nbr_ssid; ++i)
				tlv_length += 1 + m->data[i].ssid_len + 2;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_I1B(&m->nbr_ssid, &p);

			for (i = 0; i < m->nbr_ssid; ++i) {
				size_t len = 0;

				_I1B(&m->data[i].ssid_len, &p);
				len = m->data[i].ssid_len;

				_InB(m->data[i].ssid, &p, len);
				_I2B(&m->data[i].vid, &p);
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_PROFILE2_ERR_CODE:
		{
			// Multiap spec Section 17.2.51
			// tlv structure pointer
			struct tlv_profile2_err_code *m;

			// calculate length of tlv structure
			m = (struct tlv_profile2_err_code *)memory_structure;
			tlv_length = 1;

			if (m->reason_code == 0x07 || m->reason_code == 0x08)
				tlv_length += 6;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_I1B(&m->reason_code, &p);
			if (m->reason_code == 0x07 || m->reason_code == 0x08)
				_InB(&m->bssid, &p, 6);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_AP_RADIO_ADV_CAPABILITY:
		{
			// Multiap spec Section 17.2.52
			// tlv structure pointer
			struct tlv_ap_radio_adv_cap *m;

			// calculate length of tlv structure
			m = (struct tlv_ap_radio_adv_cap *)memory_structure;
			tlv_length = 6 + 1 + 1;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->radio_id, &p, 6);
			temp = 0;

			MODIFY_BIT(7, temp, m->combined_front_back);
			MODIFY_BIT(6, temp, m->combined_p1_p2);
			_I1B(&temp, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_ASSOCIATION_STATUS_NOTIF:
		{
			// Multiap spec Section 17.2.53
			// tlv structure pointer
			struct tlv_assoc_status_notif *m;

			// calculate length of tlv structure
			m = (struct tlv_assoc_status_notif *)memory_structure;
			tlv_length = 1 + m->nbr_bssid * (6 + 1);

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_I1B(&m->nbr_bssid, &p);

			for (i = 0; i < m->nbr_bssid; ++i) {
				_InB(&m->data[i].bssid, &p, 6);
				_I1B(&m->data[i].association_allowance, &p);
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_SOURCE_INFO:
		{
			// Multiap spec Section 17.2.54
			// tlv structure pointer
			struct tlv_source_info *m;

			// calculate length of tlv structure
			m = (struct tlv_source_info *)memory_structure;
			tlv_length = 6;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->mac, &p, 6);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_TUNNELED_MSG_TYPE:
		{
			// Multiap spec Section 17.2.55
			// tlv structure pointer
			struct tlv_tunnel_msg_type *m;

			// calculate length of tlv structure
			m = (struct tlv_tunnel_msg_type *)memory_structure;
			tlv_length = 1;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_I1B(&m->tunnel_protocol_type, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_TUNNELED:
		{
			// Multiap spec Section 17.2.56
			// tlv structure pointer
			struct tlv_tunneled *m;

			// calculate length of tlv structure
			m = (struct tlv_tunneled *)memory_structure;
			tlv_length = m->tlv_len;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(m->frame_body, &p, tlv_length);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_PROFILE2_STEERING_REQ:
		{
			// Multiap spec Section 17.2.57
			// tlv structure pointer
			struct tlv_profile2_steer_req *m;

			// calculate length of tlv structure
			m = (struct tlv_profile2_steer_req *)memory_structure;
			tlv_length = 6 + 1  + 2 + 2 + 1 + m->sta_list_count * 6;
			tlv_length += 1 + m->target_bssid_count * (6 + 1 + 1 + 1);

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->bssid, &p, 6);
			_I1B(&m->req_mode, &p);

			MODIFY_BIT(7, temp, m->req_mode);
			MODIFY_BIT(6, temp, m->btm_imminent_bit);
			MODIFY_BIT(5, temp, m->btm_abridged_bit);

			_I1B(&temp, &p);

			_I2B(&m->steering_opt_win, &p);
			_I2B(&m->btm_dis_timer, &p);
			_I1B(&m->sta_list_count, &p);

			for (i = 0; i < m->sta_list_count; ++i)
				_InB(&m->sta_mac[i], &p, 6);

			_I1B(&m->target_bssid_count, &p);

			for (i = 0; i < m->target_bssid_count; ++i) {
				_InB(&m->target_data[i].bssid, &p, 6);
				_I1B(&m->target_data[i].bss_op_class, &p);
				_I1B(&m->target_data[i].bss_ch, &p);
				_I1B(&m->target_data[i].reason_code, &p);
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_UNSUCCESS_ASSOCIATION_POLICY:
		{
			// Multiap spec Section 17.2.58
			// tlv structure pointer
			struct tlv_unsuccess_assoc_policy *m;

			// calculate length of tlv structure
			m = (struct tlv_unsuccess_assoc_policy *)memory_structure;
			tlv_length = 1 + 4;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			temp = 0;
			MODIFY_BIT(7, temp, m->report);
			_I1B(&temp, &p);
			_I4B(&m->max_reporting_rate, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_METRIC_COLLECTION_INTERVAL:
		{
			// Multiap spec Section 17.2.59
			// tlv structure pointer
			struct tlv_metric_collection_interval *m;

			// calculate length of tlv structure
			m = (struct tlv_metric_collection_interval *)memory_structure;
			tlv_length = 4;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_I4B(&m->collection_interval, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_RADIO_METRICS:
		{
			// Multiap spec Section 17.2.60
			// tlv structure pointer
			struct tlv_radio_metrics *m;

			// calculate length of tlv structure
			m = (struct tlv_radio_metrics *)memory_structure;
			tlv_length = 6 + 1 + 1 + 1 + 1;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->radio_id, &p, 6);
			_I1B(&m->noise, &p);
			_I1B(&m->transmit, &p);
			_I1B(&m->receive_self, &p);
			_I1B(&m->receive_other, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_AP_EXTENDED_METRICS:
		{
			// Multiap spec Section 17.2.61
			// tlv structure pointer
			struct tlv_ap_ext_metrics *m;

			// calculate length of tlv structure
			m = (struct tlv_ap_ext_metrics *)memory_structure;
			tlv_length = 6 + (6 * 4);

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->bssid, &p, 6);
			_I4B(&m->uni_bytes_sent, &p);
			_I4B(&m->uni_bytes_recv, &p);
			_I4B(&m->multi_bytes_sent, &p);
			_I4B(&m->multi_bytes_recv, &p);
			_I4B(&m->bro_bytes_sent, &p);
			_I4B(&m->bro_bytes_recv, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_ASSOCIATED_STA_EXT_LINK_METRICS:
		{
			// Multiap spec Section 17.2.62
			// tlv structure pointer
			struct tlv_assoc_sta_ext_link_metric *m;

			// calculate length of tlv structure
			m = (struct tlv_assoc_sta_ext_link_metric *)memory_structure;
			tlv_length = 6 + 1 + m->nbr_bssid * (6 + (4 * 4));

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->mac, &p, 6);
			_I1B(&m->nbr_bssid, &p);

			for (i = 0; i < m->nbr_bssid; ++i) {
				_InB(&m->data[i].bssid, &p, 6);
				_I4B(&m->data[i].last_data_dwn_rate, &p);
				_I4B(&m->data[i].last_data_up_rate, &p);
				_I4B(&m->data[i].uti_recv, &p);
				_I4B(&m->data[i].uti_transmit, &p);
			}

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_STATUS_CODE:
		{
			// tlv structure pointer
			// Multiap spec Section 17.2.63
			struct tlv_status_code *m;

			// calculate length of tlv structure
			m = (struct tlv_status_code *)memory_structure;
			tlv_length = 2;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_I2B(&m->status_code, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_REASON_CODE:
		{
			// Multiap spec Section 17.2.64
			// tlv structure pointer
			struct tlv_reason_code *m;

			// calculate length of tlv structure
			m = (struct tlv_reason_code *)memory_structure;
			tlv_length = 2;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_I2B(&m->reason_code, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_BACKHAUL_STA_RADIO_CAPABILITY:
		{
			// Multiap spec Section 17.2.65
			// tlv structure pointer
			struct tlv_backhaul_sta_radio_cap *m;

			// calculate length of tlv structure
			m = (struct tlv_backhaul_sta_radio_cap *)memory_structure;
			tlv_length = 6 + 1;

			if (m->mac_include)
				tlv_length += 6;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->radio_id, &p, 6);
			temp = 0;
			MODIFY_BIT(7, temp, m->mac_include);
			_I1B(&temp, &p);

			if (m->mac_include)
				_InB(&m->mac, &p, 6);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	case MAP_TLV_BACKHAUL_BSS_CONFIG:
		{
			// Multiap spec Section 17.2.66
			// tlv structure pointer
			struct tlv_backhaul_bss_config *m;

			// calculate length of tlv structure
			m = (struct tlv_backhaul_bss_config *)memory_structure;
			tlv_length = 6 + 1;

			ret = calloc((1 + 2 + tlv_length), sizeof(uint8_t));
			if (!ret) {
				PRINT_ERR("Insufficient memory!");
				return NULL;
			}

			p = ret;
			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			_InB(&m->bssid, &p, 6);

			MODIFY_BIT(7, temp, m->p1);
			MODIFY_BIT(6, temp, m->p2);
			_I1B(&temp, &p);

			if (len)
				*len = 1 + 2 + tlv_length;
			return (uint8_t *) ret;
		}
	default:
		{
			PRINT_ERR("Unknown TLV|0x%02x|", *memory_structure);
		}

	}
	return NULL;
}

static void _free_tlv_cstruct(uint8_t *memory_structure, bool all)
{
	size_t i, j, k;

	if (memory_structure == NULL)
		return;

	switch (*memory_structure) {
		/* 1905 related tlvs */
	case TLV_TYPE_END_OF_MESSAGE:
	case TLV_TYPE_AL_MAC_ADDRESS_TYPE:
	case TLV_TYPE_MAC_ADDRESS_TYPE:
	case TLV_TYPE_SEARCHED_ROLE:
	case TLV_TYPE_AUTOCONFIG_FREQ_BAND:
	case TLV_TYPE_SUPPORTED_ROLE:
	case TLV_TYPE_SUPPORTED_FREQ_BAND:
		{
			if (all)
				free(memory_structure);

			return;
		}

	case TLV_TYPE_DEVICE_INFORMATION_TYPE:
		{
			struct tlv_device_info *m;

			m = (struct tlv_device_info *)memory_structure;

			if (m->local_interfaces_nr > 0 && NULL != m->local_interfaces)
				free(m->local_interfaces);

			if (all)
				free(m);

			return;
		}

	case TLV_TYPE_VENDOR_SPECIFIC:
		{
			struct tlv_vendor_specific *m;

			m = (struct tlv_vendor_specific *)memory_structure;

			if (m->m_nr > 0)
				free(m->m);

			if (all)
				free(m);

			return;
		}

	case TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES:
		{
			struct tlv_device_bridge_cap *m;

			m = (struct tlv_device_bridge_cap *)memory_structure;

			for (i = 0; i < m->bridging_tuples_nr; i++) {
				if (m->bridging_tuples[i].bridging_tuple_macs_nr > 0 &&
				    m->bridging_tuples[i].bridging_tuple_macs != NULL)
					free(m->bridging_tuples[i].bridging_tuple_macs);
			}
			if (m->bridging_tuples_nr > 0 &&
			    m->bridging_tuples != NULL)
				free(m->bridging_tuples);

			if (all)
				free(m);

			return;
		}

	case TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST:
		{
			struct tlv_non1905_neighbor *m;

			m = (struct tlv_non1905_neighbor *)memory_structure;

			if (m->non_1905_neighbors_nr > 0 &&
			    m->non_1905_neighbors != NULL)
				free(m->non_1905_neighbors);

			if (all)
				free(m);

			return;
		}

	case TLV_TYPE_NEIGHBOR_DEVICE_LIST:
		{
			struct tlv_neighbor_device *m;

			m = (struct tlv_neighbor_device *)memory_structure;

			if (m->neighbors_nr)
				free(m->neighbors);

			if (all)
				free(m);

			return;
		}

	case TLV_TYPE_TRANSMITTER_LINK_METRIC:
		{
			struct tlv_tx_link_metric *m;

			m = (struct tlv_tx_link_metric *)memory_structure;

			if (m->transmitter_link_metrics_nr > 0 &&
			    m->transmitter_link_metrics != NULL)
				free(m->transmitter_link_metrics);

			if (all)
				free(m);

			return;
		}

	case TLV_TYPE_RECEIVER_LINK_METRIC:
		{
			struct tlv_rx_link_metric *m;

			m = (struct tlv_rx_link_metric *)memory_structure;

			if (m->receiver_link_metrics_nr > 0 &&
			    m->receiver_link_metrics != NULL)
				free(m->receiver_link_metrics);

			if (all)
				free(m);

			return;
		}

	case TLV_TYPE_WSC:
		{
			struct tlv_wsc *m;

			m = (struct tlv_wsc *)memory_structure;

			if (m->wsc_frame_size > 0 && m->wsc_frame != NULL)
				free(m->wsc_frame);

			if (all)
				free(m);

			return;
		}
		/* map related tlvs */
	case MAP_TLV_CLIENT_ASSOCIATION_EVENT:
	case MAP_TLV_AP_RADIO_IDENTIFIER:
	case MAP_TLV_AP_CAPABILITY:
	case MAP_TLV_AP_HT_CAPABILITIES:
	case MAP_TLV_AP_VHT_CAPABILITIES:
	case MAP_TLV_TRANSMIT_POWER_LIMIT:
	case MAP_TLV_CHANNEL_SELECTION_RESPONSE:
	case MAP_TLV_CLIENT_INFO:
	case MAP_TLV_AP_METRICS:
	case MAP_TLV_STA_MAC_ADDRESS:
	case MAP_TLV_STEERING_BTM_REPORT:
	case MAP_TLV_BACKHAUL_STEERING_REQUEST:
	case MAP_TLV_BACKHAUL_STEERING_RESPONSE:
	case MAP_TLV_ASSOCIATED_STA_TRAFFIC_STATS:
	case MAP_TLV_ERROR_CODE:
	case MAP_TLV_CHANNEL_SCAN_REPORTING_POLICY:
	case MAP_TLV_UNSUCCESS_ASSOCIATION_POLICY:
	case MAP_TLV_METRIC_COLLECTION_INTERVAL:
	case MAP_TLV_RADIO_METRICS:
	case MAP_TLV_AP_EXTENDED_METRICS:
	case MAP_TLV_SOURCE_INFO:
	case MAP_TLV_TUNNELED_MSG_TYPE:
	case MAP_TLV_PROFILE2_ERR_CODE:
	case MAP_TLV_AP_RADIO_ADV_CAPABILITY:
	case MAP_TLV_MULTIAP_PROFILE:
	case MAP_TLV_STATUS_CODE:
	case MAP_TLV_REASON_CODE:
	case MAP_TLV_BACKHAUL_STA_RADIO_CAPABILITY:
	case MAP_TLV_BACKHAUL_BSS_CONFIG:
	case MAP_TLV_PROFILE2_AP_CAPABILITY:
	case MAP_TLV_DEFAULT_8021Q_SETTINGS:
		{
			if (all)
				free(memory_structure);
			break;
		}

	case MAP_TLV_CLIENT_CAPABILITY_REPORT:
		{
			struct tlv_client_cap_report *m;

			m = (struct tlv_client_cap_report *)memory_structure;

			if(m->result_code == 0x00)
				free(m->frame_body);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_SEARCHED_SERVICE:
		{
			struct tlv_searched_service *m;

			m = (struct tlv_searched_service *)memory_structure;

			if (m->searched_services_list)
				free(m->service);

			if (all)
				free(m);

			break;
		}
	case MAP_TLV_AP_RADIO_BASIC_CAPABILITIES:
		{
			struct tlv_ap_radio_basic_cap *m;

			m = (struct tlv_ap_radio_basic_cap *)memory_structure;

			if (m->operating_classes_nr > 0) {
				for (i = 0; i < m->operating_classes_nr; ++i)
					if (m->operating_class[i].non_op_ch_nr > 0)
						free(m->operating_class[i].channel);

				free(m->operating_class);
			}

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_AP_HE_CAPABILITIES:
		{
			struct tlv_ap_he_cap *m;

			m = (struct tlv_ap_he_cap *)memory_structure;

			if (m->he_mcs_len)
				free(m->he_mcs_supported);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_STEERING_POLICY:
		{
			struct tlv_steering_policy *m;

			m = (struct tlv_steering_policy *)memory_structure;

			if (m->local_disallowed_sta_nr)
				free(m->local_disallowed_sta_macs);

			if (m->btm_disallowed_sta_nr)
				free(m->btm_disallowed_sta_macs);

			if (m->control_policy_radio_nr)
				free(m->control_policy);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_METRIC_REPORTING_POLICY:
		{
			struct tlv_metric_report_policy *m;

			m = (struct tlv_metric_report_policy *)memory_structure;

			if (m->metric_reporting_policy_radio_nr)
				free(m->metric_reporting_policy);
			if (all)
				free(m);
			break;
		}
	case MAP_TLV_CHANNEL_PREFERENCE:
		{
			struct tlv_channel_pref *m;

			m = (struct tlv_channel_pref *)memory_structure;

			for (i = 0; i < m->ch_preference_op_class_nr; ++i)
				if (m->op_class[i].channel_nr)
					free(m->op_class[i].channel_list);

			if (m->ch_preference_op_class_nr)
				free(m->op_class);
			if (all)
				free(m);
			break;
		}
	case MAP_TLV_RADIO_OPERATION_RESTRICTION:
		{
			struct tlv_radio_oper_restrict *m;

			m = (struct tlv_radio_oper_restrict *)memory_structure;

			for (i = 0; i < m->op_restricted_op_class_nr; ++i) {
				if (m->restricted_op_class[i].channel_nr)
					free(m->restricted_op_class[i].restricted_channels);
			}

			if (m->op_restricted_op_class_nr)
				free(m->restricted_op_class);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_OPERATING_CHANNEL_REPORT:
		{
			struct tlv_oper_ch_report *m;

			m = (struct tlv_oper_ch_report *)memory_structure;

			if (m->op_ch_op_class_nr)
				free(m->op_ch_op_class);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_AP_METRIC_QUERY:
		{
			struct tlv_ap_metric_query *m;

			m = (struct tlv_ap_metric_query *)memory_structure;

			if (m->bssid_nr)
				free(m->ap_metric_query_bssid);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_ASSOCIATED_STA_LINK_METRICS:
		{
			struct tlv_assoc_sta_link_metrics *m;

			m = (struct tlv_assoc_sta_link_metrics *)memory_structure;

			if (m->bssid_nr)
				free(m->sta_link_metrics_bssid);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_UNASSOCIATED_STA_LINK_METRICS_QUERY:
		{
			struct tlv_unassoc_sta_link_metrics_query *m;

			m = (struct tlv_unassoc_sta_link_metrics_query *)memory_structure;

			for (i = 0; i < m->ch_nr; ++i) {
				if (m->unassoc_sta_link_metrics_query_ch[i].sta_mac_nr)
					free(m->unassoc_sta_link_metrics_query_ch[i].sta_mac);

			}
			if (m->ch_nr > 0)
				free(m->unassoc_sta_link_metrics_query_ch);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_UNASSOCIATED_STA_LINK_METRICS_RESPONSE:
		{
			struct tlv_unassoc_sta_link_metrics_resp *m;

			m = (struct tlv_unassoc_sta_link_metrics_resp *)memory_structure;

			if (m->sta_nr > 0)
				free(m->unassoc_sta_link_metrics_sta);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_BEACON_METRICS_QUERY:
		{
			struct tlv_beacon_metrics_query *m;

			m = (struct tlv_beacon_metrics_query *)memory_structure;

			for (i = 0; i < m->ch_reports_nr; ++i) {
				if (m->ch_report[i].op_class)
					free(m->ch_report[i].ch_list);
			}
			if (m->ch_reports_nr)
				free(m->ch_report);

			if (m->elem_ids_nr)
				free(m->elem_list);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_BEACON_METRICS_RESPONSE:
		{
			struct tlv_beacon_metrics_resp *m;

			m = (struct tlv_beacon_metrics_resp *)memory_structure;

			if (m->report_elems_nr > 0)
				free(m->report_elem);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_STEERING_REQUEST:
		{
			struct tlv_steer_Req *m;

			m = (struct tlv_steer_Req *)memory_structure;

			if (m->sta_list_cnt)
				free(m->steering_req_macs);

			if (m->target_bssid_list_cnt)
				free(m->steering_req_target_bssids);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_CLIENT_ASSOCIATION_CONTROL_REQUEST:
		{
			struct tlv_client_assoc_control_req *m;

			m = (struct tlv_client_assoc_control_req *)memory_structure;

			if (m->sta_list_cnt)
				free(m->client_assoc_ctrl_req_stas);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_HIGHER_LAYER_DATA:
		{
			struct tlv_higher_layer_data *m;

			m = (struct tlv_higher_layer_data *)memory_structure;

			if (m->tlv_len)
				free(m->payload);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_SUPPORTED_SERVICE:
		{
			struct tlv_supp_service *m;

			m = (struct tlv_supp_service *)memory_structure;

			if (m->supported_services_list)
				free(m->supported_services);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_AP_OPERATIONAL_BSS:
		{
			struct tlv_ap_oper_bss *m;

			m = (struct tlv_ap_oper_bss *)memory_structure;

			for (i = 0; i < m->radios_nr; i++) {
				if (m->radio[i].bss_nr)
					free(m->radio[i].bss);
			}
			if (m->radios_nr)
				free(m->radio);

			if (all)
				free(m);
			break;
		}

	case MAP_TLV_ASSOCIATED_CLIENTS:
		{
			struct tlv_assoc_client *m;

			m = (struct tlv_assoc_client *)memory_structure;

			for (i = 0; i < m->bss_nr; i++) {
				if (m->bss[i].assoc_clients_nr)
					free(m->bss[i].clients);
			}
			if (m->bss_nr)
				free(m->bss);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_CHANNEL_SCAN_CAPABILITY:
		{
			struct tlv_ch_scan_cap *m;

			m = (struct tlv_ch_scan_cap *)memory_structure;

			for (i = 0; i < m->nbr_radios; ++i) {
				for (j = 0; j < m->radio[i].nbr_op_classes; ++j)
					if (m->radio[i].op_class[j].nbr_ch)
						free(m->radio[i].op_class[j].ch);

				if (m->radio[i].nbr_op_classes)
					free(m->radio[i].op_class);
			}
			if (m->nbr_radios)
				free(m->radio);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_CHANNEL_SCAN_REQ:
		{
			struct tlv_ch_scan_req *m;

			m = (struct tlv_ch_scan_req *)memory_structure;

			for (i = 0; i < m->nbr_radios; ++i) {
				for (j = 0; j < m->radio_data[i].nbr_op_class; ++j)
					if (m->radio_data[i].op_class_data[j].nbr_ch)
						free(m->radio_data[i].op_class_data[j].ch);

				if (m->radio_data[i].nbr_op_class)
					free(m->radio_data[i].op_class_data);
			}
			if (m->nbr_radios)
				free(m->radio_data);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_CHANNEL_SCAN_RES:
		{
			struct tlv_ch_scan_res *m;

			m = (struct tlv_ch_scan_res *)memory_structure;

			if (m->scan_res == CH_SCAN_STATUS_SUCCESS) {
				if (m->time_len)
					free(m->timestamp);

				for (i = 0; i < m->nbr_neighbors; ++i) {
					if (m->neighbor_data[i].ssid_len)
						free(m->neighbor_data[i].ssid);

					if (m->neighbor_data[i].ch_band_len)
						free(m->neighbor_data[i].ch_bandwidth);
				}
				if (m->nbr_neighbors)
					free(m->neighbor_data);
			}

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_TIMESTAMP:
		{
			struct tlv_timestamp *m;

			m = (struct tlv_timestamp *)memory_structure;

			if (m->time_len)
					free(m->time);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_CAC_REQ:
		{
			struct tlv_cac_req *m;

			m = (struct tlv_cac_req *)memory_structure;

			if (m->nbr_radios)
					free(m->radio_data);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_CAC_TERMINATION:
		{
			struct tlv_cac_term *m;

			m = (struct tlv_cac_term *)memory_structure;

			if (m->nbr_radio)
					free(m->radio_data);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_CAC_COMPLETION_REPORT:
		{
			struct tlv_cac_comp_report *m;

			m = (struct tlv_cac_comp_report *)memory_structure;

			for (i = 0; i < m->nbr_radios; ++i)
					if (m->radio_data[i].nbr_pairs)
							free(m->radio_data[i].pair_data);

			if (m->nbr_radios)
					free(m->radio_data);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_CAC_STATUS_REPORT:
		{
			struct tlv_cac_status_report *m;

			m = (struct tlv_cac_status_report *)memory_structure;

			if (m->nbr_available_ch)
					free(m->ch_data);

			if (m->nbr_pairs_duration)
					free(m->duration_pair_data);

			if (m->nbr_pairs_coundown)
					free(m->count_pair_data);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_CAC_CAPABILITY:
		{
			struct tlv_cac_cap *m;

			m = (struct tlv_cac_cap *)memory_structure;

			for (i = 0; i < m->nbr_radios; ++i) {
				typeof(*m->radio_data) *cp = &m->radio_data[i];

				for (j = 0; j < cp->nbr_cac_types; ++j) {
					typeof(*cp->cac_data) *dp = &cp->cac_data[j];

					for (k = 0; k < dp->nbr_op_classes; ++k) {
						typeof(*dp->op_class_data) *op = &dp->op_class_data[k];

						if (op->nbr_ch)
								free(op->ch);
					}
					if (dp->nbr_op_classes)
							free(dp->op_class_data);
				}
				if (cp->nbr_cac_types)
						free(cp->cac_data);
			}
			if (m->nbr_radios)
					free(m->radio_data);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_TRAFFIC_SEPARATION_POLICY:
		{
			struct tlv_traffic_sep_policy *m;

			m = (struct tlv_traffic_sep_policy *)memory_structure;

			for (i = 0; i < m->nbr_ssid; ++i)
					if (m->data[i].ssid_len)
							free(m->data[i].ssid);

			if (m->nbr_ssid)
					free(m->data);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_ASSOCIATION_STATUS_NOTIF:
		{
			struct tlv_assoc_status_notif *m;

			m = (struct tlv_assoc_status_notif *)memory_structure;

			if (m->nbr_bssid)
					free(m->data);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_TUNNELED:
		{
			struct tlv_tunneled *m;

			m = (struct tlv_tunneled *)memory_structure;

			if (m->tlv_len)
					free(m->frame_body);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_PROFILE2_STEERING_REQ:
		{
			struct tlv_profile2_steer_req *m;

			m = (struct tlv_profile2_steer_req *)memory_structure;

			if (m->sta_list_count)
					free(m->sta_mac);

			if (m->target_bssid_count)
					free(m->target_data);

			if (all)
				free(m);
			break;
		}
	case MAP_TLV_ASSOCIATED_STA_EXT_LINK_METRICS:
		{
			struct tlv_assoc_sta_ext_link_metric *m;

			m = (struct tlv_assoc_sta_ext_link_metric *)memory_structure;

			if (m->nbr_bssid > 0)
				free(m->data);

			if (all)
				free(m);
			break;
		}
	default:
		{
			PRINT_ERR("Free not defined for |0x%02x|", *memory_structure);
			break;
		}
	}
}

void map_free_tlv_cstruct(uint8_t *tlv)
{
	_free_tlv_cstruct(tlv, true);
}

void tlv_free_internal_members(uint8_t *tlv)
{
	_free_tlv_cstruct(tlv, false);
}

bool map_compare_tlv_cstructs(uint8_t *memory_structure_1, uint8_t *memory_structure_2)
{
	bool ret = false;
	uint8_t *tlv1_stream, *tlv2_stream;
	uint16_t tlv1_len, tlv2_len;

	if (NULL == memory_structure_1 || NULL == memory_structure_2)
		return false;

	if (*memory_structure_1 != *memory_structure_2)
		return false;

	tlv1_stream = map_put_tlv_cstruct(memory_structure_1, &tlv1_len);
	tlv2_stream = map_put_tlv_cstruct(memory_structure_2, &tlv2_len);

	if ((tlv1_len == tlv2_len) &&
	    (0 == memcmp(tlv1_stream, tlv2_stream, tlv1_len)))
		ret = true;

	free(tlv1_stream);
	free(tlv2_stream);

	return ret;
}

char *map_stringify_tlv_type(uint8_t tlv_type)
{
	switch (tlv_type) {
		/* 1905 related tlvs */
	case TLV_TYPE_END_OF_MESSAGE:
		return "TLV_TYPE_END_OF_MESSAGE";
	case TLV_TYPE_AL_MAC_ADDRESS_TYPE:
		return "TLV_TYPE_AL_MAC_ADDRESS_TYPE";
	case TLV_TYPE_DEVICE_INFORMATION_TYPE:
		return "TLV_TYPE_DEVICE_INFORMATION_TYPE";
	case TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES:
		return "TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES";
	case TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST:
		return "TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST";
	case TLV_TYPE_NEIGHBOR_DEVICE_LIST:
		return "TLV_TYPE_NEIGHBOR_DEVICE_LIST";
	case TLV_TYPE_TRANSMITTER_LINK_METRIC:
		return "TLV_TYPE_TRANSMITTER_LINK_METRIC";
	case TLV_TYPE_RECEIVER_LINK_METRIC:
		return "TLV_TYPE_RECEIVER_LINK_METRIC";
	case TLV_TYPE_SEARCHED_ROLE:
		return "TLV_TYPE_SEARCHED_ROLE";
	case TLV_TYPE_AUTOCONFIG_FREQ_BAND:
		return "TLV_TYPE_AUTOCONFIG_FREQ_BAND";
	case TLV_TYPE_SUPPORTED_ROLE:
		return "TLV_TYPE_SUPPORTED_ROLE";
	case TLV_TYPE_SUPPORTED_FREQ_BAND:
		return "TLV_TYPE_SUPPORTED_FREQ_BAND";
	case TLV_TYPE_WSC:
		return "TLV_TYPE_WSC";
		/* map profile-1 tlvs */
	case MAP_TLV_SUPPORTED_SERVICE:
		return "MAP_TLV_SUPPORTED_SERVICE";
	case MAP_TLV_SEARCHED_SERVICE:
		return "MAP_TLV_SEARCHED_SERVICE";
	case MAP_TLV_AP_RADIO_IDENTIFIER:
		return "MAP_TLV_AP_RADIO_IDENTIFIER";
	case MAP_TLV_AP_OPERATIONAL_BSS:
		return "MAP_TLV_AP_OPERATIONAL_BSS";
	case MAP_TLV_ASSOCIATED_CLIENTS:
		return "MAP_TLV_ASSOCIATED_CLIENTS";
	case MAP_TLV_AP_CAPABILITY:
		return "MAP_TLV_AP_CAPABILITY";
	case MAP_TLV_AP_RADIO_BASIC_CAPABILITIES:
		return "MAP_TLV_AP_RADIO_BASIC_CAPABILITIES";
	case MAP_TLV_AP_HT_CAPABILITIES:
		return "MAP_TLV_AP_HT_CAPABILITIES";
	case MAP_TLV_AP_VHT_CAPABILITIES:
		return "MAP_TLV_AP_VHT_CAPABILITIES";
	case MAP_TLV_AP_HE_CAPABILITIES:
		return "MAP_TLV_AP_HE_CAPABILITIES";
	case MAP_TLV_STEERING_POLICY:
		return "MAP_TLV_STEERING_POLICY";
	case MAP_TLV_METRIC_REPORTING_POLICY:
		return "MAP_TLV_METRIC_REPORTING_POLICY";
	case MAP_TLV_CHANNEL_PREFERENCE:
		return "MAP_TLV_CHANNEL_PREFERENCE";
	case MAP_TLV_RADIO_OPERATION_RESTRICTION:
		return "MAP_TLV_RADIO_OPERATION_RESTRICTION";
	case MAP_TLV_TRANSMIT_POWER_LIMIT:
		return "MAP_TLV_TRANSMIT_POWER_LIMIT";
	case MAP_TLV_CHANNEL_SELECTION_RESPONSE:
		return "MAP_TLV_CHANNEL_SELECTION_RESPONSE";
	case MAP_TLV_OPERATING_CHANNEL_REPORT:
		return "MAP_TLV_OPERATING_CHANNEL_REPORT";
	case MAP_TLV_CLIENT_INFO:
		return "MAP_TLV_CLIENT_INFO";
	case MAP_TLV_CLIENT_CAPABILITY_REPORT:
		return "MAP_TLV_CLIENT_CAPABILITY_REPORT";
	case MAP_TLV_CLIENT_ASSOCIATION_EVENT:
		return "MAP_TLV_CLIENT_ASSOCIATION_EVENT";
	case MAP_TLV_AP_METRIC_QUERY:
		return "MAP_TLV_AP_METRIC_QUERY";
	case MAP_TLV_AP_METRICS:
		return "MAP_TLV_AP_METRICS";
	case MAP_TLV_STA_MAC_ADDRESS:
		return "MAP_TLV_STA_MAC_ADDRESS";
	case MAP_TLV_ASSOCIATED_STA_LINK_METRICS:
		return "MAP_TLV_ASSOCIATED_STA_LINK_METRICS";
	case MAP_TLV_UNASSOCIATED_STA_LINK_METRICS_QUERY:
		return "MAP_TLV_UNASSOCIATED_STA_LINK_METRICS_QUERY";
	case MAP_TLV_UNASSOCIATED_STA_LINK_METRICS_RESPONSE:
		return "MAP_TLV_UNASSOCIATED_STA_LINK_METRICS_RESPONSE";
	case MAP_TLV_BEACON_METRICS_QUERY:
		return "MAP_TLV_BEACON_METRICS_QUERY";
	case MAP_TLV_BEACON_METRICS_RESPONSE:
		return "MAP_TLV_BEACON_METRICS_RESPONSE";
	case MAP_TLV_STEERING_REQUEST:
		return "MAP_TLV_STEERING_REQUEST";
	case MAP_TLV_STEERING_BTM_REPORT:
		return "MAP_TLV_STEERING_BTM_REPORT";
	case MAP_TLV_CLIENT_ASSOCIATION_CONTROL_REQUEST:
		return "MAP_TLV_CLIENT_ASSOCIATION_CONTROL_REQUEST";
	case MAP_TLV_BACKHAUL_STEERING_REQUEST:
		return "MAP_TLV_BACKHAUL_STEERING_REQUEST";
	case MAP_TLV_BACKHAUL_STEERING_RESPONSE:
		return "MAP_TLV_BACKHAUL_STEERING_RESPONSE";
	case MAP_TLV_HIGHER_LAYER_DATA:
		return "MAP_TLV_HIGHER_LAYER_DATA";
	case MAP_TLV_ASSOCIATED_STA_TRAFFIC_STATS:
		return "MAP_TLV_ASSOCIATED_STA_TRAFFIC_STATS";
	case MAP_TLV_ERROR_CODE:
		return "MAP_TLV_ERROR_CODE";
		/* map profile 2 type tlvs */
	case MAP_TLV_CHANNEL_SCAN_REPORTING_POLICY:
		return "MAP_TLV_CHANNEL_SCAN_REPORTING_POLICY";
	case MAP_TLV_CHANNEL_SCAN_CAPABILITY:
		return "MAP_TLV_CHANNEL_SCAN_CAPABILITY";
	case MAP_TLV_CHANNEL_SCAN_REQ:
		return "MAP_TLV_CHANNEL_SCAN_REQ";
	case MAP_TLV_CHANNEL_SCAN_RES:
		return "MAP_TLV_CHANNEL_SCAN_RES";
	case MAP_TLV_TIMESTAMP:
		return "MAP_TLV_TIMESTAMP";
	case MAP_TLV_CAC_REQ:
		return "MAP_TLV_CAC_REQ";
	case MAP_TLV_CAC_TERMINATION:
		return "MAP_TLV_CAC_TERMINATION";
	case MAP_TLV_CAC_COMPLETION_REPORT:
		return "MAP_TLV_CAC_COMPLETION_REPORT";
	case MAP_TLV_CAC_STATUS_REPORT:
		return "MAP_TLV_CAC_STATUS_REPORT";
	case MAP_TLV_CAC_CAPABILITY:
		return "MAP_TLV_CAC_CAPABILITY";
	case MAP_TLV_MULTIAP_PROFILE:
		return "MAP_TLV_MULTIAP_PROFILE";
	case MAP_TLV_PROFILE2_AP_CAPABILITY:
		return "MAP_TLV_PROFILE2_AP_CAPABILITY";
	case MAP_TLV_DEFAULT_8021Q_SETTINGS:
		return "MAP_TLV_DEFAULT_8021Q_SETTINGS";
	case MAP_TLV_TRAFFIC_SEPARATION_POLICY:
		return "MAP_TLV_TRAFFIC_SEPARATION_POLICY";
	case MAP_TLV_PROFILE2_ERR_CODE:
		return "MAP_TLV_PROFILE2_ERR_CODE";
	case MAP_TLV_AP_RADIO_ADV_CAPABILITY:
		return "MAP_TLV_AP_RADIO_ADV_CAPABILITY";
	case MAP_TLV_ASSOCIATION_STATUS_NOTIF:
		return "MAP_TLV_ASSOCIATION_STATUS_NOTIF";
	case MAP_TLV_SOURCE_INFO:
		return "MAP_TLV_SOURCE_INFO";
	case MAP_TLV_TUNNELED_MSG_TYPE:
		return "MAP_TLV_TUNNELED_MSG_TYPE";
	case MAP_TLV_TUNNELED:
		return "MAP_TLV_TUNNELED";
	case MAP_TLV_PROFILE2_STEERING_REQ:
		return "MAP_TLV_PROFILE2_STEERING_REQ";
	case MAP_TLV_UNSUCCESS_ASSOCIATION_POLICY:
		return "MAP_TLV_UNSUCCESS_ASSOCIATION_POLICY";
	case MAP_TLV_METRIC_COLLECTION_INTERVAL:
		return "MAP_TLV_METRIC_COLLECTION_INTERVAL";
	case MAP_TLV_RADIO_METRICS:
		return "MAP_TLV_RADIO_METRICS";
	case MAP_TLV_AP_EXTENDED_METRICS:
		return "MAP_TLV_AP_EXTENDED_METRICS";
	case MAP_TLV_ASSOCIATED_STA_EXT_LINK_METRICS:
		return "MAP_TLV_ASSOCIATED_STA_EXT_LINK_METRICS";
	case MAP_TLV_STATUS_CODE:
		return "MAP_TLV_STATUS_CODE";
	case MAP_TLV_REASON_CODE:
		return "MAP_TLV_REASON_CODE";
	case MAP_TLV_BACKHAUL_STA_RADIO_CAPABILITY:
		return "MAP_TLV_BACKHAUL_STA_RADIO_CAPABILITY";
	case MAP_TLV_BACKHAUL_BSS_CONFIG:
		return "MAP_TLV_BACKHAUL_BSS_CONFIG";
	default:
		PRINT_ERR("Unknown TLV type(0x%x)", tlv_type);
	}

	return "UNKNOWN";
}

uint16_t map_get_first_tlv_length(uint8_t *tlv_stream)
{
	uint8_t *p;
	uint8_t type;
	uint16_t len;

	if (tlv_stream == NULL)
		return 0;

	p = tlv_stream;
	_E1B(&p, &type);
	_E2B(&p, &len);

	// tlv packet does not have length of tlv_type(1 Byte) and
	// tlv_len(2 Bytes) included
	return len + 3;
}
